





#include "rtl8305s.h"
#include "assert.h"
#include "asicRegs.h"

#define RTL8305S_BASIC_DRIVER


void _rtl8305s_smiZBit(void) {
	uint32 i;
	REG32(PABCDIR) = (REG32(PABCDIR)& 0x3FFFFFFF) | 0x80000000;
	REG32(PABCDAT) = (REG32(PABCDAT) & 0x3FFFFFFF);
	for(i=0; i<25; i++);
}


void _rtl8305s_smiReadBit(uint8 * data) {
	uint32 i;
	REG32(PABCDIR) = (REG32(PABCDIR)& 0x3FFFFFFF) | 0x80000000;
	REG32(PABCDAT) = (REG32(PABCDAT) & 0x3FFFFFFF) | 0x80000000;
	for(i=0; i<25; i++);
	REG32(PABCDAT) = (REG32(PABCDAT) & 0x3FFFFFFF);
	*data = (REG32(PABCDAT) & 0x40000000)?1:0;
}


void _rtl8305s_smiWriteBit(uint8 data) {
	uint32 i;
	
	REG32(PABCDIR) = REG32(PABCDIR) | 0xC0000000;
	if(data) {
		REG32(PABCDAT) = (REG32(PABCDAT) & 0x3FFFFFFF) | 0x40000000;
		for(i=0; i<25; i++);
		REG32(PABCDAT) = (REG32(PABCDAT) & 0x3FFFFFFF) | 0xC0000000;
	} else {
		REG32(PABCDAT) = (REG32(PABCDAT) & 0x3FFFFFFF);
		for(i=0; i<25; i++);
		REG32(PABCDAT) = (REG32(PABCDAT) & 0x3FFFFFFF) | 0x80000000;
	}
}


void rtl8305s_smiRead(uint8 phyad, uint8 regad, uint16 * data) {
	int32 i;
	uint8 readBit;

	
	REG32(PABCCNR) = REG32(PABCCNR) & 0x3FFFFFFF;
	REG32(PABIMR) = REG32(PABIMR) & 0xFFFFFFF;
	
	for(i=0; i<32; i++)
		_rtl8305s_smiWriteBit(1);
	
	_rtl8305s_smiWriteBit(0);
	_rtl8305s_smiWriteBit(1);
	
	_rtl8305s_smiWriteBit(1);
	_rtl8305s_smiWriteBit(0);
	
	for(i=4; i>=0; i--) 
		_rtl8305s_smiWriteBit((phyad>>i)&0x1);
	
	for(i=4; i>=0; i--) 
		_rtl8305s_smiWriteBit((regad>>i)&0x1);
	
	_rtl8305s_smiZBit();
	_rtl8305s_smiReadBit(&readBit);
	
	*data = 0;
	for(i=15; i>=0; i--) {
		_rtl8305s_smiReadBit(&readBit);
		*data = (*data<<1) | readBit;
	}
	_rtl8305s_smiZBit();
}


void rtl8305s_smiWrite(uint8 phyad, uint8 regad, uint16 data) {
	int32 i;

	
	REG32(PABCCNR) = REG32(PABCCNR) & 0x3FFFFFFF;
	REG32(PABIMR) = REG32(PABIMR) & 0xFFFFFFF;
	
	for(i=0; i<32; i++)
		_rtl8305s_smiWriteBit(1);
	
	_rtl8305s_smiWriteBit(0);
	_rtl8305s_smiWriteBit(1);
	
	_rtl8305s_smiWriteBit(0);
	_rtl8305s_smiWriteBit(1);
	
	for(i=4; i>=0; i--) 
		_rtl8305s_smiWriteBit((phyad>>i)&0x1);
	
	for(i=4; i>=0; i--) 
		_rtl8305s_smiWriteBit((regad>>i)&0x1);
	
	_rtl8305s_smiWriteBit(1);
	_rtl8305s_smiWriteBit(0);
	
	for(i=15; i>=0; i--) 
		_rtl8305s_smiWriteBit((data>>i)&0x1);
	_rtl8305s_smiZBit();
}


void rtl8305s_smiReadBit(uint8 phyad, uint8 regad, uint8 bit, uint8 * value) {
	uint16 regData;

	if(value == NULL)
		return;
	if(bit>=16)
		* value = 0;
	else {
		rtl8305s_smiRead(phyad, regad, &regData);
		if(regData & (1<<bit)) 
			* value = 1;
		else
			* value = 0;
	}
}

void rtl8305s_smiWriteBit(uint8 phyad, uint8 regad, uint8 bit, uint8 value) {
	uint16 regData;
	
	if(bit>=16)
		return;
	rtl8305s_smiRead(phyad, regad, &regData);
	if(value) 
		regData = regData | (1<<bit);
	else
		regData = regData & ~(1<<bit);
	rtl8305s_smiWrite(phyad, regad, regData);
}

#define RTL8305S_ASIC_DRIVER


int32 rtl8305s_setAsicEthernetPHY(uint32 port, int8 autoNegotiation, uint32 advCapability, uint32 speed, int8 fullDuplex) {
	uint16 ctrlReg;

	if(port >= RTL8305S_PORT_NUMBER || 
		advCapability < RTL8305S_ETHER_AUTO_100FULL || advCapability > RTL8305S_ETHER_AUTO_10HALF ||
		(speed != 100 && speed != 10)	)
		return FAILED;

	if(advCapability == RTL8305S_ETHER_AUTO_100FULL)
		rtl8305s_smiWrite(port, 4, CAPABLE_PAUSE | CAPABLE_100BASE_TX_FD | CAPABLE_100BASE_TX_HD| CAPABLE_10BASE_TX_FD | CAPABLE_10BASE_TX_HD | 0x1);
	else if(advCapability == RTL8305S_ETHER_AUTO_100HALF)
		rtl8305s_smiWrite(port, 4, CAPABLE_PAUSE | CAPABLE_100BASE_TX_HD| CAPABLE_10BASE_TX_FD | CAPABLE_10BASE_TX_HD | 0x1);
	else	if(advCapability == RTL8305S_ETHER_AUTO_10FULL)
		rtl8305s_smiWrite(port, 4, CAPABLE_PAUSE | CAPABLE_10BASE_TX_FD | CAPABLE_10BASE_TX_HD | 0x1);
	else if(advCapability == RTL8305S_ETHER_AUTO_10HALF)
		rtl8305s_smiWrite(port, 4, CAPABLE_PAUSE | CAPABLE_10BASE_TX_HD | 0x1);
	
	
	if(autoNegotiation == TRUE) 
		ctrlReg = ENABLE_AUTONEGO | RESTART_AUTONEGO; 
	else
		ctrlReg = 0;
	if(speed == 100) 
		ctrlReg |= SPEED_SELECT_100M;
	if(fullDuplex == TRUE)
		ctrlReg |= SELECT_FULL_DUPLEX;
	rtl8305s_smiWrite(port, 0, ctrlReg);

	return SUCCESS;
}


int32 rtl8305s_getAsicEthernetPHY(uint32 port, int8 *autoNegotiation, uint32 *advCapability, uint32 *speed, int8 *fullDuplex) {
	uint16 regData;
	
	if(port >= RTL8305S_PORT_NUMBER || autoNegotiation == NULL || advCapability == NULL || speed == NULL || fullDuplex == NULL)
		return FAILED;
		
	rtl8305s_smiRead(port, 0, &regData);
	*autoNegotiation = (regData & ENABLE_AUTONEGO)? TRUE: FALSE;
	*speed = (regData & SPEED_SELECT_100M)? 100: 10;
	*fullDuplex = (regData & SELECT_FULL_DUPLEX)? TRUE: FALSE;

	rtl8305s_smiRead(port, 4, &regData);
	if(regData & CAPABLE_100BASE_TX_FD)
		*advCapability = RTL8305S_ETHER_AUTO_100FULL;
	else if(regData & CAPABLE_100BASE_TX_HD)
		*advCapability = RTL8305S_ETHER_AUTO_100HALF;
	else if(regData & CAPABLE_10BASE_TX_FD)
		*advCapability = RTL8305S_ETHER_AUTO_10FULL;
	else if(regData & CAPABLE_10BASE_TX_HD)
		*advCapability = RTL8305S_ETHER_AUTO_10HALF;
	
	return SUCCESS;
}


int32 rtl8305s_getAsicPHYLinkStatus(uint32 port, int8 *linkUp) {
	uint8 bitValue;
	
	rtl8305s_smiReadBit(port, 1, 2, &bitValue);
	*linkUp = bitValue == 1? TRUE: FALSE;

	return SUCCESS;
}


int32 rtl8305s_getAsicPHYAutoNegotiationDone(uint32 port, int8 *done) {
	uint8 bitValue;
	
	rtl8305s_smiReadBit(port, 1, 5, &bitValue);
	*done = bitValue == 1? TRUE: FALSE;

	return SUCCESS;
}


int32 rtl8305s_setAsicPHYLoopback(uint32 port, int8 enabled) {
	rtl8305s_smiWriteBit(port, 0, 14, enabled==TRUE?1:0);
	return SUCCESS;
}


int32 rtl8305s_getAsicPHYLoopback(uint32 port, int8 *enabled) {
	uint8 bitValue;
	
	rtl8305s_smiReadBit(port, 0, 14, &bitValue);
	*enabled = bitValue == 1? TRUE: FALSE;

	return SUCCESS;
}


#define RTL8305S_TABLE_DRIVER



int32 rtl8305s_init(void) {
	
	rtl8651_setAsicEthernetMII(8,P5_LINK_MII_MAC, TRUE);
	
	return SUCCESS;
}


