

#include "rtl8651_alg_ftp.h"

#define _RTL8651_L4_MAX_FTP_ALIAS_BUF_SIZE 576
static uint8 *_rtl8651_l4_ftp_alias_buf[2];

#define TRACE( msg ) printk( "in file %s : line %d : function %s : %s\n", __FILE__, __LINE__, __FUNCTION__, msg )


 
struct rcvdelimit {
 	#define RCVDELIMITER_LENGTH 8
 	uint32 length;
	uint8 delimiter[RCVDELIMITER_LENGTH];
	uint16 jump_table[256];
 } PASVpattern;

enum _rtl8651_l4FtpMessageType
{
	_RTL8651_L4_FTP_PORT_COMMAND,
	_RTL8651_L4_FTP_EPRT_COMMAND,
	_RTL8651_L4_FTP_227_REPLY,
	_RTL8651_L4_FTP_229_REPLY,
	_RTL8651_L4_FTP_UNKNOWN_MESSAGE
};

static uint32 _rtl8651_true_addr;	   
static uint16 _rtl8651_true_port;	   



static int32 _rtl8651_l4ParseFtpPortCommand(int8 *sptr, int32 dlen)
{
	int8      ch;
	int32     i, state;
	uint32    addr;
	uint16    port;
	uint8     octet;

	
	
	if (dlen < 18)
		return 0;

	addr = port = octet = 0;
	state = -4;
	for (i = 0; i < dlen; i++)
	{
		ch = sptr[i];
		switch (state)
		{
			case -4:
				if (ch == 'P')
					state++;
				else
					return 0;
				break;
			case -3:
				if (ch == 'O')
					state++;
				else
					return 0;
				break;
			case -2:
				if (ch == 'R')
					state++;
				else
					return 0;
				break;
			case -1:
				if (ch == 'T')
					state++;
				else
					return 0;
				break;

			case 0:
				if (_rtl8651_isspace(ch))
					break;
				else
					state++;
			case 1:
			case 3:
			case 5:
			case 7:
			case 9:
			case 11:
				if (_rtl8651_isdigit(ch))
				{
					octet = ch - '0';
					state++;
				}
				else
					return 0;
				break;
			case 2:
			case 4:
			case 6:
			case 8:
				if (_rtl8651_isdigit(ch))
					octet = 10 * octet + ch - '0';
				else if (ch == ',')
				{
					addr = (addr << 8) + octet;
					state++;
				}
				else
					return 0;
				break;
			case 10:
			case 12:
				if (_rtl8651_isdigit(ch))
					octet = 10 * octet + ch - '0';
				else if (ch == ',' || state == 12)
				{
					port = (port << 8) + octet;
					state++;
				}
				else
					return 0;
				break;
		}
	}

	if (state == 13)
	{
		_rtl8651_true_addr = addr;
		_rtl8651_true_port = port;
		return 1;
	}
	else
		return 0;
}



static int32 _rtl8651_l4ParseFtpEprtCommand(int8 *sptr, int32 dlen)
{
	int8      ch, delim;
	int32     i, state;
	uint32    addr;
	uint16    port;
	uint8     octet;

	

	
	if (dlen < 18)
		return 0;

	addr = port = octet = 0;
	delim = '|';					   
	state = -4;
	for (i = 0; i < dlen; i++)
	{
		ch = sptr[i];
		switch (state)
		{
			case -4:
				if (ch == 'E')
					state++;
				else
					return 0;
				break;
			case -3:
				if (ch == 'P')
					state++;
				else
					return 0;
				break;
			case -2:
				if (ch == 'R')
					state++;
				else
					return 0;
				break;
			case -1:
				if (ch == 'T')
					state++;
				else
					return 0;
				break;

			case 0:
				if (!_rtl8651_isspace(ch))
				{
					delim = ch;
					state++;
				}
				break;
			case 1:
				if (ch == '1')		   
					state++;
				else
					return 0;
				break;
			case 2:
				if (ch == delim)
					state++;
				else
					return 0;
				break;
			case 3:
			case 5:
			case 7:
			case 9:
				if (_rtl8651_isdigit(ch))
				{
					octet = ch - '0';
					state++;
				}
				else
					return 0;
				break;
			case 4:
			case 6:
			case 8:
			case 10:
				if (_rtl8651_isdigit(ch))
					octet = 10 * octet + ch - '0';
				else if (ch == '.' || state == 10)
				{
					addr = (addr << 8) + octet;
					state++;
				}
				else
					return 0;
				break;
			case 11:
				if (_rtl8651_isdigit(ch))
				{
					port = ch - '0';
					state++;
				}
				else
					return 0;
				break;
			case 12:
				if (_rtl8651_isdigit(ch))
					port = 10 * port + ch - '0';
				else if (ch == delim)
					state++;
				else
					return 0;
				break;
		}
	}

	if (state == 13)
	{
		_rtl8651_true_addr  = addr;
		_rtl8651_true_port = port;
		return 1;
	}
	else
		return 0;
}



static int32 _rtl8651_l4ParseFtp227Reply(int8 * sptr, int32 dlen)
{
	int8      ch;
	int32     i, state;
	uint32    addr;
	uint16    port;
	uint8     octet;

	

	
	if (dlen < 17)
		return 0;

	addr = port = octet = 0;

	state = -3;
	for (i = 0; i < dlen; i++)
	{
		ch = sptr[i];
		switch (state)
		{
			case -3:
				if (ch == '2')
					state++;
				else
					return 0;
				break;
			case -2:
				if (ch == '2')
					state++;
				else
					return 0;
				break;
			case -1:
				if (ch == '7')
					state++;
				else
					return 0;
				break;

			case 0:
				if (ch == '(')
					state++;
				break;
			case 1:
			case 3:
			case 5:
			case 7:
			case 9:
			case 11:
				if (_rtl8651_isdigit(ch))
				{
					octet = ch - '0';
					state++;
				}
				else
					return 0;
				break;
			case 2:
			case 4:
			case 6:
			case 8:
				if (_rtl8651_isdigit(ch))
					octet = 10 * octet + ch - '0';
				else if (ch == ',')
				{
					addr = (addr << 8) + octet;
					state++;
				}
				else
					return 0;
				break;
			case 10:
			case 12:
				if (_rtl8651_isdigit(ch))
					octet = 10 * octet + ch - '0';
				else if (ch == ',' || (state == 12 && ch == ')'))
				{
					port = (port << 8) + octet;
					state++;
				}
				else
					return 0;
				break;
		}
	}

	if (state == 13)
	{		
		_rtl8651_true_port = port;
		_rtl8651_true_addr = addr;
		return 1;
	}
	else
		return 0;
}



static int32 _rtl8651_l4ParseFtp229Reply(int8 * sptr, int32 dlen)
{
	int8      ch, delim;
	int32     i, state;
	uint16    port;

	

	
	if (dlen < 11)
		return 0;

	port = 0;
	delim = '|';					   

	state = -3;
	for (i = 0; i < dlen; i++)
	{
		ch = sptr[i];
		switch (state)
		{
			case -3:
				if (ch == '2')
					state++;
				else
					return 0;
				break;
			case -2:
				if (ch == '2')
					state++;
				else
					return 0;
				break;
			case -1:
				if (ch == '9')
					state++;
				else
					return 0;
				break;

			case 0:
				if (ch == '(')
					state++;
				break;
			case 1:
				delim = ch;
				state++;
				break;
			case 2:
			case 3:
				if (ch == delim)
					state++;
				else
					return 0;
				break;
			case 4:
				if (_rtl8651_isdigit(ch))
				{
					port = ch - '0';
					state++;
				}
				else
					return 0;
				break;
			case 5:
				if (_rtl8651_isdigit(ch))
					port = 10 * port + ch - '0';
				else if (ch == delim)
					state++;
				else
					return 0;
				break;
			case 6:
				if (ch == ')')
					state++;
				else
					return 0;
				break;
		}
	}

	if (state == 7)
	{
		_rtl8651_true_port = port;
		return 1;
	}
	else
		return 0;
}



static int32
_rtl8651_l4NewFtpClientMessage( struct ip *pip,
			  rtl8651_tblDrv_naptTcpUdpFlowEntry_t *tb,
			  int32 msgType )
{
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry;
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *ftp_tb;
	struct tcphdr *tc;
	uint32 srcIP, dstIP, extIP;
	uint16 srcPort, dstPort, extPort;

	
	tc = (struct tcphdr *)((int8 *)pip + ((pip->ip_vhl & 0xf) << 2));

	
	srcIP = _rtl8651_true_addr;
	srcPort = _rtl8651_true_port;
	dstIP = PKTGET_UINT32_UNALIGNED( &pip->ip_dst );
	dstPort =  tc->th_dport - 1;		
	extIP = 0;
	extPort = 0;

	
	if (_rtl8651_true_port < IPPORT_RESERVED)
		return FAILED;
	
	
	if (tb->natlink == 1)
	{
		ftp_tb = tb;	
		extIP = tb->insideGlobalIpAddr;
		extPort = _rtl8651_true_port;
	}
	
	else
	{
		
		ftp_tb = &entry;
		if (_rtl8651_getOutboundNaptFlow( TRUE, srcIP, srcPort, dstIP, dstPort, &ftp_tb ) == SUCCESS)
		{
			extIP = ftp_tb->insideGlobalIpAddr;
			extPort = ftp_tb->insideGlobalPort;
		}
		
		else	
		{
			
			_rtl8651_softNaptPolicyRoute( srcIP, dstIP, IPPROTO_TCP, srcPort, dstPort, &extIP, &extPort );
			if (extIP == 0 || extPort == 0)
			{
				
				if( _rtl8651_getUsableExtIpAndPort( TRUE, srcIP, srcPort, dstIP, dstPort, &extIP, &extPort ) != SUCCESS )
				{
					TRACE( "cannot get a pair of external IP and port for NAPT!" );
					return FAILED;
				}
			}

			
			if (_rtl8651_addNaptConnection( TRUE, TRUE | ALG_QOS, TCP_LISTEN_FLOW, 
				srcIP, srcPort, &extIP, &extPort, dstIP, dstPort) != SUCCESS)
			{
				TRACE( "adding NAPT connection failed!" );
				return FAILED;
			}

			
			ftp_tb = &entry;
			if( _rtl8651_getOutboundNaptFlow( TRUE, srcIP, srcPort, dstIP, dstPort, &ftp_tb ) != SUCCESS)
			{
				TRACE( "Error occurs during searching a NAPT flow!" );
				return FAILED;
			}
		}
	}
	
	
	if (ftp_tb != NULL)
	{
		int8 *stemp = (int8 *)_rtl8651_l4_ftp_alias_buf[0];
		int8 *sptr;
		int32 a1, a2, a3, a4, p1, p2;
		int32 slen, hlen, tlen, dlen;
		int32 delta;

		
		hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >> 4) & 0xf)) << 2;
		tlen = ntohs( pip->ip_len );
		dlen = tlen - hlen;
			
		
		a1 = (extIP >> 24) & 0xff;
		a2 = (extIP >> 16) & 0xff;
		a3 = (extIP >> 8) & 0xff;
		a4 = extIP & 0xff;

		
		switch (msgType)
		{
			case _RTL8651_L4_FTP_PORT_COMMAND:
				
				p1 = (extPort >> 8) & 0xff;
				p2 = extPort & 0xff;

				
				sprintf( stemp, "PORT %d,%d,%d,%d,%d,%d\r\n", a1, a2, a3, a4, p1, p2 );
				break;
			case _RTL8651_L4_FTP_EPRT_COMMAND:
				
				sprintf( stemp, "EPRT |1|%d.%d.%d.%d|%d|\r\n", a1, a2, a3, a4, extPort );
				break;
			default:
				TRACE( "Unknown message type!" );
				return FAILED;
		}
		slen = strlen( stemp );	
		
		
		sptr = (int8 *)pip;
		sptr += hlen;
		_strncpy( sptr, stemp, _RTL8651_L4_MAX_FTP_ALIAS_BUF_SIZE - hlen );

		
		tb->tcpAckModified = 1;
		delta = _rtl8651_l4GetDeltaSeqOut( pip, tc, tb );
		_rtl8651_l4AddSeq( pip, tb, delta + slen - dlen );

		
		pip->ip_len = htons( hlen + slen );

		
		tc->th_sum = 0;
		pip->ip_sum = 0;
		tc->th_sum = _rtl8651_tcpChecksum( pip);
		pip->ip_sum = _rtl8651_ipChecksum( pip );
	}
	
	return SUCCESS;
}



static int32
_rtl8651_l4NewFtpServerMessage(struct ip *pip,
			  rtl8651_tblDrv_naptTcpUdpFlowEntry_t *tb,
			  int32 msgType)
{
	uint32 srcIP, dstIP, extIP;
	uint16 srcPort, dstPort, extPort;

	
	srcIP = _rtl8651_true_addr;
	srcPort = _rtl8651_true_port;
	dstIP = PKTGET_UINT32_UNALIGNED( &pip->ip_dst );
	dstPort =  0;
	extIP = tb->insideGlobalIpAddr;
	extPort = srcPort;
	
	
	if (_rtl8651_true_port < IPPORT_RESERVED)
		return FAILED;
		
	
	if (tb->natlink != 1)
	{
		#ifdef _RTL_UPNP
		rtl8651_addUpnpMap( UPNP_TCP|UPNP_ONESHOT|UPNP_ALG_QOS, dstIP, dstPort, extIP, extPort, srcIP, srcPort );
		#endif
	}

	
	int8 *stemp = (int8 *)_rtl8651_l4_ftp_alias_buf[0];
	int8 *sptr;
	int32 a1, a2, a3, a4, p1, p2;
	int32 slen, hlen, tlen, dlen;
	int32 delta;
	struct tcphdr *tc;

	
	tc = (struct tcphdr *)((int8 *)pip + ((pip->ip_vhl & 0xf) << 2));
	hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >> 4) & 0xf)) << 2;	
	tlen = ntohs( pip->ip_len );
	dlen = tlen - hlen;

	
	a1 = (extIP >> 24) & 0xff;
	a2 = (extIP >> 16) & 0xff;
	a3 = (extIP >> 8) & 0xff;
	a4 = extIP & 0xff;

	
	switch (msgType)
	{
		case _RTL8651_L4_FTP_227_REPLY:
			p1 = (extPort >> 8) & 0xff;
			p2 = extPort & 0xff;
			sprintf( stemp, "227 Entering Passive Mode (%d,%d,%d,%d,%d,%d)\r\n", a1, a2, a3, a4, p1, p2 );
			break;
		case _RTL8651_L4_FTP_EPRT_COMMAND:
			sprintf( stemp, "EPRT |1|%d.%d.%d.%d|%d|\r\n", a1, a2, a3, a4, _rtl8651_true_port );
			break;
		case _RTL8651_L4_FTP_229_REPLY:
			sprintf( stemp, "229 Entering Extended Passive Mode (|||%d|)\r\n", _rtl8651_true_port );
			break;
	}
	slen = strlen( stemp );		
		
	
	sptr = (int8 *)pip;
	sptr += hlen;
	_strncpy( sptr, stemp, _RTL8651_L4_MAX_FTP_ALIAS_BUF_SIZE - hlen );

	
	tb->tcpAckModified = 1;
	delta = _rtl8651_l4GetDeltaSeqOut( pip, tc, tb );
	_rtl8651_l4AddSeq( pip, tb, delta + slen - dlen );

	
	pip->ip_len = htons( hlen + slen );

	
	tc->th_sum = 0;
	pip->ip_sum = 0;
	tc->th_sum = _rtl8651_tcpChecksum( pip);
	pip->ip_sum = _rtl8651_ipChecksum( pip );
		
	return SUCCESS;
}



void _rtl8651_configPasvFtpClient( uint8 isTCP, uint32 srcIP, uint16 srcPort, uint32 extIP, uint16 extPort, uint32 dstIP, uint16 dstPort, rtl8651_alg_qos_t *algQosEntry )
{
	memset( algQosEntry, 0, sizeof(rtl8651_alg_qos_t) );
	algQosEntry->algType = RTL8651_ALG_QOS_TYPE_PASV_FTP_CLIENT;
	if (isTCP == 1)
		algQosEntry->isTCP = 1;
	else
		algQosEntry->isTCP = 0;
	algQosEntry->srcIP = srcIP;
	algQosEntry->srcPort = srcPort;
	algQosEntry->extIP = extIP;
	algQosEntry->extPort = extPort;
	algQosEntry->dstIP = dstIP;
	algQosEntry->dstPort = dstPort;
	algQosEntry->isOneShot = 1;		
	algQosEntry->liveTime = RTL8651_ALG_QOS_PASV_FTP_LIVETIME;
	algQosEntry->isHigh = alg_qoshl[0][RTL8651_ALG_FTP_IDX];
	strcpy( algQosEntry->ifName, "eth1" );	
	algQosEntry->groupId = alg_qosqid[0][RTL8651_ALG_FTP_IDX];
	algQosEntry->qosRatio = alg_qosRatio[0][RTL8651_ALG_FTP_IDX];

	return;
}



void _rtl8651_configPasvFtpServer( uint8 isTCP, uint32 srcIP, uint16 srcPort, uint32 extIP, uint16 extPort, uint32 dstIP, uint16 dstPort, rtl8651_alg_qos_t *algQosEntry )
{
	memset( algQosEntry, 0, sizeof(rtl8651_alg_qos_t) );
	algQosEntry->algType = RTL8651_ALG_QOS_TYPE_PASV_FTP_SERVER;
	if (isTCP == 1)
		algQosEntry->isTCP = 1;
	else
		algQosEntry->isTCP = 0;
	algQosEntry->srcIP = srcIP;
	algQosEntry->srcPort = srcPort;
	algQosEntry->extIP = extIP;
	algQosEntry->extPort = extPort;
	algQosEntry->dstIP = dstIP;
	algQosEntry->dstPort = dstPort;
	algQosEntry->isOneShot = 1;		
	algQosEntry->liveTime = RTL8651_ALG_QOS_PASV_FTP_LIVETIME;
	algQosEntry->isHigh = alg_qoshl[0][RTL8651_ALG_FTP_IDX];
	strcpy( algQosEntry->ifName, "eth1" );	
	algQosEntry->groupId = alg_qosqid[0][RTL8651_ALG_FTP_IDX];
	algQosEntry->qosRatio = alg_qosRatio[0][RTL8651_ALG_FTP_IDX];

	return;
}



void _rtl8651_configActiveFtp( uint8 isTCP, uint32 srcIP, uint16 srcPort, uint32 extIP, uint16 extPort, uint32 dstIP, uint16 dstPort, rtl8651_alg_qos_t *algQosEntry )
{
	memset( algQosEntry, 0, sizeof(rtl8651_alg_qos_t) );
	algQosEntry->algType = RTL8651_ALG_QOS_TYPE_ACTIVE_FTP;
	if (isTCP == 1)
		algQosEntry->isTCP = 1;
	else
		algQosEntry->isTCP = 0;
	algQosEntry->srcIP = srcIP;
	algQosEntry->srcPort = srcPort;
	algQosEntry->extIP = extIP;
	algQosEntry->extPort = extPort;
	algQosEntry->dstIP = dstIP;
	algQosEntry->dstPort = dstPort;
	algQosEntry->isOneShot = 1;
	algQosEntry->liveTime = RTL8651_ALG_QOS_ACTIVE_FTP_LIVETIME;
	algQosEntry->isHigh = alg_qoshl[0][RTL8651_ALG_FTP_IDX];
	strcpy( algQosEntry->ifName, "eth1" );
	algQosEntry->groupId = alg_qosqid[0][RTL8651_ALG_FTP_IDX];
	algQosEntry->qosRatio = alg_qosRatio[0][RTL8651_ALG_FTP_IDX];

	return;
}



static int32 init_ftp = 0;


int32 rtl8651_l4AliasFtpInit( void *arg )
{
	
	int8 pasvCmdStr[] = "PASV\r\n";
	int bufSize = _RTL8651_L4_MAX_FTP_ALIAS_BUF_SIZE + 64;
	
	if (init_ftp)
		return SUCCESS;
	
	
	_rtl8651_l4_ftp_alias_buf[0] = (uint8 *)rtlglue_malloc( bufSize );
	_rtl8651_l4_ftp_alias_buf[1] = (uint8 *)rtlglue_malloc( bufSize );
	assert(_rtl8651_l4_ftp_alias_buf[0]);
	assert(_rtl8651_l4_ftp_alias_buf[1]);
	memset( _rtl8651_l4_ftp_alias_buf[0], 0, bufSize );
	memset( _rtl8651_l4_ftp_alias_buf[1], 0, bufSize );

	
	rtl8651_memcpy( (void *)PASVpattern.delimiter, pasvCmdStr, sizeof(pasvCmdStr) + 1 );
	PASVpattern.length = strlen(pasvCmdStr);
	mBuf_getBMjmpTable( (uint8 *)pasvCmdStr, (uint16 *)&PASVpattern.jump_table[0], PASVpattern.length, TRUE );

	
	_rtl8651_true_addr = 0;
	_rtl8651_true_port = 0;

	
	init_ftp = 1;

	return SUCCESS;
}




void rtl8651_l4AliasFtpReInit( void )
{
	int8 pasvCmdStr[] = "PASV\r\n";
	int bufSize = _RTL8651_L4_MAX_FTP_ALIAS_BUF_SIZE + 64;

	
	_rtl8651_true_addr = 0;
	_rtl8651_true_port = 0;
	if (_rtl8651_l4_ftp_alias_buf[0] == NULL)
	{
		_rtl8651_l4_ftp_alias_buf[0] = (uint8 *)rtlglue_malloc( bufSize );
		assert(_rtl8651_l4_ftp_alias_buf[0]);
		memset( _rtl8651_l4_ftp_alias_buf[0], 0, bufSize );
	}
	if (_rtl8651_l4_ftp_alias_buf[1] == NULL)
	{
		_rtl8651_l4_ftp_alias_buf[1] = (uint8 *)rtlglue_malloc( bufSize );
		assert(_rtl8651_l4_ftp_alias_buf[1]);
		memset( _rtl8651_l4_ftp_alias_buf[1], 0, bufSize );
	}

	
	rtl8651_memcpy( (void *)PASVpattern.delimiter, pasvCmdStr, sizeof(pasvCmdStr) + 1 );
	PASVpattern.length = strlen(pasvCmdStr);
	mBuf_getBMjmpTable( (uint8 *)pasvCmdStr, (uint16 *)&PASVpattern.jump_table[0], PASVpattern.length, TRUE );

	return;
}




int32 rtl8651_l4AliasHandleFtpClientOutbound(
	struct rtl_pktHdr *indat, 
	struct ip *inhdr,
	rtl8651_tblDrv_naptTcpUdpFlowEntry_t *link	
)
{
	struct rtl_pktHdr *pkthdrPtr = (struct rtl_pktHdr *)indat;
	struct rtl_mBuf *mbuf = pkthdrPtr->ph_mbuf;
	struct ip *inip = (struct ip *)(inhdr);
	rtl8651_tblDrv_naptTcpUdpFlowEntry_t *tb = (rtl8651_tblDrv_naptTcpUdpFlowEntry_t *)link;
	struct ip *pip;
	struct tcphdr *tc;
	int8 *sptr;
	int8 *newPacket;
	int32 offset;
	int32 msgType;
	int32 hlen, tlen, dlen;

	
	if ((alg_supports[indat->ph_dsid] & RTL8651_ALG_FTP) == 0)
		return SUCCESS;

	
	offset = ((uint8 *)inip) - mbuf->m_data;
	if (offset < 0 || offset > 40 || tb == NULL)
		return FWDENG_USER_DROP;
	
	
	if (mbuf->m_pkthdr->ph_len > _RTL8651_L4_MAX_FTP_ALIAS_BUF_SIZE)
		return FWDENG_USER_DROP;

	
	newPacket =  (int8 *)_rtl8651_l4_ftp_alias_buf[1];
	
	
	mBuf_copyToUserBuffer( mbuf, offset, mbuf->m_pkthdr->ph_len - offset, newPacket );	
	pip = (struct ip *)newPacket;	
	
	
	tc = (struct tcphdr *)((int8 *)pip + ((pip->ip_vhl & 0xf) << 2));	
	hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >> 4) & 0xf)) << 2;	
	tlen = ntohs( pip->ip_len );		
	dlen = tlen - hlen;				

	
	if (dlen <= 0)
		return SUCCESS;

	
	sptr = (int8 *)pip;
	sptr += hlen;

	
	if (tb->tcpFtpLastLineCRLFTermed == 1)
	{
		msgType = _RTL8651_L4_FTP_UNKNOWN_MESSAGE;

		
		if (_rtl8651_l4ParseFtpPortCommand( sptr, dlen ))
			msgType = _RTL8651_L4_FTP_PORT_COMMAND;
		else if (_rtl8651_l4ParseFtpEprtCommand( sptr, dlen ))
			msgType = _RTL8651_L4_FTP_EPRT_COMMAND;
		
		
		if (msgType != _RTL8651_L4_FTP_UNKNOWN_MESSAGE)
		{
			
			#ifdef RTL8651_ALG_QOS
			
			uint32 srcIP = _rtl8651_true_addr;
			uint16 srcPort = _rtl8651_true_port;
			uint32 dstIP = PKTGET_UINT32_UNALIGNED( &pip->ip_dst );
			uint16 dstPort = tc->th_dport - 1;

			
			rtl8651_alg_qos_t algQosEntry;
			_rtl8651_configActiveFtp( 1, srcIP, srcPort, 0, 0, dstIP, dstPort, &algQosEntry );
		
			
			if (_rtl8651_algQos_addEntry( &algQosEntry ) != SUCCESS)
				TRACE( "adding FTP ALG QoS failed!" );
			#endif

			
			if (_rtl8651_l4NewFtpClientMessage( pip, tb, msgType ) != SUCCESS )
				return FWDENG_USER_DROP;

			
			tlen = ntohs( pip->ip_len );		
			mBuf_copyToMbuf( mbuf, offset, tlen, newPacket );
			assert( mbuf->m_len == mbuf->m_pkthdr->ph_len );
		}
	}

	
	sptr = (int8 *)pip;
	if ((sptr[tlen - 2] == '\r') && (sptr[tlen - 1] == '\n'))
		link->tcpFtpLastLineCRLFTermed = 1;
	else
		link->tcpFtpLastLineCRLFTermed = 0;

	return SUCCESS;
}



int32 rtl8651_l4AliasHandleFtpClientInbound(
	struct rtl_pktHdr *indat, 
	struct ip *inhdr,
	rtl8651_tblDrv_naptTcpUdpFlowEntry_t *link
)
{
	struct rtl_pktHdr *pkthdrPtr = (struct rtl_pktHdr *)indat;
	struct rtl_mBuf *mbuf = pkthdrPtr->ph_mbuf;
	ip_t *inip = (ip_t *)(inhdr);
	rtl8651_tblDrv_naptTcpUdpFlowEntry_t *tb = (rtl8651_tblDrv_naptTcpUdpFlowEntry_t *)link;
	struct tcphdr *tc;
	ip_t *pip;
	int8 *newPacket;
	int32 msgType;
	int32 hlen, tlen, dlen;
	int32 offset;
	int8 *sptr;
	rtl8651_alg_qos_t algQosEntry;
	uint32 srcIP, dstIP;
	uint16 srcPort, dstPort;


	
	if ((alg_supports[indat->ph_dsid] & RTL8651_ALG_FTP) == 0)
		return SUCCESS;

	
	offset = ((uint8 *)inip) - mbuf->m_data;
	if (offset < 0 || offset > 40 || tb == NULL)
		return FWDENG_USER_DROP;
	
	
	if (mbuf->m_pkthdr->ph_len > _RTL8651_L4_MAX_FTP_ALIAS_BUF_SIZE)
		return FWDENG_USER_DROP;

	
	newPacket =  (int8 *)_rtl8651_l4_ftp_alias_buf[1];
	
	
	mBuf_copyToUserBuffer( mbuf, offset, mbuf->m_pkthdr->ph_len - offset, newPacket );	
	pip = (struct ip *)newPacket;	
	
	
	tc = (struct tcphdr *)((int8 *)pip + ((pip->ip_vhl & 0xf) << 2));	
	hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >> 4) & 0xf)) << 2;	
	tlen = ntohs( pip->ip_len );		
	dlen = tlen - hlen;				

	
	if (dlen <= 0)
		return SUCCESS;

	
	sptr = (int8 *)pip;
	sptr += hlen;

	
	if (tb->tcpFtpLastLineCRLFTermed == 1)
	{
		#ifdef RTL8651_ALG_QOS
		msgType = _RTL8651_L4_FTP_UNKNOWN_MESSAGE;

		
		if (_rtl8651_l4ParseFtp227Reply( sptr, dlen ))
			msgType = _RTL8651_L4_FTP_227_REPLY;
		else if (_rtl8651_l4ParseFtp229Reply( sptr, dlen ))
		{
			msgType = _RTL8651_L4_FTP_229_REPLY;
			_rtl8651_true_addr = tb->dstIpAddr;
		}
		
		
		if (msgType != _RTL8651_L4_FTP_UNKNOWN_MESSAGE)
		{
			
			dstIP = _rtl8651_true_addr;
			dstPort = _rtl8651_true_port;

			
			srcIP = tb->insideLocalIpAddr;
			srcPort = 0;

			
			_rtl8651_configPasvFtpClient( 1, srcIP, srcPort, 0, 0, dstIP, dstPort, &algQosEntry );
		
			
			if (_rtl8651_algQos_addEntry( &algQosEntry ) != SUCCESS)
				TRACE( "Failed to add FTP ALG QoS flow!" );
		}
		#endif
	}

	
	sptr = (int8 *)pip;
	if ((sptr[tlen - 2] == '\r') && (sptr[tlen - 1] == '\n'))
		link->tcpFtpLastLineCRLFTermed = 1;
	else
		link->tcpFtpLastLineCRLFTermed = 0;
	
	return SUCCESS;
}






int32 rtl8651_l4AliasHandleFtpServerOutbound(
	struct rtl_pktHdr *indat, 
	struct ip *inhdr,
	rtl8651_tblDrv_naptTcpUdpFlowEntry_t *link
)
{
	struct rtl_pktHdr *pkthdrPtr = (struct rtl_pktHdr *)indat;
	struct rtl_mBuf *mbuf = pkthdrPtr->ph_mbuf;
	struct ip *inip = (struct ip *)(inhdr);
	rtl8651_tblDrv_naptTcpUdpFlowEntry_t *tb = (rtl8651_tblDrv_naptTcpUdpFlowEntry_t *)link;
	struct ip *pip;
	struct tcphdr *tc;
	int8 *sptr;
	int8 *newPacket;
	int32 offset;
	int32 msgType;
	int32 hlen, tlen, dlen;
	int bufSize = _RTL8651_L4_MAX_FTP_ALIAS_BUF_SIZE + 64;

	
	if ((alg_supports[indat->ph_dsid] & RTL8651_ALG_FTP) == 0)
		return SUCCESS;

	
	offset = (uint8 *)inip - mbuf->m_data;
	if (offset < 0 || offset > 40 || tb == NULL)
		return FWDENG_USER_DROP;
	
	
	if (mbuf->m_pkthdr->ph_len > _RTL8651_L4_MAX_FTP_ALIAS_BUF_SIZE)
		return FWDENG_USER_DROP;

	
	newPacket =  (int8 *)_rtl8651_l4_ftp_alias_buf[1];
	memset( newPacket, 0, bufSize );
	
	
	mBuf_copyToUserBuffer( mbuf, offset, mbuf->m_pkthdr->ph_len - offset, newPacket );	
	pip = (struct ip *)newPacket;	
	
	
	tc = (struct tcphdr *)((int8 *)pip + ((pip->ip_vhl & 0xf) << 2));	
	hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >> 4) & 0xf)) << 2;	
	tlen = ntohs( pip->ip_len );		
	dlen = tlen - hlen;				

	
	if (dlen <= 0)
		return SUCCESS;

	
	sptr = (int8 *)pip;
	sptr += hlen;

	
	if (tb->tcpFtpLastLineCRLFTermed == 1)
	{
		msgType = _RTL8651_L4_FTP_UNKNOWN_MESSAGE;

		
		if (_rtl8651_l4ParseFtp227Reply(sptr, dlen))
			msgType = _RTL8651_L4_FTP_227_REPLY;
		else if (_rtl8651_l4ParseFtp229Reply(sptr, dlen))
		{
			msgType = _RTL8651_L4_FTP_229_REPLY;
			_rtl8651_true_addr = tb->insideLocalIpAddr;		
		}
		
		
		if (msgType != _RTL8651_L4_FTP_UNKNOWN_MESSAGE)
		{
			
			#ifdef RTL8651_ALG_QOS
			
			uint32 srcIP = _rtl8651_true_addr;
			uint16 srcPort = _rtl8651_true_port;

			
			uint32 dstIP = PKTGET_UINT32_UNALIGNED( &pip->ip_dst );
			uint16 dstPort = 0;

			
			rtl8651_alg_qos_t algQosEntry;
			_rtl8651_configPasvFtpServer( 1, srcIP, srcPort, 0, 0, dstIP, dstPort, &algQosEntry );
		
			
			if (_rtl8651_algQos_addEntry( &algQosEntry ) != SUCCESS)
				TRACE( "adding FTP ALG QoS failed!" );
			#endif

			
			if (_rtl8651_l4NewFtpServerMessage( pip, tb, msgType ) != SUCCESS )
				return FWDENG_USER_DROP;

			
			tlen = ntohs( pip->ip_len );		
			mBuf_copyToMbuf( mbuf, offset, tlen, newPacket );
			assert( mbuf->m_len == mbuf->m_pkthdr->ph_len );
		}
	}
	

	
	sptr = (int8 *)pip;
	if ((sptr[tlen - 2] == '\r') && (sptr[tlen - 1] == '\n'))
		link->tcpFtpLastLineCRLFTermed = 1;
	else
	{
		link->tcpFtpLastLineCRLFTermed = 0;
	}
	
	return SUCCESS;
}




int32 rtl8651_l4AliasHandleFtpServerInbound(
	struct rtl_pktHdr *indat, 
	struct ip *inhdr,
	rtl8651_tblDrv_naptTcpUdpFlowEntry_t *link	
)
{
	struct rtl_pktHdr *pkthdrPtr = (struct rtl_pktHdr *)indat;
	struct rtl_mBuf *mbuf = pkthdrPtr->ph_mbuf;	
	ip_t *inip = (ip_t *)(inhdr);
	rtl8651_tblDrv_naptTcpUdpFlowEntry_t *tb = (rtl8651_tblDrv_naptTcpUdpFlowEntry_t *)link;
	struct tcphdr *tc;
	rtl8651_alg_qos_t algQosEntry;
	int32 hlen, tlen, dlen;
	int8 *sptr;
	int8 *newPacket;	
	ip_t *pip;
	int32 msgType;
	int32 offset;
	uint32 srcIP, extIP, dstIP;
	uint16 srcPort, extPort, dstPort;
	int bufSize = _RTL8651_L4_MAX_FTP_ALIAS_BUF_SIZE + 64;


	
	if ((alg_supports[indat->ph_dsid] & RTL8651_ALG_FTP) == 0)
		return SUCCESS;

	
	offset = ((uint8 *)inip) - mbuf->m_data;
	if (offset < 0 || offset > 40 || tb == NULL)
		return FWDENG_USER_DROP;
	
	
	if (mbuf->m_pkthdr->ph_len > _RTL8651_L4_MAX_FTP_ALIAS_BUF_SIZE)
		return FWDENG_USER_DROP;

	
	newPacket = (int8 *)_rtl8651_l4_ftp_alias_buf[1];
	memset( newPacket, 0, bufSize );
	
	
	mBuf_copyToUserBuffer( mbuf, offset, mbuf->m_pkthdr->ph_len - offset, newPacket );	
	pip = (struct ip *)newPacket;	
	
	
	tc = (struct tcphdr *)((int8 *)pip + ((pip->ip_vhl & 0xf) << 2));	
	hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >> 4) & 0xf)) << 2;	
	tlen = ntohs( pip->ip_len );		
	dlen = tlen - hlen;			

	
	if (dlen <= 0)
		return SUCCESS;

	
	sptr = (int8 *)pip;
	sptr += hlen;

	
	if (tb->tcpFtpLastLineCRLFTermed == 1)
	{
		#ifdef RTL8651_ALG_QOS
		msgType = _RTL8651_L4_FTP_UNKNOWN_MESSAGE;

		
		if (_rtl8651_l4ParseFtpPortCommand( sptr, dlen ))
			msgType = _RTL8651_L4_FTP_PORT_COMMAND;
		else if (_rtl8651_l4ParseFtpEprtCommand( sptr, dlen ))
			msgType = _RTL8651_L4_FTP_EPRT_COMMAND;

		
		if (msgType != _RTL8651_L4_FTP_UNKNOWN_MESSAGE)
		{
   			
			dstIP = _rtl8651_true_addr;
			dstPort = _rtl8651_true_port;

			
			srcIP = tb->insideLocalIpAddr;
			srcPort = tb->insideLocalPort - 1;

			
			_rtl8651_configActiveFtp( 1, srcIP, srcPort, 0, 0, dstIP, dstPort, &algQosEntry );
		
			
			if (_rtl8651_algQos_addEntry( &algQosEntry ) != SUCCESS)
				TRACE( "Failed to add FTP ALG QoS flow!" );

			
			if (tb->natlink != 1)
			{
			  #ifdef _RTL_UPNP
			  extIP = tb->insideGlobalIpAddr;
			  extPort = tb->insideLocalPort - 1;    
			  rtl8651_addUpnpMap( UPNP_TCP|UPNP_ONESHOT|UPNP_ALG_QOS, dstIP, dstPort, extIP, extPort, srcIP, srcPort );
			  #endif
			}
		}
		#endif
	}

	
	if ((sptr[dlen - 2] == '\r') && (sptr[dlen - 1] == '\n'))
		link->tcpFtpLastLineCRLFTermed = 1;
	else
		link->tcpFtpLastLineCRLFTermed = 0;

	return SUCCESS;
}

