


#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"

#include "rtl8651_alg_init.h"
#include "rtl8651_alg_ftp.h"
#include "rtl8651_alg_h323.h"


#include "rtl8651_alg_dx7play.h"
#include "rtl8651_alg_pptp.h"

#include "rtl8651_alg_ipsec.h"
#include "rtl8651_alg_ymsg.h"
#include "rtl8651_alg_icq.h"
#include "rtl8651_alg_irc_dcc.h"
#include "rtl8651_alg_l2tp.h"
#include "rtl8651_alg_sip.h"
#include "rtl8651_alg_vdolive.h"
#include "rtl8651_alg_cs.h"
#include "rtl8651_alg_quake.h"
#include "rtl8651_alg_battlenet.h"
#include "rtl8651_alg_realaudio.h"
#include "rtl8651_alg_rtsp.h"
#include "rtl8651_alg_cuseeme.h"
#include "rtl8651_alg_tftp.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_config.h"

uint32 alg_supports[RTL8651_MAX_DIALSESSION]={0};
uint32 alg_serverips[RTL8651_MAX_DIALSESSION][32]={{0}};
uint32 alg_qosqid[RTL8651_MAX_DIALSESSION][32]={{0}};
uint32 alg_qoshl[RTL8651_MAX_DIALSESSION][32]={{0}};
uint32 alg_qosRatio[RTL8651_MAX_DIALSESSION][32]={{0}};

struct rtl8651_algMaxSession algMaxSession={32,32,32};



#ifdef RTL8651_ALG_MODULE_UPDATE


#define RTL8651_ALG_MODULE_NUM RTL8651_ALG_MAX_IDX
#define RTL8651_ALG_PORT_NUM 28


#define ALG_TYPE_CLIENT	 		1	
#define ALG_TYPE_SERVER 		2	
#define ALG_TYPE_SERVERPORT	4	

typedef struct {
	
	uint16 startPort;
	uint16 endPort;
	uint8 isTcp;

	
	uint8 algType;

	
	
	uint16 algIdx;

	
	int32 (*alg_init)(void *);
	alg_funcptr alg_outboundClientcb;
	alg_funcptr alg_inboundClientcb;
	alg_funcptr alg_outboundServercb;
	alg_funcptr alg_inboundServercb;
} _rtl8651_algPortConfig_t;



static _rtl8651_algPortConfig_t _rtl8651_algPortConfig[RTL8651_ALG_PORT_NUM] = {
	
	
	{
		1723, 1723, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER | ALG_TYPE_SERVERPORT, 
		RTL8651_ALG_PPTP_IDX, 
		rtl8651_l4AliasPptpInit, 
		rtl8651_l4AliasHandlePptpClientOutbound, 
		rtl8651_l4AliasHandlePptpClientInbound, 
		rtl8651_l4AliasHandlePptpServerOutbound, 
		rtl8651_l4AliasHandlePptpServerInbound
	},


	
	
	{
		500, 500, FALSE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER | ALG_TYPE_SERVERPORT, 
		RTL8651_ALG_IPSEC_IDX, 
		rtl8651_l4AliasIpsecInit,
		rtl8651_l4AliasHandleIpsecClientOutbound,
		rtl8651_l4AliasHandleIpsecClientInbound,
		NULL,
		NULL
	},


	
	
	{
		1701, 1701, FALSE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER, 
		RTL8651_ALG_L2TP_IDX, 
		rtl8651_l4AliasL2tpInit, 
		rtl8651_l4AliasHandleL2tpClientOutbound, 
		rtl8651_l4AliasHandleL2tpClientInbound, 
		NULL, 
		NULL
	},


	
	
	{
		21, 21, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER | ALG_TYPE_SERVERPORT, 
		RTL8651_ALG_FTP_IDX, 
		rtl8651_l4AliasFtpInit,
		rtl8651_l4AliasHandleFtpClientOutbound,
		rtl8651_l4AliasHandleFtpClientInbound,
		rtl8651_l4AliasHandleFtpServerOutbound,
		rtl8651_l4AliasHandleFtpServerInbound
	},


	
	
	{
		1503, 1503, TRUE, 
		ALG_TYPE_SERVERPORT, 
		RTL8651_ALG_NETMEETING_IDX, 
		NULL, 
		NULL, 
		NULL, 
		NULL, 
		NULL
	}, 

	
	{
		1719, 1719, FALSE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER, 
		RTL8651_ALG_NETMEETING_IDX, 
		rtl8651_l4AliasH323Init,
      		rtl8651_l4AliasHandleClientH225RASOut,
	      	rtl8651_l4AliasHandleClientH225RASIn,
	      	rtl8651_l4AliasHandleServerH225RASOut,
	      	rtl8651_l4AliasHandleServerH225RASIn
	}, 

	
	{
		1720, 1720, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER | ALG_TYPE_SERVERPORT, 
		RTL8651_ALG_NETMEETING_IDX, 
		rtl8651_l4AliasH323Init,
      		rtl8651_l4AliasHandleClientH225CSOut,
	      	rtl8651_l4AliasHandleClientH225CSIn,
	      	rtl8651_l4AliasHandleServerH225CSOut,
	      	rtl8651_l4AliasHandleServerH225CSIn
	},

	
	{
		17200, 17200, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER, 
		RTL8651_ALG_NETMEETING_IDX, 
		NULL,
      		NULL,
      		NULL,
      		rtl8651_l4AliasHandleServerH245Out,
      		rtl8651_l4AliasHandleServerH245In
	}, 


	
	
	{
		5190, 5190, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER, 
		RTL8651_ALG_ICQ_IDX, 
		NULL, 
		rtl8651_l4AliasHandleIcqClientOutbound,
		NULL,
		NULL,
		NULL
	}, 

	
	{
		5190, 5190, FALSE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER, 
		RTL8651_ALG_ICQ_IDX, 
		NULL, 
		rtl8651_l4AliasHandleIcqUdpClientOutbound,
		NULL,
		NULL,
		NULL
	}, 

	
	{
		24500, 24500 + _RTL8651_ALG_ICQ_MAX_CLIENT - 1, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER,
		RTL8651_ALG_ICQ_IDX, 
		NULL,
		NULL,
		NULL,
		rtl8651_l4AliasHandleIcqServerOutbound,
		NULL
	}, 


	
	
	{
		5050, 5050, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER,
		RTL8651_ALG_YAHOOMSG_IDX, 
		rtl8651_l4AliasYmsgInit,
		rtl8651_l4AliasHandleYmsgClientOutbound,
		NULL,
		NULL,
		NULL
	}, 


	
	
	{
		6667, 6668, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER,
		RTL8651_ALG_IRC_IDX, 
		NULL,
		rtl8651_l4AliasHandleIrcDccClientOutbound,
		NULL,
		NULL,
		NULL
	}, 


	
	
	{
		47624, 47624, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER | ALG_TYPE_SERVERPORT, 
		RTL8651_ALG_DX7_IDX, 
		NULL,
		rtl8651_l4AliasHandleDx7playClientOutbound, 
		NULL,
		NULL,
		rtl8651_l4AliasHandleDx7playServerInbound
	}, 

	
	{
		2300, 2300, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER, 
		RTL8651_ALG_DX7_IDX, 
		NULL, 
		rtl8651_l4AliasHandleDx7playClientOutbound2, 
		NULL, 
		NULL,
		rtl8651_l4AliasHandleDx7playServerInbound2
	}, 


	
	
	{
		5060, 5060, FALSE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER | ALG_TYPE_SERVERPORT,  
		RTL8651_ALG_SIP_IDX, 
		rtl8651_l4AliasSipInit, 
		rtl8651_l4AliasHandleSipClientOutbound,
		rtl8651_l4AliasHandleSipClientInbound,
		rtl8651_l4AliasHandleSipClientOutbound,
		rtl8651_l4AliasHandleSipClientInbound
	}, 


	
	
	{
		2000, 2001, TRUE, 
		ALG_TYPE_SERVERPORT, 
		RTL8651_ALG_ICUII_IDX, 
		NULL, 
		NULL, 
		NULL, 
		NULL, 
		NULL
	}, 

	
	{
		2002, 2002, FALSE, 
		ALG_TYPE_SERVERPORT, 
		RTL8651_ALG_ICUII_IDX, 
		NULL, 
		NULL, 
		NULL, 
		NULL, 
		NULL
	}, 


	
	
	{
		7000, 7000, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER,
		RTL8651_ALG_VDOLIVE_IDX, 
		NULL,
		rtl8651_l4AliasHandleVDOliveClientOutbound,
		NULL,
		NULL,
		NULL
	}, 


	
	
	{
		27910, 27910, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER,
		RTL8651_ALG_QUAKE_IDX, 
		NULL,
		rtl8651_l4AliasHandleQuakeClientOutbound,
		rtl8651_l4AliasHandleQuakeClientInbound,
		NULL,
		NULL
	}, 

	
	{
		27960, 27960, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER,
		RTL8651_ALG_QUAKE_IDX, 
		NULL,
		rtl8651_l4AliasHandleQuakeClientOutbound,
		rtl8651_l4AliasHandleQuakeClientInbound,
		NULL,
		NULL
	}, 


	
	
	{
		27015, 27015, FALSE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER, 
		RTL8651_ALG_CS_IDX, 
		NULL,
		rtl8651_l4AliasHandleCSClientOutbound,
		NULL,
		NULL,
		NULL
	}, 


	
	
	{
		6112, 6112, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER | ALG_TYPE_SERVERPORT, 
		RTL8651_ALG_BATTLENET_IDX, 
		NULL,
		NULL,
		rtl8651_l4AliasHandleBattlenetClientInbound,
		NULL,
		NULL
	}, 

	
	{
		6112, 6112, FALSE, 
		ALG_TYPE_SERVERPORT, 
		RTL8651_ALG_BATTLENET_IDX, 
		NULL,
		NULL,
		NULL,
		NULL,
		NULL
	}, 


	
	
	{
		7070, 7070, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER, 
		RTL8651_ALG_REALAUDIO_IDX, 
		NULL,
		rtl8651_l4AliasHandleRealAudioClientOutbound,
		NULL,
		NULL,
		NULL
	}, 

	
	{
		554, 554, TRUE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER, 
		RTL8651_ALG_REALAUDIO_IDX, 
		rtl8651_l4AliasRTSPInit,
		rtl8651_l4AliasHandleRTSPClientOutbound,
		rtl8651_l4AliasHandleRTSPClientInbound,
		NULL,
		NULL
	}, 


	
	
	{
		7648, 7648, FALSE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER, 
		RTL8651_ALG_CUSEEME_IDX, 
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleCUseeMeClientOutbound,
		(alg_funcptr)rtl8651_l4AliasHandleCUseeMeClientInbound,
		NULL,
		NULL
	}, 


	
	
	{
		69, 69, FALSE, 
		ALG_TYPE_CLIENT | ALG_TYPE_SERVER | ALG_TYPE_SERVERPORT, 
		RTL8651_ALG_TFTP_IDX, 
		NULL,
		rtl8651_l4AliasHandleTftpClientOutbound,
		NULL,
		NULL,
		NULL
	}
};



static uint32 _algMask[RTL8651_ALG_MODULE_NUM] = {
	RTL8651_ALG_PPTP,
	RTL8651_ALG_IPSEC,
	RTL8651_ALG_L2TP,
	RTL8651_ALG_FTP,
	RTL8651_ALG_NETMEETING,
	RTL8651_ALG_ICQ,
	RTL8651_ALG_YAHOOMSG,
	RTL8651_ALG_IRC,
	RTL8651_ALG_DX7,
	RTL8651_ALG_SIP,
	RTL8651_ALG_ICUII,
	RTL8651_ALG_VDOLIVE,
	RTL8651_ALG_QUAKE,
	RTL8651_ALG_CS,
	RTL8651_ALG_BATTLENET,
	RTL8651_ALG_REALAUDIO,
	RTL8651_ALG_CUSEEME,
	RTL8651_ALG_TFTP
};
#endif



uint32 rtl8651_getAlgStatus(void)
{
	return alg_supports[0];
}

uint32 rtl8651a_getAlgStatus(uint32 SessionID)
{
	uint32 dsid=_rtl8651_SessionIDToDSID(SessionID);
	if(dsid>=RTL8651_MAX_DIALSESSION)
		return 0;
	else
		return alg_supports[dsid];
}


uint32 rtl8651_flushAlgServerIn(void)
{
	rtl8651a_flushAlgServerIn(0);
	return SUCCESS;
}
	

#ifdef RTL8651_ALG_MODULE_UPDATE

uint32 rtl8651a_flushAlgServerIn( uint32 SessionID )
{
	uint32 dsid = _rtl8651_SessionIDToDSID( SessionID );
	uint32 gip = _rtl8651_getExtIpAddrByDsid( dsid );

	_rtl8651_algPortConfig_t *pEntry;
	uint16 startPort, endPort, portNum;
	uint8 isTcp;
	uint8 index;
	for (pEntry = &_rtl8651_algPortConfig[0]; 
		pEntry < &_rtl8651_algPortConfig[RTL8651_ALG_PORT_NUM]; 
		pEntry++)
	{
				
		if (pEntry->algType & ALG_TYPE_SERVERPORT)	
		{
			startPort = pEntry->startPort;			
			endPort = pEntry->endPort;
			portNum = endPort - startPort + 1;
			isTcp = pEntry->isTcp;
			index = pEntry->algIdx;

			if (portNum == 1)	
			{
				rtl8651a_delNaptServerPortMapping( dsid, isTcp, gip, startPort, alg_serverips[dsid][index], startPort );
			}
			else		
			{
				rtl8651a_delNaptServerPortRange( dsid, isTcp, gip, startPort, alg_serverips[dsid][index], startPort, portNum );
			}
		}
	}

	return SUCCESS;
}

#else
uint32 rtl8651a_flushAlgServerIn(uint32 SessionID)
{
	uint32 dsid=_rtl8651_SessionIDToDSID(SessionID);
	ipaddr_t gip=_rtl8651_getExtIpAddrByDsid(dsid);
	

	
	rtl8651a_delNaptServerPortMapping(dsid,1,gip, 21,alg_serverips[dsid][RTL8651_ALG_FTP_IDX], 21);

	
	rtl8651a_delNaptServerPortMapping(dsid,1,gip, 1720,alg_serverips[dsid][RTL8651_ALG_NETMEETING_IDX], 1720);
	rtl8651a_delNaptServerPortMapping(dsid,1,gip, 1503,alg_serverips[dsid][RTL8651_ALG_NETMEETING_IDX], 1503);	

	
	rtl8651a_delNaptServerPortMapping(dsid,0,gip, 5060,alg_serverips[dsid][RTL8651_ALG_SIP_IDX], 5060);

	
	rtl8651a_delNaptServerPortMapping(dsid,0,gip, 500,alg_serverips[dsid][RTL8651_ALG_IPSEC_IDX], 500);

	
	rtl8651a_delNaptServerPortMapping(dsid,1,gip, 47624,alg_serverips[dsid][RTL8651_ALG_DX7_IDX], 47624);

	
	rtl8651a_delNaptServerPortRange(dsid,1,gip, 2000,alg_serverips[dsid][RTL8651_ALG_ICUII_IDX], 2000,2);
	rtl8651a_delNaptServerPortMapping(dsid,0,gip, 2002,alg_serverips[dsid][RTL8651_ALG_ICUII_IDX], 2002);

	
	rtl8651a_delNaptServerPortMapping(dsid,1,gip, 1723,alg_serverips[dsid][RTL8651_ALG_PPTP_IDX], 1723);

	
	rtl8651a_delNaptServerPortMapping(dsid,TRUE,gip, 6112,alg_serverips[dsid][RTL8651_ALG_BATTLENET_IDX], 6112);
	rtl8651a_delNaptServerPortMapping(dsid,FALSE,gip, 6112,alg_serverips[dsid][RTL8651_ALG_BATTLENET_IDX], 6112);	

	
	rtl8651a_delNaptServerPortMapping(dsid, FALSE, gip, TFTP_UDP_DPORT, alg_serverips[dsid][RTL8651_ALG_TFTP_IDX], TFTP_UDP_DPORT);

	return SUCCESS;	
}
#endif	

void rtl8651_setAlgStatus(uint32 alg_support,ipaddr_t *alg_serverip)
{
	 rtl8651a_setAlgStatus(0,alg_support,alg_serverip);
}

void rtl8651_setAlgConfig(uint32 alg_support,ipaddr_t *alg_serverip,struct rtl8651_algMaxSession maxSession)
{
	 rtl8651a_setAlgConfig(0,alg_support,alg_serverip,maxSession);
}


int32 rtl8651a_setAlgQosQueueId(uint32 SessionID,uint32 alg_idx,uint32 queueID,uint32 isHigh, uint32 ratio)
{
	if(alg_idx>=32) return FAILED;
	alg_qosqid[SessionID][alg_idx]=queueID;
	alg_qoshl[SessionID][alg_idx]=isHigh;	
	alg_qosRatio[SessionID][alg_idx] = ratio;
	return SUCCESS;
}

int32 rtl8651_setAlgQosQueueId(uint32 alg_idx,uint32 queueID,uint32 isHigh, uint32 ratio)
{
	return rtl8651a_setAlgQosQueueId(0,alg_idx,queueID,isHigh, ratio);
}

int32 _rtl8651_addAlgQosNaptConnection(int8 fromDrv, int16 assigned, int8 flowType, ipaddr_t insideLocalIpAddr, uint16 insideLocalPort, 
			ipaddr_t *insideGlobalIpAddr, uint16 *insideGlobalPort, ipaddr_t dstIpAddr, uint16 dstPort,uint32 AlgIdx) 
{
	

if((assigned&ALG_QOS)&&(assigned&EXTIPPORT_ASSIGNED)&&(alg_qosqid[0][AlgIdx]!=0))
{
	rtl8651_tblDrvAclRule_t rule;

	memset(&rule, 0, sizeof(rtl8651_tblDrvAclRule_t));
	rule.actionType_ = RTL8651_ACL_DROP;

	switch(flowType) {

	case UDP_FLOW: 
		rule.ruleType_	= RTL8651_ACL_UDP;
		rule.udpSrcPortUB_	= insideLocalPort;
		rule.udpSrcPortLB_		= insideLocalPort;


#if 0		
		if(dstPort)
		{
			rule.udpDstPortUB_	= dstPort;
			rule.udpDstPortLB_		= dstPort;
		}
		else			
#endif					
		{
			rule.udpDstPortUB_	= 65535;
			rule.udpDstPortLB_		= 0;
		}
		break;

	default: 
		rule.ruleType_	= RTL8651_ACL_TCP;
		rule.tcpSrcPortUB_		= insideLocalPort;
		rule.tcpSrcPortLB_		= insideLocalPort;

#if 0
		if(dstPort)
		{
			rule.tcpDstPortUB_	= dstPort;
			rule.tcpDstPortLB_	= dstPort;
		}
		else	
#endif					
		{		
			rule.tcpDstPortUB_	= 65535;
			rule.tcpDstPortLB_	= 0;
		}
		break;		
	}
	
	rule.srcIpAddr_		= insideLocalIpAddr;
	rule.srcIpAddrMask_	= 0xffffffff;

#if 0
	if(dstIpAddr)
	{
		rule.dstIpAddr_		= dstIpAddr;
		rule.dstIpAddrMask_	= 0xffffffff;	
	}
#endif	

	{
		int r=0;
		if(alg_qoshl[0][AlgIdx])  
			r=rtl8651_addFlowToUnlimitedQueue("eth1", &rule, alg_qosRatio[0][AlgIdx], alg_qosqid[0][AlgIdx]);
		else
			r=rtl8651_addFlowToLimitedQueue("eth1", &rule, alg_qosRatio[0][AlgIdx], alg_qosqid[0][AlgIdx]);
		if(r) printk("addFlowToUn/LimiteQueue Failed , lport=%d r=%d ishigh=%d qid=%d\n",insideLocalPort,r,alg_qoshl[0][AlgIdx],alg_qosqid[0][AlgIdx]);
	}

}

return _rtl8651_addNaptConnection(fromDrv, assigned, flowType,  insideLocalIpAddr,insideLocalPort, 
			insideGlobalIpAddr, insideGlobalPort, dstIpAddr,  dstPort) ;


}


uint32 rtl8651a_setAlgStatus(uint32 SessionID,uint32 alg_support,ipaddr_t *alg_serverip)
{
	struct rtl8651_algMaxSession maxSession;	
	maxSession.ipsecMaxSession=32;
	maxSession.l2tpMaxSession=32;
	maxSession.pptpMaxSession=32;	
	rtl8651a_setAlgConfig(SessionID,alg_support,alg_serverip,maxSession);
	return SUCCESS;
}



#ifdef RTL8651_ALG_MODULE_UPDATE


uint32 rtl8651a_setAlgConfig( uint32 sessionID, uint32 alg_support, ipaddr_t *alg_serverip, struct rtl8651_algMaxSession maxSession )
{
	uint32 dsid = _rtl8651_SessionIDToDSID( sessionID );
	uint32 gip = _rtl8651_getExtIpAddrByDsid( dsid );		

	
	rtl8651_memcpy( &algMaxSession, &maxSession, sizeof(maxSession) );

	
	rtl8651_memcpy( alg_serverips[dsid], alg_serverip, 32 * sizeof(uint32) );

	
	alg_supports[dsid] = alg_support;
	uint32 totalAlgSupport = alg_supports[0] | alg_supports[1];

	
	rtl8651a_flushAlgServerIn( dsid );
	_rtl8651_flushALGRule();

	
	_rtl8651_algPortConfig_t *pEntry;
	uint8 algType;
	uint8 isTcp;
	uint16 startPort, endPort;
	uint16 portNum;
	uint8 isServer, isClient;
	uint8 isServerPort;
	uint16 algIdx;
	uint32 localIp;
	for (pEntry = &_rtl8651_algPortConfig[0]; 
		pEntry < &_rtl8651_algPortConfig[RTL8651_ALG_PORT_NUM]; 
		pEntry++)
	{
		algIdx = pEntry->algIdx;
		if ((totalAlgSupport & _algMask[algIdx]) == 0)
			continue;
	
		
		algType = pEntry->algType;
		isServer = (algType & ALG_TYPE_SERVER) ? TRUE : FALSE;
		isClient = (algType & ALG_TYPE_CLIENT) ? TRUE : FALSE;
		isServerPort  = (algType & ALG_TYPE_SERVERPORT) ? TRUE : FALSE;

		
		startPort = pEntry->startPort;
		endPort = pEntry->endPort;
		isTcp = pEntry->isTcp;

		
		if (isServer | isClient)
		{
			rtl8651_fwdEngineRegisterAlgModule( startPort, endPort, isTcp, 
				isServer, isClient, 
				pEntry->alg_init, 
				pEntry->alg_outboundClientcb, 
				pEntry->alg_inboundClientcb, 
				pEntry->alg_outboundServercb, 
				pEntry->alg_inboundServercb);

			
			isTcp |= PURE_SW;
		}

		
		localIp = alg_serverip[algIdx];
		if (isServerPort && ((alg_support & _algMask[algIdx]) != 0) && (localIp != 0))
		{
			portNum = endPort - startPort + 1;
			if (portNum == 1)	
			{
				rtl8651a_addNaptServerPortMapping( dsid, isTcp, gip, startPort, localIp, startPort );
			}
			else		
			{
				rtl8651a_addNaptServerPortRange( dsid, isTcp, gip, startPort, localIp, startPort, portNum ); 
			}
		}
	}

	return SUCCESS;
}

#else
uint32 rtl8651a_setAlgConfig(uint32 SessionID,uint32 alg_support,ipaddr_t *alg_serverip,struct rtl8651_algMaxSession maxSession)
{
	int i;
	uint32 dsid=_rtl8651_SessionIDToDSID(SessionID);
	ipaddr_t gip=_rtl8651_getExtIpAddrByDsid(dsid);
	
	if(gip==0) return NOT_OK;

	rtl8651_memcpy(&algMaxSession,&maxSession,sizeof(maxSession));
	
	alg_supports[dsid]=alg_support;

	rtl8651a_flushAlgServerIn(dsid);


	
	#define FTPALG_INSIDE_SERVER_PASSIVE_MODE

	if((alg_support&RTL8651_ALG_FTP)&&(alg_serverip[RTL8651_ALG_FTP_IDX]))
	{		
		rtl8651a_addNaptServerPortMapping(dsid,(1|2),gip, 21,alg_serverip[RTL8651_ALG_FTP_IDX], 21);	
	}

	rtl8651_fwdEngineRegisterAlgModule(21, TRUE,
		rtl8651_l4AliasFtpInit,
		(alg_funcptr)rtl8651_l4AliasHandleFtpClientOutbound,
		(alg_funcptr)rtl8651_l4AliasHandleFtpClientInbound,
		(alg_funcptr)rtl8651_l4AliasHandleFtpServerOutbound,
		(alg_funcptr)rtl8651_l4AliasHandleFtpServerInbound);

	
	
	if((alg_support&RTL8651_ALG_NETMEETING)&&(alg_serverip[RTL8651_ALG_NETMEETING_IDX]))
	{		
		
		rtl8651a_addNaptServerPortMapping(dsid,(1|2),gip, 1720,alg_serverip[RTL8651_ALG_NETMEETING_IDX], 1720);
		rtl8651a_addNaptServerPortMapping(dsid,(1),gip, 1503,alg_serverip[RTL8651_ALG_NETMEETING_IDX], 1503);
	}
	
	rtl8651_fwdEngineRegisterAlgModule(1719, FALSE,
		rtl8651_l4AliasH323Init,
		(alg_funcptr)rtl8651_l4AliasHandleClientH225RASOut,
		(alg_funcptr)rtl8651_l4AliasHandleClientH225RASIn,
		(alg_funcptr)rtl8651_l4AliasHandleServerH225RASOut,
		(alg_funcptr)rtl8651_l4AliasHandleServerH225RASIn);

	rtl8651_fwdEngineRegisterAlgModule(1720, TRUE,
		rtl8651_l4AliasH323Init,
		(alg_funcptr)rtl8651_l4AliasHandleClientH225CSOut,
		(alg_funcptr)rtl8651_l4AliasHandleClientH225CSIn,
		(alg_funcptr)rtl8651_l4AliasHandleServerH225CSOut,
		(alg_funcptr)rtl8651_l4AliasHandleServerH225CSIn);
	

       rtl8651_fwdEngineRegisterAlgModule(17200, TRUE,
		NULL,
		NULL, 
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleServerH245Out,
		(alg_funcptr)rtl8651_l4AliasHandleServerH245In);

		
	rtl8651_fwdEngineRegisterAlgModule(1723, TRUE,
		rtl8651_l4AliasPptpInit,
		(alg_funcptr)rtl8651_l4AliasHandlePptpClientOutbound,
		(alg_funcptr)rtl8651_l4AliasHandlePptpClientInbound,
		(alg_funcptr)rtl8651_l4AliasHandlePptpServerOutbound,
		(alg_funcptr)rtl8651_l4AliasHandlePptpServerInbound);

		
	if((alg_support&RTL8651_ALG_PPTP_SERVER)&&(alg_serverip[RTL8651_ALG_PPTP_IDX]))
	{		
		
		rtl8651a_addNaptServerPortMapping(dsid,(1|2),gip, 1723,alg_serverip[RTL8651_ALG_PPTP_IDX], 1723);
	}

	
	rtl8651_fwdEngineRegisterAlgModule(500, FALSE,
		rtl8651_l4AliasIpsecInit,
		(alg_funcptr)rtl8651_l4AliasHandleIpsecClientOutbound,
		(alg_funcptr)rtl8651_l4AliasHandleIpsecClientInbound,
		NULL,
		NULL);

		
	if((alg_support&RTL8651_ALG_IPSEC_SERVER)&&(alg_serverip[RTL8651_ALG_IPSEC_IDX]))
	{		
		rtl8651a_addNaptServerPortMapping(dsid,(2),gip, 500,alg_serverip[RTL8651_ALG_IPSEC_IDX], 500);
	}

	
	
	rtl8651_fwdEngineRegisterAlgModule(1701, FALSE,
		rtl8651_l4AliasL2tpInit,
		(alg_funcptr)rtl8651_l4AliasHandleL2tpClientOutbound,
		(alg_funcptr)rtl8651_l4AliasHandleL2tpClientInbound,
		NULL,
		NULL);

	
	if((alg_support&RTL8651_ALG_SIP)&&(alg_serverip[RTL8651_ALG_SIP_IDX]))
	{		
		rtl8651a_addNaptServerPortMapping(dsid,(0|2),gip, 5060,alg_serverip[RTL8651_ALG_SIP_IDX], 5060);
	}

	rtl8651_fwdEngineRegisterAlgModule(5060, FALSE,
		rtl8651_l4AliasSipInit,
		(alg_funcptr)rtl8651_l4AliasHandleSipClientOutbound,
		(alg_funcptr)rtl8651_l4AliasHandleSipClientInbound,
		(alg_funcptr)rtl8651_l4AliasHandleSipClientOutbound,
		(alg_funcptr)rtl8651_l4AliasHandleSipClientInbound);	

	


	
	if((alg_support&RTL8651_ALG_DX7)&&(alg_serverip[RTL8651_ALG_DX7_IDX]))
	{		
		rtl8651a_addNaptServerPortMapping(dsid,(1|2),gip, 47624,alg_serverip[RTL8651_ALG_DX7_IDX], 47624);
	}

	
	rtl8651_fwdEngineRegisterAlgModule(47624, TRUE,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleDx7playClientOutbound, 
		NULL,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleDx7playServerInbound); 


		rtl8651_fwdEngineRegisterAlgModule(2300, TRUE,
			NULL,
			(alg_funcptr)rtl8651_l4AliasHandleDx7playClientOutbound2, 
			NULL, 
			NULL,
			(alg_funcptr)rtl8651_l4AliasHandleDx7playServerInbound2); 


		
	if((alg_support&RTL8651_ALG_ICUII)&&(alg_serverip[RTL8651_ALG_ICUII_IDX]))
	{		
		rtl8651a_addNaptServerPortRange(dsid,1,gip, 2000,alg_serverip[RTL8651_ALG_ICUII_IDX], 2000,2);
		rtl8651a_addNaptServerPortMapping(dsid,(0|2),gip, 2002,alg_serverip[RTL8651_ALG_ICUII_IDX], 2002);
	}	


	rtl8651_fwdEngineRegisterAlgModule(5190, TRUE,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleIcqClientOutbound,
		NULL,
		NULL,
		NULL);

	rtl8651_fwdEngineRegisterAlgModule(5190, FALSE,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleIcqUdpClientOutbound,
		NULL,
		NULL,
		NULL);	


	for(i=0;i<_RTL8651_ALG_ICQ_MAX_CLIENT;i++)		
		rtl8651_fwdEngineRegisterAlgModule(24500+i, TRUE,
			NULL,
			NULL,
			NULL,
			(alg_funcptr)rtl8651_l4AliasHandleIcqServerOutbound,
			NULL);

				
	rtl8651_fwdEngineRegisterAlgModule(5050, TRUE,
		rtl8651_l4AliasYmsgInit,
		(alg_funcptr)rtl8651_l4AliasHandleYmsgClientOutbound,
		NULL,
		NULL,
		NULL);

		
	rtl8651_fwdEngineRegisterAlgModule(6667, TRUE,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleIrcDccClientOutbound,
		NULL,
		NULL,
		NULL);

	rtl8651_fwdEngineRegisterAlgModule(6668, TRUE,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleIrcDccClientOutbound,
		NULL,
		NULL,
		NULL);

	
	rtl8651_fwdEngineRegisterAlgModule(7000, TRUE,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleVDOliveClientOutbound,
		NULL,
		NULL,
		NULL);

	
	rtl8651_fwdEngineRegisterAlgModule(27910, TRUE,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleQuakeClientOutbound,
		(alg_funcptr)rtl8651_l4AliasHandleQuakeClientInbound,
		NULL,
		NULL);	

	
	rtl8651_fwdEngineRegisterAlgModule(27960, TRUE,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleQuakeClientOutbound,
		(alg_funcptr)rtl8651_l4AliasHandleQuakeClientInbound,
		NULL,
		NULL);

	
	rtl8651_fwdEngineRegisterAlgModule(6112, TRUE,
		NULL,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleBattlenetClientInbound,
		NULL,
		NULL);

	
	if((alg_support&RTL8651_ALG_BATTLENET)&&(alg_serverip[RTL8651_ALG_BATTLENET_IDX]))
	{		
		rtl8651a_addNaptServerPortMapping(dsid,(FALSE|PURE_SW),gip, 6112,alg_serverip[RTL8651_ALG_BATTLENET_IDX], 6112);
		rtl8651a_addNaptServerPortMapping(dsid,(TRUE|PURE_SW),gip, 6112,alg_serverip[RTL8651_ALG_BATTLENET_IDX], 6112);		
	}	
	
	
	rtl8651_fwdEngineRegisterAlgModule(7070, TRUE,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleRealAudioClientOutbound,
		NULL,
		NULL,
		NULL);

	
	rtl8651_fwdEngineRegisterAlgModule(554, TRUE,
		rtl8651_l4AliasRTSPInit,
		(alg_funcptr)rtl8651_l4AliasHandleRTSPClientOutbound,
		(alg_funcptr)rtl8651_l4AliasHandleRTSPClientInbound,
		NULL,
		NULL);

	
	rtl8651_fwdEngineRegisterAlgModule(7648, FALSE,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleCUseeMeClientOutbound,
		(alg_funcptr)rtl8651_l4AliasHandleCUseeMeClientInbound,
		NULL,
		NULL);		

	
	rtl8651_fwdEngineRegisterAlgModule(27015, FALSE,
		NULL,
		(alg_funcptr)rtl8651_l4AliasHandleCSClientOutbound,
		NULL,
		NULL,
		NULL);

	
	if((alg_support&RTL8651_ALG_TFTP)&&(alg_serverip[RTL8651_ALG_TFTP_IDX]))
	{		
		rtl8651a_addNaptServerPortMapping(	dsid,
											FALSE,
											gip,
											TFTP_UDP_DPORT,
											alg_serverip[RTL8651_ALG_TFTP_IDX],
											TFTP_UDP_DPORT);
	}	
	rtl8651_fwdEngineRegisterAlgModule(	TFTP_UDP_DPORT,
										FALSE,
										NULL,
										(alg_funcptr)rtl8651_l4AliasHandleTftpClientOutbound,
										NULL,
										NULL,
										NULL);

#if 0 
	
	rtl8651_fwdEngineRegisterAlgModule(1863, TRUE,
		rtl8651_l4AliasMmsInit,
		rtl8651_l4AliasHandleMmsCleintOutbound,
		NULL,
		NULL,
		NULL);
		
		
	rtl8651_fwdEngineRegisterAlgModule(137, FALSE,
		NULL,
		rtl8651_l4AliasHandleClientUdpNbtNSOut,
		rtl8651_l4AliasHandleClientUdpNbtNSIn,
		NULL,
		NULL);		
	rtl8651_fwdEngineRegisterAlgModule(138, FALSE,
		NULL,
		rtl8651_l4AliasHandleClientUdpNbtOut,
		rtl8651_l4AliasHandleClientUdpNbtIn,
		NULL,
		NULL);	
		
		
	rtl8651_fwdEngineRegisterAlgModule(RTSP_CONTROL_PORT_NUMBER_1, TRUE,
		rtl8651_l4AliasSmediaInit,
		rtl8651_l4AliasHandleSmediaCleintOutbound,
		NULL,
		NULL,
		NULL);		
	rtl8651_fwdEngineRegisterAlgModule(RTSP_CONTROL_PORT_NUMBER_2, TRUE,
		NULL,
		rtl8651_l4AliasHandleSmediaCleintOutbound,
		NULL,
		NULL,
		NULL);	
				
#endif 

	rtl8651_memcpy(alg_serverips[dsid],alg_serverip,32*sizeof(uint32));
	return OK;

}
#endif


int32 rtl8651_algOneSecondTimer(void){
	rtl8651_l4AliasPptpHouseKeeping();
	rtl8651_l4AliasIpsecHouseKeeping();
	rtl8651_l4AliasRtspHouseKeeping();
	return SUCCESS;
}

void rtl8651_algReInit(void)
{
	rtl8651_l4AliasHandleBattlenetReInit();
	rtl8651_l4AliasHandleCsReInit();
	rtl8651_l4AliasHandleDx7playReInit();
	rtl8651_l4AliasFtpReInit();
	rtl8651_l4AliasH323ReInit();
	rtl8651_l4AliasIcqReInit();	
	rtl8651_l4AliasIpsecReInit();	
	rtl8651_l4AliasPptpReInit();	
	rtl8651_l4AliasSipReInit();
	rtl8651_l4AliasYmsgReInit();
}



