

#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_alg_init.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_l2tp.h"

static uint32 last_ip=0;
static uint32 session_cnt=0;
static struct l2tp_session *session=NULL;


int32 rtl8651_l4AliasL2tpInit(void *arg)
{
	if(session) rtlglue_free(session);
	session = (struct l2tp_session *)rtlglue_malloc(sizeof(struct l2tp_session)*algMaxSession.l2tpMaxSession);
	memset((void*)session,0,sizeof(struct l2tp_session)*algMaxSession.l2tpMaxSession);	
	last_ip=0;
	session_cnt=0;
	return SUCCESS;
}


int32 rtl8651_l4AliasHandleL2tpClientOutbound(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *tb)
{
	struct udphdr *udp;
	int32 tlen;
	struct l2tp_header *l2tp;
	
	
	

	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_L2TP)==0) return FWDENG_USER_DROP;

	

	udp = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));

	if(udp->uh_sport!=udp->uh_dport) return SUCCESS;	

	tlen = ntohs(pip->ip_len); 

	if(tlen-36<0) return SUCCESS; 

	
	l2tp = (struct l2tp_header *)(udp+1);
	
	if(last_ip==0) last_ip=PKTGET_UINT32_UNALIGNED(&pip->ip_src.s_addr);

	#ifdef _RTL_UPNP
	
	if(tb->insideLocalPort==1701)
	{
		rtl8651_addUpnpMap(UPNP_ONESHOT, tb->dstIpAddr, 1701, 
			tb->insideGlobalIpAddr , 1701,
			tb->insideLocalIpAddr , 1701);
			tb->insideGlobalPort=1701;
	}
	#endif 	

	if((ntohs(l2tp->flag)&0xf4ff)==0xc002) 
	{
		if((l2tp->session_id==0)&&(l2tp->tunnel_id==0))
		{
			last_ip=PKTGET_UINT32_UNALIGNED(&pip->ip_src.s_addr);
			
		}
	}
	
	return SUCCESS;
}


int32 rtl8651_l4AliasHandleL2tpClientInbound(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *tb)
{

	struct udphdr *udp;
	int32 i;
	int32 tlen;
	struct l2tp_header *l2tp;

	
	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_L2TP)==0) return FWDENG_USER_DROP;

	
	if(last_ip==0)  return SUCCESS;

	udp = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));

	if(udp->uh_sport!=udp->uh_dport) return SUCCESS;

	tlen = ntohs(pip->ip_len); 
	if(tlen-36<0)  return SUCCESS;  
	l2tp = (struct l2tp_header *)(udp+1);
	
	
	for(i=0;i<algMaxSession.l2tpMaxSession;i++)
	{
		if((session[i].tunnel_id==l2tp->tunnel_id)&&(l2tp->tunnel_id!=0))
		{
			tb->insideLocalIpAddr=session[i].redirect_ip; 
			
			return SUCCESS;
		}
	}

	
	if((ntohs(l2tp->flag)&0xf4ff)==0xc002) 
	{
		for(i=0;i<algMaxSession.l2tpMaxSession;i++) 
		{	
			if(session[i].redirect_ip==last_ip)
			{
				session[i].tunnel_id=l2tp->tunnel_id;
				
				break;
			}
		}

		if(i==algMaxSession.l2tpMaxSession) 
		{
			session[session_cnt].tunnel_id=l2tp->tunnel_id;
			session[session_cnt].redirect_ip=last_ip;								
			
			++session_cnt;
			session_cnt%=algMaxSession.l2tpMaxSession;
		}
	}

	tb->insideLocalIpAddr=last_ip;
	return SUCCESS;
}



