

#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_smedia.h"

#define RTSP_PORT_GROUP            2

#define ISDIGIT(a) (((a) >= '0') && ((a) <= '9'))

#define _RTL8651_L4_MAX_SMEDIA_ALIAS_BUF_SIZE 576

static uint8 *_rtl8651_l4_smedia_alias_buf[2]; 

static int32 _rtl8651_smedia_search_string(int8 * , int32 , const int8 * );
static int32 _rtl8651_alias_rtsp_out(struct ip *, struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *,int8 * , const int8 * );
static int32 _rtl8651_alias_pna_out(struct ip *, struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *,int8 * , int32 );
						   
#define MIN_PORT 12000
#define MAX_PORT 14000

static uint16 _rtl8651_rtsp_port;


static uint16 _rtl8651_getRrtpPort(void){
	_rtl8651_rtsp_port++;	
	if(_rtl8651_rtsp_port < MIN_PORT || _rtl8651_rtsp_port > MAX_PORT)
		_rtl8651_rtsp_port = MIN_PORT;
		
	return _rtl8651_rtsp_port;	
}						   


int32
rtl8651_l4AliasHandleSmediaCleintOutbound(struct rtl_pktHdr *pkthdrPtr, struct ip *pip, 
				struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{
	uint32     hlen, tlen, dlen;
	struct tcphdr *tc;
	int8     *data;
	const int8 *setup = "SETUP", *pna = "PNA", *str200 = "200";
	const int8 *okstr = "OK", *client_port_str = "client_port";
	const int8 *server_port_str = "server_port";
	int32     i, parseOk;

	struct rtl_mBuf *m ;
	
	m  = pkthdrPtr->ph_mbuf;
	
	
	if ( mBuf_getPktlen(m) > _RTL8651_L4_MAX_SMEDIA_ALIAS_BUF_SIZE || mBuf_getPktlen(m) > _RTL8651_L4_MAX_SMEDIA_ALIAS_BUF_SIZE)
		return SUCCESS;

	
	mBuf_copyToUserBuffer(m, 0, mBuf_getPktlen(m), (int8*)_rtl8651_l4_smedia_alias_buf[0]);
	pip = (struct ip *) _rtl8651_l4_smedia_alias_buf[0];
	
	tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
	hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >>4)&0xf)) << 2;	
	tlen = ntohs(pip->ip_len);
	dlen = tlen - hlen;

	data = (int8 *) pip;
	data += hlen;

	
	if ((ntohs(tc->th_dport) == RTSP_CONTROL_PORT_NUMBER_1) ||
		(ntohs(tc->th_dport) == RTSP_CONTROL_PORT_NUMBER_2))
	{
		if (dlen >= strlen(setup))
		{
			if (memcmp(data, setup, strlen(setup)) == 0)
			{
				_rtl8651_alias_rtsp_out(pip, link, data, client_port_str);
				return SUCCESS;
			}
		}
		if (dlen >= strlen(pna))
		{
			if (memcmp(data, pna, strlen(pna)) == 0)
			{
				_rtl8651_alias_pna_out(pip, link, data, dlen);
			}
		}
	}
	else
	{

		

		if (dlen >= strlen(str200))
		{

			for (parseOk = 0, i = 0; i <= dlen - strlen(str200); i++)
			{
				if (memcmp(&data[i], str200, strlen(str200)) == 0)
				{
					parseOk = 1;
					break;
				}
			}
			if (parseOk)
			{
				i += strlen(str200);   
				while (data[i] == ' ') 
					i++;

				if ((dlen - i) >= strlen(okstr))
				{

					if (memcmp(&data[i], okstr, strlen(okstr)) == 0)
						_rtl8651_alias_rtsp_out(pip, link, data, server_port_str);

				}
			}
		}
	}

	
	tc->th_sum = 0;
	pip->ip_sum = 0;				
	tc->th_sum = _rtl8651_tcpChecksum(pip);
	pip->ip_sum = _rtl8651_ipChecksum(pip);	

	
	mBuf_copyToMbuf(m, 0, ntohs(pip->ip_len), (int8*)_rtl8651_l4_smedia_alias_buf[0]);

	return SUCCESS;
}


static    int32
_rtl8651_smedia_search_string(int8 * data, int32 dlen, const int8 * search_str)
{
	int32     i, j, k;
	int32     search_str_len;

	search_str_len = strlen(search_str);
	for (i = 0; i < dlen - search_str_len; i++)
	{
		for (j = i, k = 0; j < dlen - search_str_len; j++, k++)
		{
			if (data[j] != search_str[k] &&
				data[j] != search_str[k] - ('a' - 'A'))
			{
				break;
			}
			if (k == search_str_len - 1)
			{
				return j + 1;
			}
		}
	}
	return -1;
}


static    int32
_rtl8651_alias_rtsp_out(struct ip *pip,
			   struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link, int8 * data, const int8 * port_str)
{
	uint32     hlen = 0, tlen = 0, dlen = 0, port_dlen = 0, new_dlen = 0;
	struct tcphdr *tc;
	int32     i, j, pos, state, delta;
	uint16    p[2], new_len;
	uint16    sport, eport, base_port;
	uint16    salias = 0, ealias = 0, base_alias = 0;
	const int8 *transport_str = "transport:";
	int8      newdata[2048], *port_data, *port_newdata, stemp[80];
	int32     links_created = 0, pkt_updated = 0;

	
	tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
	hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >>4)&0xf)) << 2;
	tlen = ntohs(pip->ip_len);
	dlen = tlen - hlen;

	
	pos = _rtl8651_smedia_search_string(data, dlen, transport_str);
	if (pos < 0)
	{
		return -1;
	}
	port_data = data + pos;
	port_dlen = dlen - pos;

	rtl8651_memcpy(newdata, data, pos);
	port_newdata = newdata + pos;

	while ( port_dlen > strlen(port_str))
	{
		
		pos = _rtl8651_smedia_search_string(port_data, port_dlen, port_str);
		if (pos < 0)
		{
			break;
		}

		rtl8651_memcpy(port_newdata, port_data, pos + 1);
		port_newdata += (pos + 1);

		p[0] = p[1] = 0;
		sport = eport = 0;
		state = 0;
		for ( i = pos; i < port_dlen; i++)
		{
			switch (state)
			{
				case 0:
					if (port_data[i] == '=')
					{
						state++;
					}
					break;
				case 1:
					if (_rtl8651_isdigit(port_data[i]))
					{
						p[0] = p[0] * 10 + port_data[i] - '0';
					}
					else
					{
						if (port_data[i] == ';')
						{
							state = 3;
						}
						if (port_data[i] == '-')
						{
							state++;
						}
					}
					break;
				case 2:
					if (_rtl8651_isdigit(port_data[i]))
					{
						p[1] = p[1] * 10 + port_data[i] - '0';
					}
					else
					{
						state++;
					}
					break;
				case 3:
					base_port = p[0];
					sport = htons(p[0]);
					eport = htons(p[1]);

					if (!links_created)
					{

						links_created = 1;
						
						
						if (0 == (salias = _rtl8651_getRrtpPort()) || link->natlink )
						{
						   salias = 0; 
						}
						else
						{

							base_alias = ntohs(salias);
							for (j = 0; j < RTSP_PORT_GROUP; j++)
							{
								
								
#ifdef _RTL_UPNP
								
								rtl8651_addUpnpMap(UPNP_ONESHOT, link->dstIpAddr, 0, 
									link->insideGlobalIpAddr, base_alias + j,
									link->insideLocalIpAddr, base_port + j);		
#endif 										
							}
						}
						ealias = htons(base_alias + (RTSP_PORT_GROUP - 1));
					}

					if (salias )
					{

						pkt_updated = 1;

						
						sprintf(stemp, "%d", ntohs(salias));
						rtl8651_memcpy(port_newdata, stemp, strlen(stemp));
						port_newdata += strlen(stemp);

						if (eport != 0 )
						{
							*port_newdata = '-';
							port_newdata++;

							
							sprintf(stemp, "%d", ntohs(ealias));
							rtl8651_memcpy(port_newdata, stemp, strlen(stemp));
							port_newdata += strlen(stemp);
						}

						*port_newdata = ';';
						port_newdata++;
					}
					state++;
					break;
			}
			if (state > 3)
			{
				break;
			}
		}
		port_data += i;
		port_dlen -= i;
	}

	if (!pkt_updated)
		return -1;

	if(link->natlink == 0){
		rtl8651_memcpy(port_newdata, port_data, port_dlen);
		port_newdata += port_dlen;
		*port_newdata = '\0';

		
		new_dlen = port_newdata - newdata;
		rtl8651_memcpy(data, newdata, new_dlen);
	}

	link->tcpAckModified = 1;
	delta = _rtl8651_l4GetDeltaSeqOut(pip, tc, link);
	_rtl8651_l4AddSeq(pip, link, delta + new_dlen - dlen);

	new_len = htons(hlen + new_dlen);
	pip->ip_len = new_len;

	
	tc->th_sum = 0;
	pip->ip_sum = 0;				
	tc->th_sum = _rtl8651_tcpChecksum(pip);
	pip->ip_sum = _rtl8651_ipChecksum(pip);	

	return 0;
}



static int32
_rtl8651_alias_pna_out(struct ip *pip,
			  struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link, int8 * data, int32 dlen)
{
	uint16    msg_id, msg_len;
	int8     *work;
	uint16    alias_port, port;

	work = data;
	work += 5;
	while (work + 4 < data + dlen)
	{
		rtl8651_memcpy(&msg_id, work, 2);
		work += 2;
		rtl8651_memcpy(&msg_len, work, 2);
		work += 2;
		if (ntohs(msg_id) == 0)
		{
			
			return 0;
		}
		if ((ntohs(msg_id) == 1) || (ntohs(msg_id) == 7))
		{
			rtl8651_memcpy(&port, work, 2);
			
		
#ifdef _RTL_UPNP		
		alias_port = _rtl8651_getRrtpPort();
		rtl8651_addUpnpMap(UPNP_ONESHOT, link->dstIpAddr, 0, 
				link->insideGlobalIpAddr, alias_port,
				link->insideLocalIpAddr, port);	
		rtl8651_memcpy(work, &alias_port, 2);
#endif 			
		}
		work += ntohs(msg_len);
	}
	return 0;
}



static int32 init_smedia = 0;
int32 rtl8651_l4AliasSmediaInit(void *arg){
	
	
	
	if(init_smedia)
		return SUCCESS;
		
	init_smedia = 1;	
	_rtl8651_l4_smedia_alias_buf[0] = (uint8 *)rtlglue_malloc(_RTL8651_L4_MAX_SMEDIA_ALIAS_BUF_SIZE+64);
	_rtl8651_l4_smedia_alias_buf[1] = (uint8 *)rtlglue_malloc(_RTL8651_L4_MAX_SMEDIA_ALIAS_BUF_SIZE+64);
	assert(_rtl8651_l4_smedia_alias_buf[0]);
	assert(_rtl8651_l4_smedia_alias_buf[1]);

	return SUCCESS;
}


