


#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_errno.h"
#include "rtl_glue.h"
#include "mbuf.h"

#include "mbuf.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_dns_local.h"





static	_rtl8651_dnsDB_ctrl_t _rtl8651_dns_ctrlBlock;
uint32	_rtl8651_dns_enableTTLSuppression			= FALSE;	
uint32	_rtl8651_dns_suppressedTTL					= 0;		

static int __inet_aton(const char *cp, ipaddr_t *addrptr);
static ipaddr_t __inet_addr(const char *cp);

static _rtl8651_dnsDB_t *__rtl8651_dns_getDnsDB(uint8 process, char *domain, _rtl8651_dnsDB_t *start, uint8 isExactlyMatch);
static _rtl8651_dnsDB_t *__rtl8651_dns_addAndLinkDnsDB(uint8 process);
static int32 __rtl8651_dns_UnlinkAndDelDnsDB(_rtl8651_dnsDB_t *dnsDb_p);
static int32 __rtl8651_dns_destroyDnsDB(uint8 process, _rtl8651_dnsDB_t *dnsDb_p);
static int32 __rtl8651_dns_flushDataBase(void);
static int32 _rtl8651_dns_dataBase_init(rtl8651_fwdEngineInitPara_t *para);
static int32 _rtl8651_dns_dataBase_reinit(void);
static int32 _rtl8651_dns_funcCallBack(char *mappedDomain, ipaddr_t mappedIp);



static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameMap_p = NULL;
static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameLru_h = NULL;
static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameLru_t = NULL;
static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameFree_p = NULL;
static uint32 _rtl8651_dns_cnameTimeout = 0;

static int32 _rtl8651_dns_cnameMapping_init(rtl8651_fwdEngineInitPara_t *para);
static void _rtl8651_dns_cnameMapping_reinit(void);
static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameMapping_find(char *name);
static inline _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameMapping_getEntry(void);
static void _rtl8651_dns_cnameMapping_refresh(_rtl8651_dns_cnameMapping_t *entry);
static int32 _rtl8651_dns_cnameMapping_add(char *name1, char *name2);
static void _rtl8651_dns_cnameMapping_del(_rtl8651_dns_cnameMapping_t *delEntry);
static void _rtl8651_dns_cnameMapping_TimeUpdate(uint32 secPassed);
static int32 _rtl8651_dns_procDnsCallBack(ipaddr_t mapped_ip, char *mapped_domain, int32 (*dns_proc)(char *, ipaddr_t));

static int32 _RTL8651_DNS_GET_NAMELEN(uint8 *name, uint32 maxLen, uint32 *len_p);
static int32 _RTL8651_DNS_GET_INFO(_rtl8651_dns_dnsDomain_t domainType, uint8 *outputPtr, uint32 maxOutputLen, uint8 *inputPtr, uint8 *dnsData_p, uint32 dnsDataLen);
static int32 _rtl8651_dns_parseDnsPkt(uint8 *dns_p, uint32 dnsLen, uint8 *pktModified, int32 (*dns_proc)(char *, ipaddr_t));
static int32 _rtl8651_dns_parseDnsRR(uint8 *dns_p, uint32 totalLen, uint8 *rr_p, uint32 maxLen, uint32 *rrLen_p, uint8 *pktModified, int32 (*dns_proc)(char *, ipaddr_t));




#define RTL8651_DNS_DATABASE

#define __rtl_isupper(c)	(((c) >= 'A') && ((c) <= 'Z'))


static inline void __strlowerncpy(char *dst, const char *src, int32 len)
{
	int32 cpLen = 0;
	char p;

	while ((cpLen < len) && (src[cpLen] != 0))
	{
		p = src[cpLen];
 		dst[cpLen] = __rtl_isupper(p)?(p-'A'+'a'):p;
		cpLen ++;
	}
}

static inline int __inet_aton(const char *cp, ipaddr_t *addrptr)
{
	ipaddr_t addr;
	int value;
	int part;

	addr = 0;
	for (part = 1; part <= 4; part++)
	{
		if (!isdigit(*cp))
			return FAILED;

		value = 0;
		while (isdigit(*cp))
		{
			value *= 10;
			value += *cp++ - '0';
			if (value > 255)
				return FAILED;
		}

		if (part < 4)
		{
			if (*cp++ != '.')
				return FAILED;
		} else
		{
			char c = *cp++;
			if (c != '\0' && !isspace(c))
			return FAILED;
		}

		addr <<= 8;
		addr |= value;
	}

	
	if (addrptr)
	{
		PKTSET_UINT32_UNALIGNED(addr, addrptr);
	}

	return SUCCESS;
}

static ipaddr_t __inet_addr(const char *cp)
{
	ipaddr_t a;

	if (__inet_aton(cp, &a) != SUCCESS)
	{
		return 0;
	}else
	{
		return a;
	}
}


static _rtl8651_dnsDB_t *__rtl8651_dns_getDnsDB(
	uint8 process,
	char *domain,
	_rtl8651_dnsDB_t *start,
	uint8 isExactlyMatch)
{
	_rtl8651_dnsDB_t *ptr = NULL;
	uint32 domainLen = 0;

	if (process >= _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		RTL_DNS_ERR("Invalid process %d", process);
		goto out;
	}

	domainLen = strlen(domain);

	if ((domainLen >= RTL8651_DNS_DOMAIN_MAXLENGTH) || (domainLen == 0))
	{
		goto out;
	}

	ptr = (start)?(start->next):_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process];

	while (ptr)
	{
		void *matchPos;

		
		if (	(isExactlyMatch == TRUE) && (strlen(ptr->domain) != strlen(domain)))
		{
			goto next;
		}

		if ((matchPos = (void*)strstr(domain, ptr->domain)) != NULL)
		{
			
			if (	(isExactlyMatch == TRUE) && (matchPos != domain))
			{
				goto next;
			}
			
			goto out;
		}
	next:
		ptr = ptr->next;
	}

out:
	return ptr;
}


static _rtl8651_dnsDB_t *__rtl8651_dns_addAndLinkDnsDB(uint8 process)
{
	_rtl8651_dnsDB_t *ptr = NULL;

	if (process >= _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		RTL_DNS_ERR("Invalid process %d", process);
		goto out;
	}

	if (_rtl8651_dns_ctrlBlock.freeDnsDB)
	{
		
		ptr = _rtl8651_dns_ctrlBlock.freeDnsDB;
		_rtl8651_dns_ctrlBlock.freeDnsDB = (_rtl8651_dns_ctrlBlock.freeDnsDB)->next;

		
		ptr->hdr = &(_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process]);
		ptr->prev = NULL;
		ptr->next = _rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process];

		if (_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process])
		{
			(_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process])->prev = ptr;
		}

		_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process] = ptr;

		
		memset(ptr->domain, 0, RTL8651_DNS_DOMAIN_MAXLENGTH);
		ptr->procData = NULL;

	}
out:
	return ptr;
}


static int32 __rtl8651_dns_UnlinkAndDelDnsDB(_rtl8651_dnsDB_t *dnsDb_p)
{
	_rtl8651_dnsDB_t *hdr = NULL;

	if (	(dnsDb_p == NULL			) ||
		(dnsDb_p->procData != NULL	))
	{
		RTL_DNS_ERR("Invalid input:");
		RTL_DNS_ERR("\tDnsDB: %p", dnsDb_p);
		if (dnsDb_p)
		{
			RTL_DNS_ERR("\tDnsDB procData: %p (must be cleaned before delete this DB entry)", dnsDb_p->procData);
		}
		return FAILED;
	}

	
	if (dnsDb_p->next)
	{
		dnsDb_p->next->prev = dnsDb_p->prev;
	}
	if (dnsDb_p->prev)
	{
		dnsDb_p->prev->next = dnsDb_p->next;
	}

	hdr = *(_rtl8651_dnsDB_t**)(dnsDb_p->hdr);
	if (hdr == dnsDb_p)
	{
		*(dnsDb_p->hdr) = dnsDb_p->next;
	}

	
	memset(dnsDb_p, 0, sizeof(_rtl8651_dnsDB_t));
	dnsDb_p->next = _rtl8651_dns_ctrlBlock.freeDnsDB;
	_rtl8651_dns_ctrlBlock.freeDnsDB = dnsDb_p;

	return SUCCESS;

}


static int32 __rtl8651_dns_destroyDnsDB(uint8 process, _rtl8651_dnsDB_t *dnsDb_p)
{
	
	if (dnsDb_p->procData)
	{
		
		if (_rtl8651_dns_ctrlBlock.callBack[process].destructor)
		{
			RTL_DNS_INFO("[START] destructor(%p) for process (%d), procData (%p)",
							_rtl8651_dns_ctrlBlock.callBack[process].destructor,
							process,
							dnsDb_p->procData);

			_rtl8651_dns_ctrlBlock.callBack[process].destructor(dnsDb_p->procData);

			RTL_DNS_INFO("[END] destructor OK");
		} else
		{
			RTL_DNS_WARN("No destructor for process (%d), procData (%p)",
							process,
							dnsDb_p->procData);
		}

		
		dnsDb_p->procData = NULL;
	}

	return __rtl8651_dns_UnlinkAndDelDnsDB(dnsDb_p);
}


static int32 __rtl8651_dns_flushDataBase(void)
{
	int32 idx, retval = FAILED;

	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		while (_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx])
		{
			retval = _rtl8651_dns_delProcessEntry(	idx,
													TRUE,	
													_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx],
													NULL,
													NULL);
			if (retval != SUCCESS)
			{
				RTL_DNS_ERR("FATAL ERROR, clear DNS structure FAILED");
				goto out;
			}
		}
	}

	
	retval = SUCCESS;
out:
	return retval;
}


static int32 _rtl8651_dns_dataBase_init(rtl8651_fwdEngineInitPara_t *para)
{
	
	memset(&_rtl8651_dns_ctrlBlock, 0 , sizeof(_rtl8651_dns_ctrlBlock));

	
	if (	(para == NULL) ||
		(para->dns_dnsDB_tblSize == 0))
	{
		if (para)
		{
			RTL_DNS_WARN("Variable err: dns table size [%d]", para->dns_dnsDB_tblSize);
		} else
		{
			RTL_DNS_WARN("Variable err: para [%p]", para);
		}
		goto err_out;
	}

	_rtl8651_dns_ctrlBlock.dnsDB_unusedProcCnt = para->dns_dnsDB_processCnt;

	
	MEM_ALLOC(_rtl8651_dns_ctrlBlock.freeDnsDB, sizeof(_rtl8651_dnsDB_t) * para->dns_dnsDB_tblSize, FAILED);
	DNS_LIST_INIT(_rtl8651_dns_ctrlBlock.freeDnsDB, _rtl8651_dnsDB_t, para->dns_dnsDB_tblSize, next);

	MEM_ALLOC(_rtl8651_dns_ctrlBlock.rtl8651_inusedDB, sizeof(_rtl8651_dnsDB_t*) * para->dns_dnsDB_processCnt, FAILED);
	MEM_ALLOC(_rtl8651_dns_ctrlBlock.callBack, sizeof(_rtl8651_dnsFuncCtrl_t*) * para->dns_dnsDB_processCnt, FAILED);

	
	memcpy(&(_rtl8651_dns_ctrlBlock.initPara), para, sizeof(rtl8651_fwdEngineInitPara_t));

	return SUCCESS;

err_out:
	return FAILED;
}


static int32 _rtl8651_dns_dataBase_reinit(void)
{
	return __rtl8651_dns_flushDataBase();
}

static int32 _rtl8651_dns_funcCallBack(char *mappedDomain, ipaddr_t mappedIp)
{
	int32 idx;
	_rtl8651_dnsDB_t *db_p;

	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.callBack[idx].proc)
		{
			db_p = NULL;
			while ((db_p = __rtl8651_dns_getDnsDB(
							idx,
							mappedDomain,
							db_p,
							(_rtl8651_dns_ctrlBlock.callBack[idx].property&RTL8651_DNSDB_PROPERTY_EXACTLYMATCH)?
							TRUE:FALSE))
					!= NULL)
			{
				_rtl8651_dns_ctrlBlock.callBack[idx].proc(db_p->domain, mappedDomain, mappedIp, db_p->procData);
			}
		}
	}
	return SUCCESS;
}


void rtl8651_dns_dataBase_dump(void)
{
	int32 idx;
	_rtl8651_dnsDB_t *db_p;

	uint32 inusedCnt = 0;

	rtlglue_drvMutexLock();

	printk("\nDNS database table:\n\n");

	for (idx = 0 ; idx < (_rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt - _rtl8651_dns_ctrlBlock.dnsDB_unusedProcCnt); idx ++)
	{
		printk("Process [%d]\n", idx);

		printk("\tProperty:\t\t\t\t0x%x\n", _rtl8651_dns_ctrlBlock.callBack[idx].property);
		printk("\tREINIT call-back function:\t\t%p\n", _rtl8651_dns_ctrlBlock.callBack[idx].reinit);
		printk("\tPROC call-back function:\t\t%p\n", _rtl8651_dns_ctrlBlock.callBack[idx].proc);
		printk("\tTimeUpdate call-back function:\t\t%p\n", _rtl8651_dns_ctrlBlock.callBack[idx].timeUpdate);
		printk("\tDELETE call-back function:\t\t%p\n", _rtl8651_dns_ctrlBlock.callBack[idx].destructor);

		printk("\n");

		db_p = _rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx];
		if (db_p)
		{
			printk("\t[DATA PTR]\t[DOMAIN]\n");
			printk("\t===================================================\n");
			while (db_p)
			{
				printk("\t%p\t%s\n", db_p->procData, db_p->domain);
				inusedCnt ++;
				db_p = db_p->next;
			}
			printk("\n");
		}
	}
	printk("\n\tInused:\t%d\n", inusedCnt);
	printk("\tFree:\t%d\n", _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_tblSize - inusedCnt);

	rtlglue_drvMutexUnlock();
}


#define RTL8651_DNS_EXTERNAL


int32 rtl8651_dns_setDnsTtlSuppression(uint8 enable, uint32 suppressedTtl)
{
	rtlglue_drvMutexLock();
	_rtl8651_dns_enableTTLSuppression = enable?TRUE:FALSE;
	_rtl8651_dns_suppressedTTL = suppressedTtl;
	rtlglue_drvMutexUnlock();

	return enable;
}


int32 rtl8651_dns_registerProcess(
	uint8 *process,
	uint32 property,
	int32 (*init)(void *para),
	void (*reinit)(void),
	int32 (*proc)(char *orig_domain, char *domain, ipaddr_t ip, void *data),
	void (*timeUpdate)(uint32 secpassed),
	int32 (*destructor)(void *data)
	)
{
	int32 retval;

	rtlglue_drvMutexLock();
	retval = _rtl8651_dns_registerProcess(	process,
											property,
											init,
											reinit,
											proc,
											timeUpdate,
											destructor);
	rtlglue_drvMutexUnlock();
	return retval;
}


int32 _rtl8651_dns_registerProcess(
	uint8 *process,
	uint32 property,
	int32 (*init)(void *para),
	void (*reinit)(void),
	int32 (*proc)(char *orig_domain, char *domain, ipaddr_t ip, void *data),
	void (*timeUpdate)(uint32 secpassed),
	int32 (*destructor)(void *data)
	)
{
	if (_rtl8651_dns_ctrlBlock.dnsDB_unusedProcCnt == 0)
	{
		RTL_DNS_ERR("NO Free entry for new DNS process! Please increase \"dns_dnsDB_processCnt\" of forwarding engine parameter.");
		return FAILED;
	}

	assert(process);

	*process = (_rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt - _rtl8651_dns_ctrlBlock.dnsDB_unusedProcCnt);

	RTL_DNS_INFO("Register function into DNS module.");
	RTL_DNS_INFO("================================");
	RTL_DNS_INFO("property:\t\t%d", property);
	RTL_DNS_INFO("init callBack:\t%p", init);
	RTL_DNS_INFO("reinit callBack:\t%p", reinit);
	RTL_DNS_INFO("proc callBack:\t%p", proc);
	RTL_DNS_INFO("timeUpdate:\t%p", timeUpdate);
	RTL_DNS_INFO("desructor:\t%p", destructor);
	RTL_DNS_INFO("================================");

	_rtl8651_dns_ctrlBlock.callBack[*process].property = property;
	_rtl8651_dns_ctrlBlock.callBack[*process].reinit = reinit;
	_rtl8651_dns_ctrlBlock.callBack[*process].proc = proc;
	_rtl8651_dns_ctrlBlock.callBack[*process].timeUpdate = timeUpdate;
	_rtl8651_dns_ctrlBlock.callBack[*process].destructor = destructor;

	if (init)
	{
		DNS_UNLOCK_ERROR_CHECK(init(&(_rtl8651_dns_ctrlBlock.initPara)), SUCCESS, TRUE, FAILED);	
	}

	_rtl8651_dns_ctrlBlock.dnsDB_unusedProcCnt --;

	RTL_DNS_INFO("--------> Registration OK.");

	return SUCCESS;
}


int32 rtl8651_dns_addProcessEntry(
	uint8 process,
	char *domain,
	void *procData,
	void **entryPtr)
{
	int32 retval;
	rtlglue_drvMutexLock();
	retval = _rtl8651_dns_addProcessEntry(	process,
											domain,
											procData,
											entryPtr);
	rtlglue_drvMutexUnlock();
	return retval;
}


int32 _rtl8651_dns_addProcessEntry(
	uint8 process,
	char *domain,
	void *procData,
	void **entryPtr)
{
	_rtl8651_dnsDB_t * dnsDb_p = NULL;
	int32 idx;
	int32 origTrapStatus = FAILED;
	uint32 domainLen = 0;

	if (process >= _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		RTL_DNS_ERR("Invalid process %d", process);
		goto err_out;
	}

	domainLen = strlen(domain);

	if ((domainLen >= RTL8651_DNS_DOMAIN_MAXLENGTH) || (domainLen == 0))
	{
		RTL_DNS_ERR("Domain name (%s) length is invalid: %d", domain, domainLen);
		goto err_out;
	}

	
	dnsDb_p = NULL;
	while ((dnsDb_p = __rtl8651_dns_getDnsDB(	process,
												domain,
												dnsDb_p,
												TRUE)) != NULL)
	{
		if (dnsDb_p->procData == procData)
		{
			RTL_DNS_WARN("Duplicate Entry: process: %d domain: %s", process, domain);
			goto ok_out;
		}
	}

	
	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx])
		{
			origTrapStatus = TRUE;
			break;
		}
	}

	
	if ((dnsDb_p = __rtl8651_dns_addAndLinkDnsDB(process)) != NULL)
	{
		
		__strlowerncpy(dnsDb_p->domain, domain, RTL8651_DNS_DOMAIN_MAXLENGTH);
		dnsDb_p->procData = procData;
	}

	
	_rtl8651_arrangeAllACL();

ok_out:

	if (entryPtr)
	{
		*entryPtr = (void*)dnsDb_p;
	}

	return SUCCESS;

err_out:
	return FAILED;
}


int32 rtl8651_dns_delProcessEntry(
	uint8 process,
	uint8 byDbPtr,						
	void *db_p,
	char *domain,
	void *procData)
{
	int32 retval;

	rtlglue_drvMutexLock();
	retval = _rtl8651_dns_delProcessEntry(	process,
											byDbPtr,
											db_p,
											domain,
											procData);
	rtlglue_drvMutexUnlock();
	return retval;
}


int32 _rtl8651_dns_delProcessEntry(
	uint8 process,
	uint8 byDbPtr,						
	void *db_p,
	char *domain,
	void *procData)
{
	_rtl8651_dnsDB_t * dnsDb_p = NULL;
	int32 idx;
	uint8 domainCmp[RTL8651_DNS_DOMAIN_MAXLENGTH];

	if (process >= _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		RTL_DNS_ERR("Invalid process %d", process);
		goto err_out;
	}

	if (byDbPtr == TRUE)
	{
		if (db_p == NULL)
		{
			RTL_DNS_ERR("delete null entry.");
			goto err_out;
		}
		
		if (__rtl8651_dns_destroyDnsDB(process, (_rtl8651_dnsDB_t *)db_p) != SUCCESS)
		{
			goto err_out;
		}
	} else
	{
		uint32 domainLen = 0;

		domainLen = strlen(domain);

		if ((domainLen >= RTL8651_DNS_DOMAIN_MAXLENGTH) || (domainLen == 0))
		{
			RTL_DNS_ERR("Domain name (%s) length is invalid: %d", domain, domainLen);
			goto err_out;
		}

		
		memset(domainCmp, 0, sizeof(domainCmp));
		__strlowerncpy(domainCmp, domain, strlen(domain));

		
		dnsDb_p = NULL;
		while ((dnsDb_p = __rtl8651_dns_getDnsDB(	process,
													domain,
													dnsDb_p,
													TRUE)) != NULL)
		{
			if (procData && (dnsDb_p->procData != procData))
				continue;

			if (__rtl8651_dns_destroyDnsDB(process, dnsDb_p) != SUCCESS)
			{
				goto err_out;
			}
		}
	}

	
	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx])	break;
	}
	if (idx == _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		_rtl8651_arrangeAllACL();
	}

	return SUCCESS;

err_out:
	return FAILED;
}


int32 rtl8651_dns_findProcessEntry(
	uint8 process,
	char *domain,
	void *procData,
	char **returnedDomain,
	void **returnedProcData,
	void **start)
{
	int32 retval;

	rtlglue_drvMutexLock();
	retval = _rtl8651_dns_findProcessEntry(	process,
											domain,
											procData,
											returnedDomain,
											returnedProcData,
											start);
	rtlglue_drvMutexUnlock();
	return retval;
}

int32 _rtl8651_dns_findProcessEntry(
	uint8 process,
	char *domain,
	void *procData,
	char **returnedDomain,
	void **returnedProcData,
	void **start)
{
	_rtl8651_dnsDB_t * dnsDb_p = NULL;

	if (process >= _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		RTL_DNS_ERR("Invalid process %d", process);
		return FALSE;
	}

	dnsDb_p = start?(*start):NULL;

	while ((dnsDb_p = __rtl8651_dns_getDnsDB(	process,
												domain,
												dnsDb_p,
												TRUE)) != NULL)
	{	
		if (procData && (procData != dnsDb_p->procData))
			continue;

		break;	
	}

	if (start)
	{
		*start = dnsDb_p;
	}
	if (returnedProcData)
	{
		*returnedProcData = dnsDb_p?dnsDb_p->procData:NULL;
	}
	if (returnedDomain)
	{
		*returnedDomain = dnsDb_p?dnsDb_p->domain:NULL;
	}
	return (dnsDb_p)?TRUE:FALSE;
}



char *rtl8651_dns_getProcessEntryDomainInfo(void *db_p)
{
	char *ret;

	rtlglue_drvMutexLock();
	ret = _rtl8651_dns_getProcessEntryDomainInfo(db_p);
	rtlglue_drvMutexUnlock();
	return ret;
}

char *_rtl8651_dns_getProcessEntryDomainInfo(void *db_p)
{
	if (db_p)
	{
		return ((_rtl8651_dnsDB_t*)db_p)->domain;
	}

	return NULL;
}


#define RTL8651_DNS_PKTPROC


void rtl8651_dns_dnsProc_dump(void)
{
	_rtl8651_dns_cnameMapping_t *ptr, *hdr;
	uint32 inusedCnt = 0, freeCnt = 0;

	rtlglue_drvMutexLock();

	printk("\nDNS Process module:\n\n");

	printk("\tTTL Suppression:\t%s\n", _rtl8651_dns_enableTTLSuppression==TRUE?"TRUE":"FALSE");
	if ( _rtl8651_dns_enableTTLSuppression==TRUE )
	{
		printk("\tTTL upper bound:\t%d\n", _rtl8651_dns_suppressedTTL);
	}
	printk("\n");


	ptr = _rtl8651_dns_cnameLru_h;
	while ( ptr )
	{
		inusedCnt ++;
		ptr = ptr->lru_next;
	}
	ptr = _rtl8651_dns_cnameFree_p;
	while ( ptr )
	{
		freeCnt ++;
		ptr = ptr->next;
	}

	printk("\nDNS CNAME mapping table:\n\n");
	printk("\tInused:\t\t%d\n", inusedCnt);
	printk("\tFree:\t\t%d\n", freeCnt);
	printk("\tTIMEOUT:\t%d\n", _rtl8651_dns_cnameTimeout);
	if (inusedCnt > 0)
	{
		int32 idx = 1;
		printk("\n\t[NO]\t[Timeout]\t[DOMAIN NAME]\n");
		printk("\t============================================================\n");
		hdr = _rtl8651_dns_cnameMap_p;
		while ( hdr )
		{
			printk("\t( %d )\n", idx ++);
			ptr = hdr;
			while ( ptr )
			{
				printk("\t\t%d\t\t%s\n", 	(ptr->updateTime + _rtl8651_dns_cnameTimeout) - _sysUpSeconds,
										ptr->addr);
				ptr = ptr->next;
			}
			hdr = hdr->hdr_next;
		}
	}
	printk("\t============================================================\n");

	rtlglue_drvMutexUnlock();
}


static int32 _rtl8651_dns_cnameMapping_init(rtl8651_fwdEngineInitPara_t *para)
{

	assert(para);

	_rtl8651_dns_cnameMap_p = NULL;
	_rtl8651_dns_cnameLru_h = NULL;
	_rtl8651_dns_cnameLru_t = NULL;
	_rtl8651_dns_cnameFree_p = NULL;
	_rtl8651_dns_cnameTimeout = 0;

	if (para->dns_dnsDB_cnameMapTblSize == 0)
	{
		RTL_DNS_WARN(	"** Disable CNAME mapping: para->dns_dnsDB_tblSize = %d",
						para->dns_dnsDB_tblSize);
		return SUCCESS;
	}

	MEM_ALLOC(_rtl8651_dns_cnameFree_p, para->dns_dnsDB_cnameMapTblSize * sizeof(_rtl8651_dns_cnameMapping_t), FAILED);
	DNS_LIST_INIT(_rtl8651_dns_cnameFree_p, _rtl8651_dns_cnameMapping_t, para->dns_dnsDB_cnameMapTblSize, next);
 
	
	_rtl8651_dns_cnameTimeout = para->dns_dnsDB_cnameMapTimeout;

	return SUCCESS;
}

static void _rtl8651_dns_cnameMapping_reinit(void)
{
	while (_rtl8651_dns_cnameMap_p)
	{
		_rtl8651_dns_cnameMapping_del(_rtl8651_dns_cnameMap_p);
	}
}

static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameMapping_find(char *name)
{
	_rtl8651_dns_cnameMapping_t *ptr, *hdr;
	uint32 nameLen = strlen(name);

	hdr = _rtl8651_dns_cnameMap_p;

	while (hdr)
	{
		ptr = hdr;

		while (ptr)
		{
			if (strlen(ptr->addr) != nameLen)
			{
				goto next;
			}

			if (strncmp(ptr->addr, name, nameLen) == 0)
			{
				return ptr;
			}

		next:
			ptr = ptr->next;
		}

		hdr = hdr->hdr_next;
	}

	return NULL;
}

static inline _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameMapping_getEntry(void)
{
	_rtl8651_dns_cnameMapping_t *entry = NULL;

	if (_rtl8651_dns_cnameFree_p)
	{	
		entry = _rtl8651_dns_cnameFree_p;
		_rtl8651_dns_cnameFree_p = _rtl8651_dns_cnameFree_p->next;
	} else
	{	
		if (_rtl8651_dns_cnameLru_h)
		{
			_rtl8651_dns_cnameMapping_del(_rtl8651_dns_cnameLru_h);
			assert(entry);
		}
	}

	
	if (entry)
	{
		memset(entry, 0, sizeof(_rtl8651_dns_cnameMapping_t));
	}

	return entry;
}

static void _rtl8651_dns_cnameMapping_refresh(_rtl8651_dns_cnameMapping_t *entry)
{
	entry->updateTime = _sysUpSeconds;

	if (_rtl8651_dns_cnameLru_t != entry)
	{
		
		if (entry->lru_next)
		{
			entry->lru_next->lru_prev = entry->lru_prev;
		}
		if (entry->lru_prev)
		{
			entry->lru_prev->lru_next = entry->lru_next;
		}
		if (_rtl8651_dns_cnameLru_h == entry)
		{
			_rtl8651_dns_cnameLru_h = entry->lru_next;
		}

		entry->lru_next = NULL;
		entry->lru_prev = NULL;

		
		if (_rtl8651_dns_cnameLru_t)
		{
			_rtl8651_dns_cnameLru_t->lru_next = entry;
			entry->lru_prev = _rtl8651_dns_cnameLru_t;
		} else
		{
			assert(_rtl8651_dns_cnameLru_h == NULL);
			_rtl8651_dns_cnameLru_h = entry;
		}

		_rtl8651_dns_cnameLru_t = entry;
	}
}


static int32 _rtl8651_dns_cnameMapping_add(char *name1, char *name2)
{
	_rtl8651_dns_cnameMapping_t *name1_p, *name2_p, *exist_p, *new_p;

	name1_p = _rtl8651_dns_cnameMapping_find(name1);
	name2_p = _rtl8651_dns_cnameMapping_find(name2);

	if ((name1_p) && (name2_p))
	{
		
		_rtl8651_dns_cnameMapping_t *first, *ptr;
		uint8 name1_ok, name2_ok;

		
		first = name1_p;
		while (first->prev)
		{
			first = first->prev;
		}

		ptr = first;
		name1_ok = FALSE;
		name2_ok = FALSE;
		while (ptr)
		{
			if (ptr == name1_p)
			{
				name1_ok = TRUE;
			}
			if (ptr == name2_p)
			{
				name2_ok = TRUE;
			}
			ptr = ptr->next;
		}

		if ((name1_ok != TRUE) || (name2_ok != TRUE))
		{	
			_rtl8651_dns_cnameMapping_del(name1_p);
			_rtl8651_dns_cnameMapping_del(name2_p);
			goto new_group;
		}

		
		_rtl8651_dns_cnameMapping_refresh(name1_p);
		_rtl8651_dns_cnameMapping_refresh(name2_p);

	} else if (name1_p || name2_p)
	{
		if ((new_p = _rtl8651_dns_cnameMapping_getEntry()) == NULL)
		{
			goto out;
		}
		if (name1_p)
		{
			exist_p = name1_p;
			strncpy(new_p->addr, name2, RTL8651_DNS_DOMAIN_MAXLENGTH);
		} else
		{
			exist_p = name2_p;
			strncpy(new_p->addr, name1, RTL8651_DNS_DOMAIN_MAXLENGTH);
		}

		goto link_to_exist;

	} else
	{	
new_group:
		
		if ((new_p = _rtl8651_dns_cnameMapping_getEntry()) == NULL)
		{
			goto out;
		}
		if ((exist_p = _rtl8651_dns_cnameMapping_getEntry()) == NULL)
		{
			_rtl8651_dns_cnameMapping_del(new_p);
			goto out;
		}

		strncpy(exist_p->addr, name1, RTL8651_DNS_DOMAIN_MAXLENGTH);

		strncpy(new_p->addr, name2, RTL8651_DNS_DOMAIN_MAXLENGTH);

		
		if (_rtl8651_dns_cnameMap_p)
		{
			_rtl8651_dns_cnameMap_p->hdr_prev = exist_p;
			exist_p->hdr_next = _rtl8651_dns_cnameMap_p;
		}
		_rtl8651_dns_cnameMap_p = exist_p;

link_to_exist:
		
		if (exist_p->next)
		{
			exist_p->next->prev =new_p;
		}
		new_p->next = exist_p->next;
		new_p->prev = exist_p;
		exist_p->next = new_p;

		_rtl8651_dns_cnameMapping_refresh(new_p);
		_rtl8651_dns_cnameMapping_refresh(exist_p);
	}

out:
	return SUCCESS;
}


static void _rtl8651_dns_cnameMapping_del(_rtl8651_dns_cnameMapping_t *delEntry)
{
	_rtl8651_dns_cnameMapping_t *first = NULL, *entry = delEntry;

proc_del:

	
	first = entry;
	while (first->prev)
	{
		first = first->prev;
	}
	if (first == entry)
	{
		first = entry->next;
	}

	
	if (entry->prev)
	{
		entry->prev->next = entry->next;
	}
	if (entry->next)
	{
		entry->next->prev = entry->prev;

		if (entry->hdr_next)
		{
			entry->next->hdr_next = entry->hdr_next;
			entry->hdr_next->hdr_prev = entry->next;
		}
		if (entry->hdr_prev)
		{
			entry->next->hdr_prev = entry->hdr_prev;
			entry->hdr_prev->hdr_next = entry->next;
		}
		if (_rtl8651_dns_cnameMap_p == entry)
		{
			_rtl8651_dns_cnameMap_p = entry->next;
		}
	} else
	{	
		if (entry->hdr_next)
		{
			entry->hdr_next->hdr_prev = entry->hdr_prev;
		}
		if (entry->hdr_prev)
		{
			entry->hdr_prev->hdr_next = entry->hdr_next;
		}
		if (_rtl8651_dns_cnameMap_p == entry)
		{
			_rtl8651_dns_cnameMap_p = entry->hdr_next;
		}
	}

	
	if (entry->lru_prev)
	{
		entry->lru_prev->lru_next = entry->lru_next;
	}
	if (entry->lru_next)
	{
		entry->lru_next->lru_prev = entry->lru_prev;
	}
	if (_rtl8651_dns_cnameLru_h == entry)
	{
		_rtl8651_dns_cnameLru_h = entry->lru_next;
	}
	if (_rtl8651_dns_cnameLru_t == entry)
	{
		_rtl8651_dns_cnameLru_t = entry->lru_prev;
	}

	memset(entry, 0, sizeof(_rtl8651_dns_cnameMapping_t));

	
	entry->next = _rtl8651_dns_cnameFree_p;
	_rtl8651_dns_cnameFree_p = entry;

	
	if (first)
	{
		if (first->next == NULL)	
		{
			entry = first;
			goto proc_del;		
		}
	}
}


static void _rtl8651_dns_cnameMapping_TimeUpdate(uint32 secPassed)
{
	while (	(_rtl8651_dns_cnameLru_h) &&
			((_rtl8651_dns_cnameLru_h->updateTime + _rtl8651_dns_cnameTimeout) < _sysUpSeconds))
	{
		_rtl8651_dns_cnameMapping_del(_rtl8651_dns_cnameLru_h);
	}
}
 
static int32 _RTL8651_DNS_GET_NAMELEN(uint8 *name, uint32 maxLen, uint32 *len_p)
{
	uint8 *ptr = name;
	uint32 len = 0, nameLenCheck = 0;

	while (1)
	{
		if ((*((uint8*)ptr) & 0xc0) == 0xc0)
		{	

			

			len += 2;			
			break;	
		} else
		{	
			uint32 added;

			

			added = (1 + (uint32)(*(uint8*)ptr));	

			len += added;

			if (added == 1)	
			{
				break;	
			}

			
			{	
				nameLenCheck += added;

				

				if (nameLenCheck > RTL8651_DNS_DOMAIN_MAXLENGTH)
				{
					RTL_DNS_ERR("Decoded Domain Length (%d) is larger than Domain name max length (%d), ignore.",
								nameLenCheck,
								RTL8651_DNS_DOMAIN_MAXLENGTH);
					return FAILED;
				}
			}

			if (len > maxLen)
			{
				RTL_DNS_WARN("Domain Name parse failed.");
				return FAILED;
			}

			ptr = (uint8*)(added + (uint32)ptr);	
		}

	}

	if (len_p)
	{
		*len_p = len;
	}

	return SUCCESS;
}


static int32 _RTL8651_DNS_GET_INFO(
	_rtl8651_dns_dnsDomain_t domainType,
	uint8 *outputPtr,
	uint32 maxOutputLen,
	uint8 *inputPtr,
	uint8 *dnsData_p,
	uint32 dnsDataLen)
{
	uint8 outTmp[RTL8651_DNS_DOMAIN_MAXLENGTH];
	uint8 *inPtr = inputPtr;
	uint8 *outPtr = outputPtr;
	uint32 outLen = 0;
	uint8 addDot = FALSE;

	
	outPtr = outTmp;
	addDot = FALSE;	

	while (1)
	{
		if ((*((uint8*)inPtr) & 0xc0) == 0xc0)
		{	
			uint16 offset;

			
			offset = (*(uint8*)inPtr)<<8 | (*(uint8*)((uint32)inPtr + 1));
			offset &= 0x3fff;	
			offset = ntohs(offset) + 0;

			if (offset > dnsDataLen)
			{
				RTL_DNS_WARN("offset : %d, dnsDataLen : %d\n", offset, dnsDataLen);
				goto err_out;
			}

			inPtr = (uint8*)((uint32)dnsData_p + offset);

		} else
		{	
			uint8 *origInPtr, *origOutPtr;
			uint8 labelLen;

			labelLen = *((uint8*)inPtr);

			PTR_CHK_SHIFT_EXIT(	(dnsDataLen + (uint32)dnsData_p - (uint32)inPtr),
									inPtr,
									sizeof(uint8),
									err_out);	

			
			if (labelLen == 0)
			{
				break;
			}

			if (addDot == TRUE)
			{
				
				origOutPtr = outPtr;
				PTR_CHK_SHIFT_EXIT(	(maxOutputLen - outLen),
										outPtr,
										sizeof(uint8),
										err_out);

				
				outLen += sizeof(uint8);
				
				*(char*)origOutPtr = '.';
			}

			
			origInPtr = inPtr;
			origOutPtr = outPtr;
			
			PTR_CHK_SHIFT_EXIT(	(dnsDataLen + (uint32)dnsData_p - (uint32)inPtr),
									inPtr,
									labelLen,
									err_out);
			PTR_CHK_SHIFT_EXIT(	(maxOutputLen - outLen),
									outPtr,
									labelLen,
									err_out);

			
			outLen += labelLen;

			
			memcpy(origOutPtr, origInPtr, labelLen);

			addDot = TRUE;	

		}
	}

	
	{
		uint8 *origOutPtr = outPtr;

		PTR_CHK_SHIFT_EXIT(	(maxOutputLen - outLen),
								outPtr,
								sizeof(uint8),
								err_out);

		
		outLen += sizeof(uint8);
		*(uint8*)origOutPtr = 0;
	}

	
	if (domainType == _RTL8651_DNS_GET_TYPE_INET_IPV4)
	{
		
		uint32 separator[16];
		uint8 digitBeforeSeparator[16];
		uint32 separatorCnt, sepIdx;
		uint32 digitLen;
		uint8 *ptr;

		uint8 *actOutputPtr = outputPtr;
		uint32 remOutLen = maxOutputLen;
		uint8 *origOutPtr;

		
		ptr = outTmp;
		digitLen = 0;
		memset(separator, 0, sizeof(separator));
		memset(digitBeforeSeparator, 0 ,sizeof(digitBeforeSeparator));
		separatorCnt = 0;

		while (1)
		{
			if (	(*(uint8*)ptr == 0) ||
				(digitLen > outLen))
			{
				RTL_DNS_WARN("ptr : %p, digitLen %d, outLen %d", ptr, digitLen, outLen);
				goto err_out;
			}

			if (	(!isdigit(*(char*)ptr)) &&
				(*(char*)ptr != '.'))
			{
				if (	(digitLen == 0)||
					(*(char*)(ptr-1) != '.'))
				{
					
					RTL_DNS_WARN("digitLen : %d, char (ptr - 1) : %c -- Error string.", digitLen, *(char*)(ptr-1));
					goto err_out;
				}
				assert(*(char*)ptr == 'i');	
				break;
			}

			if (*(char*)ptr == '.')
			{
				if (separatorCnt >= (sizeof(separator) / sizeof(separator[0])))
				{
					RTL_DNS_WARN("separatorCount: %d , size of separator: %d , sizeof separator[0]: %d",
									separatorCnt,
									sizeof(separator),
									sizeof(separator[0]));
					goto err_out;
				}

				separator[separatorCnt] = digitLen;

				if (separatorCnt == 0)
				{
					digitBeforeSeparator[separatorCnt] = digitLen;
				} else
				{
					digitBeforeSeparator[separatorCnt] = digitLen - separator[separatorCnt - 1] - 1;
				}

				separatorCnt ++;
			}

			ptr ++;
			digitLen ++;
		}

		

		

		sepIdx = separatorCnt;
		
		while (sepIdx > 0)
		{
			uint8 *numPtr = &(outTmp[(separator[sepIdx - 1] + 1)]);

			if (numPtr == ptr)
			{	
				goto nextlabel;
			}

			if (	(sepIdx >= (sizeof(digitBeforeSeparator) - sizeof(digitBeforeSeparator[0]))) ||
				(digitBeforeSeparator[sepIdx] == 0))
			{	
				goto nextlabel;
			}

			
			origOutPtr = actOutputPtr;

			PTR_SHIFT_EXIT(	remOutLen,
								actOutputPtr,
								digitBeforeSeparator[sepIdx],
								err_out);

			memcpy(origOutPtr, numPtr, digitBeforeSeparator[sepIdx]);

			
			origOutPtr = actOutputPtr;

			PTR_SHIFT_EXIT(	remOutLen,
								actOutputPtr,
								1,
								err_out);

			*(char*)origOutPtr = '.';

		nextlabel:
			sepIdx --;
		}

		
		if (digitBeforeSeparator[0] > 0)
		{
			origOutPtr = actOutputPtr;

			PTR_SHIFT_EXIT(	remOutLen,
								actOutputPtr,
								digitBeforeSeparator[0],
								err_out);

			memcpy(origOutPtr, outTmp, digitBeforeSeparator[0]);
		}

		
		{
			origOutPtr = actOutputPtr;

			PTR_CHK_SHIFT_EXIT(	remOutLen,
									actOutputPtr,
									sizeof(uint8),
									err_out);

			
			outLen += sizeof(uint8);
			*(uint8*)origOutPtr = 0;
		}

	}else	
	{	
		__strlowerncpy(outputPtr, outTmp, outLen);
	}

	RTL_DNS_INFO("Get string [%s]", outputPtr);

	return SUCCESS;

err_out:

	RTL_DNS_WARN("process failed");
	
	return FAILED;
}





static int32 _rtl8651_dns_parseDnsPkt(
	uint8 *dns_p,
	uint32 dnsLen,
	uint8 *pktModified,
	int32 (*dns_proc)(char *, ipaddr_t))
{
	uint8 *ptr = NULL;
	uint32 idx = 0, remLen = 0;
	_rtl8651_dns_dnsHeader_t *dnsHdr_p = NULL;
	uint16 qdcount = 0, ancount = 0, nscount = 0, arcount = 0;

	RTL_DNS_INFO("Parse DNS message (size %d)\n", dnsLen);
	RTL_DNS_INFO("*=======================================*");

	if (	(dns_p == NULL) ||
		(dns_proc == NULL) ||
		(dnsLen < sizeof(_rtl8651_dns_dnsHeader_t)))
	{
		goto err_out;
	}

	ptr = dns_p;
	remLen = dnsLen;

	


	
	dnsHdr_p = (_rtl8651_dns_dnsHeader_t*)ptr;

	RTL_DNS_INFO("");
	RTL_DNS_INFO("Header:");
	RTL_DNS_INFO("");
	RTL_DNS_INFO("\tID:\t\t%d", ntohs(dnsHdr_p->id));
	RTL_DNS_INFO("\tQR:\t\t%s", (dnsHdr_p->qr == 0)?"QUERY":"RESPONSE");
	RTL_DNS_INFO("\tOPCODE:\t\t%d", dnsHdr_p->opcode);
	RTL_DNS_INFO("\tAA:\t\t%d", dnsHdr_p->aa);
	RTL_DNS_INFO("\tTC:\t\t%d", dnsHdr_p->tc);
	RTL_DNS_INFO("\tRD:\t\t%d", dnsHdr_p->rd);
	RTL_DNS_INFO("\tRA:\t\t%d", dnsHdr_p->ra);
	RTL_DNS_INFO("\tRCODE:\t\t%d", dnsHdr_p->rcode);
	RTL_DNS_INFO("\tQDCOUNT:\t%d", ntohs(dnsHdr_p->qdcount));
	RTL_DNS_INFO("\tANCOUNT:\t%d", ntohs(dnsHdr_p->ancount));
	RTL_DNS_INFO("\tQNSOUNT:\t%d", ntohs(dnsHdr_p->nscount));
	RTL_DNS_INFO("\tQAROUNT:\t%d", ntohs(dnsHdr_p->arcount));

	

	
	if (dnsHdr_p->qr != _RTL8651_DNSHDR_QR_RESPONSE)
	{
		goto ok_out;
	}

	
	if (dnsHdr_p->opcode != _RTL8651_DNSHDR_OPCODE_QUERY)
	{
		RTL_DNS_INFO("opcode = %d", dnsHdr_p->opcode);
	}

	
	if (dnsHdr_p->tc == 1)
	{
		RTL_DNS_WARN("TC == 1, this DNS message is truncated");
		goto ok_out;
	}

	
	if (dnsHdr_p->opcode == _RTL8651_DNSHDR_RCODE_OK)
	{
		
	} else
	{
		if ((dnsHdr_p->ancount + dnsHdr_p->nscount + dnsHdr_p->arcount) > 0)
		{	
			RTL_DNS_WARN("Response with opcode (%d): ignore information -- ANCNT(%d) NSCNT(%d) ARCNT(%d)",
								dnsHdr_p->opcode,
								dnsHdr_p->ancount,
								dnsHdr_p->nscount,
								dnsHdr_p->arcount);
		}
		goto ok_out;
	}

	
	qdcount = dnsHdr_p->qdcount;
	ancount = dnsHdr_p->ancount;
	nscount = dnsHdr_p->nscount;
	arcount = dnsHdr_p->arcount;

	PTR_SHIFT_EXIT(remLen, ptr, sizeof(_rtl8651_dns_dnsHeader_t),  err_out);

	

	
	for ( idx = 0 ; idx < qdcount ; idx ++ )
	{
		uint32 nameLen;

		
		if (_RTL8651_DNS_GET_NAMELEN(ptr, remLen, &nameLen) != SUCCESS)
		{
			goto err_out;
		}

#ifdef RTL_DNS_DEBUG
		RTL_DNS_INFO("");
		RTL_DNS_INFO("QUERY:");
		RTL_DNS_INFO("");
		{
			uint16 qtype, qclass;

			qtype = PKTGET_UINT16_UNALIGNED((uint16*)((uint32)ptr+nameLen));
			qclass = PKTGET_UINT16_UNALIGNED((uint16*)((uint32)ptr+nameLen+sizeof(uint16)));

			RTL_DNS_INFO("\tQNAME Length:\t%d", nameLen);
			RTL_DNS_INFO("\tQTYPE:\t\t%d", qtype);
			RTL_DNS_INFO("\tQCLASS:\t\t%d", qclass);
		}
#endif
		PTR_SHIFT_EXIT(	remLen,
							ptr,
							(nameLen  + sizeof(uint16)  + sizeof(uint16) ),
							err_out);
	}

	
	for ( idx = 0 ; idx < (ancount + nscount + arcount) ; idx ++ )
	{
		uint32 rrLen = 0;

		if (_rtl8651_dns_parseDnsRR(dns_p, dnsLen, ptr, remLen, &rrLen, pktModified, dns_proc) != SUCCESS)
		{
			goto err_out;
		}

		PTR_SHIFT_EXIT(remLen, ptr, rrLen,  err_out);
	}

ok_out:

	RTL_DNS_INFO("*=======================================*");
	return SUCCESS;
err_out:

	RTL_DNS_INFO("*=======================================*");
	return FAILED;
}


static int32 _rtl8651_dns_parseDnsRR(
	uint8 *dns_p, uint32 totalLen,	
	uint8 *rr_p,					
	uint32 maxLen,				
	uint32 *rrLen_p,
	uint8 *pktModified,
	int32 (*dns_proc)(char *, ipaddr_t)
	)
{
	
	_rtl8651_dns_dnsRR_fixedField_t fixFieldBody, *rrFixField_p = &fixFieldBody;
	uint32 nameLen = 0, rrLen = 0;
	uint8 *namePtr = NULL;

	char mapped_domain[RTL8651_DNS_DOMAIN_MAXLENGTH];
	char mapped_cname[RTL8651_DNS_DOMAIN_MAXLENGTH];		
	ipaddr_t mapped_ip = 0;

	
	DNS_UNLOCK_ERROR_CHECK_JMP(	_RTL8651_DNS_GET_NAMELEN(rr_p, maxLen, &nameLen),
									SUCCESS,
									err_out); 
	namePtr = rr_p;

	

	DNS_UNLOCK_COND_JMP(maxLen < (nameLen + 10), err_out);

	rtl8651_memcpy(rrFixField_p, (void*)((uint32)rr_p + nameLen), 10);

	
	rrLen =	nameLen +					
			10 +							
			ntohs(rrFixField_p->rdlength);	

	DNS_UNLOCK_COND_JMP((rrLen > maxLen), err_out);

	RTL_DNS_INFO("");
	RTL_DNS_INFO("RR field:");
	RTL_DNS_INFO("");
	RTL_DNS_INFO("\tNAME Length:\t%d", nameLen);
	RTL_DNS_INFO("\tTYPE:\t\t%d", ntohs(rrFixField_p->type));
	RTL_DNS_INFO("\tCLASS:\t\t%d", ntohs(rrFixField_p->class));
	RTL_DNS_INFO("\tTTL:\t\t%d", ntohl(rrFixField_p->ttl));
	RTL_DNS_INFO("\tRDLEN:\t\t%d", ntohs(rrFixField_p->rdlength));

	
	switch (rrFixField_p->type)
	{
		case _RTL8651_DNSTYPE_A:
			
		case _RTL8651_DNSTYPE_CNAME:
		case _RTL8651_DNSTYPE_PTR:
			break;
		default:							
			goto ok_out;
	}

	
	switch (rrFixField_p->class)
	{
		case _RTL8651_DNSCLASS_IN:
			break;
		default:						
			goto ok_out;
	}

	

	
	memset(mapped_domain, 0, RTL8651_DNS_DOMAIN_MAXLENGTH);
	memset(mapped_cname, 0, RTL8651_DNS_DOMAIN_MAXLENGTH);
	mapped_ip = 0;

	switch (rrFixField_p->type)
	{
		case _RTL8651_DNSTYPE_A:
			
			rtl8651_memcpy(&mapped_ip, (void*)((uint32)namePtr + sizeof(_rtl8651_dns_dnsRR_fixedField_t)), sizeof(ipaddr_t));
			if (_RTL8651_DNS_GET_INFO(	_RTL8651_DNS_GET_TYPE_DOMAIN,
										mapped_domain,
										RTL8651_DNS_DOMAIN_MAXLENGTH,
										namePtr,
										dns_p, totalLen) != SUCCESS)	
			{
				goto err_out;
			}

			if (_rtl8651_dns_procDnsCallBack(mapped_ip, mapped_domain, dns_proc) != SUCCESS)
			{
				goto err_out;
			}

			break;
		case _RTL8651_DNSTYPE_PTR:
			
			if (	(_RTL8651_DNS_GET_INFO(	_RTL8651_DNS_GET_TYPE_INET_IPV4,
											mapped_cname,
											RTL8651_DNS_DOMAIN_MAXLENGTH,
											namePtr,
											dns_p, totalLen) != SUCCESS) ||
				(_RTL8651_DNS_GET_INFO(	_RTL8651_DNS_GET_TYPE_DOMAIN,
											mapped_domain,
											RTL8651_DNS_DOMAIN_MAXLENGTH,
											(void*)((uint32)namePtr + sizeof(_rtl8651_dns_dnsRR_fixedField_t)),
											dns_p, totalLen) != SUCCESS)	
				)
			{
				goto err_out;
			}

			
			mapped_ip = __inet_addr(mapped_cname);
			if (_rtl8651_dns_procDnsCallBack(mapped_ip, mapped_domain, dns_proc) != SUCCESS)
			{
				goto err_out;
			}

			break;
		case _RTL8651_DNSTYPE_CNAME:
			
			if (	(_RTL8651_DNS_GET_INFO(	_RTL8651_DNS_GET_TYPE_DOMAIN,
											mapped_domain,
											RTL8651_DNS_DOMAIN_MAXLENGTH,
											namePtr,
											dns_p, totalLen) != SUCCESS) ||
				(_RTL8651_DNS_GET_INFO(	_RTL8651_DNS_GET_TYPE_DOMAIN,
											mapped_cname,
											RTL8651_DNS_DOMAIN_MAXLENGTH,
											(void*)((uint32)namePtr + sizeof(_rtl8651_dns_dnsRR_fixedField_t)),
											dns_p, totalLen) != SUCCESS)	)
			{
				goto err_out;
			}

			if (_rtl8651_dns_cnameMapping_add(mapped_domain, mapped_cname) != SUCCESS)
			{
				assert(0);
				goto err_out;
			}

			break;

		default:
			goto ok_out;
	}

	
	if (	(_rtl8651_dns_enableTTLSuppression == TRUE) &&
		(rrFixField_p->ttl > _rtl8651_dns_suppressedTTL)	)
	{
		_rtl8651_dns_dnsRR_fixedField_t *tmp;
		tmp = (_rtl8651_dns_dnsRR_fixedField_t *)((uint32)rr_p + nameLen);
		PKTSET_UINT32_UNALIGNED(_rtl8651_dns_suppressedTTL, &(tmp->ttl));

		assert(pktModified);
		*pktModified = TRUE;		
	}
ok_out:

	
	assert(rrLen);

	if (rrLen_p)
	{
		*rrLen_p = rrLen;
	}

	return SUCCESS;
err_out:
	return FAILED;
}


static int32 _rtl8651_dns_procDnsCallBack(
	ipaddr_t mapped_ip,
	char *mapped_domain,
	int32 (*dns_proc)(char *, ipaddr_t)
	)
{
	int32 retval;
	_rtl8651_dns_cnameMapping_t *ptr;

	
	if (mapped_ip == 0)		
	{
		RTL_DNS_INFO("mapping ignored: [%s]:(%u.%u.%u.%u).", mapped_domain, NIPQUAD(mapped_ip));
		return SUCCESS;
	}

	
	if ((retval = dns_proc(mapped_domain, mapped_ip)) != SUCCESS)
	{
		return retval;
	}
	
	ptr = _rtl8651_dns_cnameMapping_find(mapped_domain);

	if (ptr)
	{
		_rtl8651_dns_cnameMapping_t *trace = ptr;

		
		while (trace->prev)
		{
			trace = trace->prev;
		}

		while (trace)
		{
			if (trace != ptr)	
			{
				if ((retval = dns_proc(trace->addr, mapped_ip)) != SUCCESS)
				{
					return retval;
				}
			}
			trace = trace->next;
		}
	}

	return SUCCESS;
}



#define RTL8651_DNS_TBLDRV_GLUE

#define SET_DNSACL_MATCHLAYER(rule) \
	rule.pktOpApp = RTL8651_ACLTBL_L3_AND_L4;


uint32 _rtl8651_tryToAddAclRuleForDns(uint32 aclStart, uint32 aclEnd, int8 *enoughFlag)
{
	int32 idx, maxIdx;
	_rtl8651_tblDrvAclRule_t rule;
	uint32 addCount = 0, aclIdx = aclStart;
	uint32 trapDns = FALSE;

	
	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx])
		{
			trapDns = TRUE;
			break;
		}
	}

	
	if (trapDns == FALSE)
	{
		goto out;
	}

#ifdef ENABLE_TCP_DNS
	maxIdx = 2;
#else
	maxIdx = 1;
#endif

	for (idx = 0 ; idx < maxIdx ; idx ++)
	{
		if (aclIdx > aclEnd) {
			*enoughFlag = 0;	
			goto out;
		}

		bzero(&rule, sizeof(_rtl8651_tblDrvAclRule_t));

		rule.actionType_				= RTL8651_ACL_CPU;

		if ( idx == 0 )
		{
			
			rule.ruleType_				= RTL8651_ACL_UDP;
			rule.udpSrcPortUB_		= 53;
			rule.udpSrcPortLB_			= 53;
			rule.udpDstPortUB_		= 0xffff;
			rule.udpDstPortLB_			= 0;
		} else
		{
#ifdef ENABLE_TCP_DNS
			
			rule.ruleType_				= RTL8651_ACL_TCP;
			rule.tcpSrcPortUB_			= 53;
			rule.tcpSrcPortLB_			= 53;
			rule.tcpDstPortUB_			= 0xffff;
			rule.tcpDstPortLB_			= 0;
#endif
		}

		SET_DNSACL_MATCHLAYER(rule);
		rtl8651_setAsicAclRule(aclIdx, &rule);
		_rtl8651_registerACLDB( aclIdx, _rtl8651_dns_procDnsPkt, NULL );

		addCount ++;
		aclIdx ++;
	}

out:
	return addCount;
}


int32 _rtl8651_dns_init(rtl8651_fwdEngineInitPara_t *para)
{
	RTL_DNS_INFO("Init DNS analysis module of ROMEDRV.");

	DNS_UNLOCK_ERROR_CHECK(_rtl8651_dns_dataBase_init(para), SUCCESS, TRUE, FAILED);
	DNS_UNLOCK_ERROR_CHECK(_rtl8651_dns_cnameMapping_init(para), SUCCESS, TRUE, FAILED);

	return SUCCESS;
}


void _rtl8651_dns_reinit(void)
{
	int32 idx;

	RTL_DNS_INFO("Re-init DNS analysis module of ROMEDRV.");

	
	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.callBack[idx].reinit)
		{
			_rtl8651_dns_ctrlBlock.callBack[idx].reinit();
		}
	}

	_rtl8651_dns_dataBase_reinit();
	_rtl8651_dns_cnameMapping_reinit();
}

void _rtl8651_dns_timeUpdate(uint32 secPassed)
{
	int32 idx;

	
	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.callBack[idx].timeUpdate)
		{
			_rtl8651_dns_ctrlBlock.callBack[idx].timeUpdate(secPassed);
		}
	}

	_rtl8651_dns_cnameMapping_TimeUpdate(secPassed);
}



#define RTL8651_DNS_FWDENG_GLUE


int32 _rtl8651_dns_procDnsPkt(int32 ruleNo, struct rtl_pktHdr *pktHdr, struct ip *pip, void *usrDefined)
{
	uint8 *ptr;
	uint32 dataLen;
	uint8 pktModified = FALSE;
	struct tcphdr *tc = NULL;
	struct udphdr *ud = NULL;

	RTL_DNS_INFO("Proc DNS packet trapped by ACL rule (%d)", ruleNo);

	if (IS_FRAGMENT_REMAIN(pip->ip_off))
	{
		goto out;
	}

	switch (pip->ip_p)
	{
		case IPPROTO_TCP:
		{
			tc = (struct tcphdr *)L4HDR_ADDR(pip);
			ptr = (uint8*)TCPDATA_ADDR(tc);
			break;
		}
		case IPPROTO_UDP:
		{
			ud = (struct udphdr *)L4HDR_ADDR(pip);
			ptr = (uint8*)((uint32)ud + sizeof(struct udphdr));
			break;
		}
		default:
			goto out;
	}

	
	dataLen = ntohs(pip->ip_len) - ((uint32)ptr - (uint32)pip);

	if (_rtl8651_dns_parseDnsPkt(	ptr,
									dataLen,
									&pktModified,
									_rtl8651_dns_funcCallBack) != SUCCESS)
	{
		RTL_DNS_WARN("_rtl8651_dns_parseDnsPkt return FAILED.");
	}

	if (pktModified == TRUE)
	{
		switch(pip->ip_p)
		{
			case IPPROTO_TCP:
				tc->th_sum = 0;
				tc->th_sum = _rtl8651_tcpChecksum(pip);
				break;
			case IPPROTO_UDP:
				if (ud->uh_sum)
				{
					ud->uh_sum = 0;
					ud->uh_sum = _rtl8651_tcpChecksum(pip);
				}
				break;
		}
	}

out:
	return FAILED;	
}

