

#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_tblAsicDrv.h"
#include "rtl8651_jhash.h"
#include "mbuf.h"

#include "rtl8651_dos.h"
#include "rtl8651_dos_local.h"



typedef struct _rtl8651_dosProc_blockIp_s {
	ipaddr_t							ip;
	uint16							vid;
	uint32							dischargeTime;

	
	
	struct _rtl8651_dosProc_blockIp_s	*next;
	struct _rtl8651_dosProc_blockIp_s	**hdr;
	
	struct _rtl8651_dosProc_blockIp_s	*jail_next;
} _rtl8651_dosProc_blockIp_t;

static uint8 _rtl8651_dosProc_sipBlock_sw[RTL8651_MAX_DIALSESSION] = {FALSE};
static uint8 _rtl8651_dosProc_sipBlock_enable = FALSE;

#define RTL8651_DOS_PROC_BLOCKIP_HASHTBL_SIZE	16
static _rtl8651_dosProc_blockIp_t *sipBlockHash[RTL8651_DOS_PROC_BLOCKIP_HASHTBL_SIZE];

static uint32 _rtl8651_dos_sipBlock_sw[RTL8651_MAX_DIALSESSION] = {0};

static uint32 _rtl8651_dos_sipBlock_prisonTerm = RTL8651_DOSPROC_SIPBLOCK_TIMEOUT;

_rtl8651_dosProc_blockIp_t *inuseJail = NULL;
_rtl8651_dosProc_blockIp_t *emptyJail = NULL;

static uint32 blockVlanIpCnt[RTL8651_VLAN_NUMBER];	

uint32 rtl8651_dosSIPBlockSw(uint32, uint32, uint32 *);

int32 _rtl8651_dosProc_blockSip_init(rtl8651_fwdEngineInitPara_t *);
void _rtl8651_dosProc_blockSip_reinit(void);
static __inline__ void _rtl8651_dosProc_blockSip(uint32, uint32, uint16, struct ip*);
static __inline__ void _rtl8651_dosProc_blockIP(uint32, uint32, uint16, ipaddr_t);

static __inline__ uint32 _rtl8651_dosProc_blockSip_hash(ipaddr_t);
static __inline__ void _rtl8651_dosProc_blockSip_imprison(uint16, ipaddr_t, uint32);

#define AttackBlockSip (\
		RTL8651_DOS_SCAN |\
		RTL8651_DOS_PINGOFDEATH |\
		RTL8651_DOS_TEARDROP |\
		RTL8651_DOS_UDPBOMB |\
		RTL8651_DOS_ECHOCHARGEN |\
		RTL8651_DOS_SYNWITHDATA |\
		RTL8651_DOS_SRC_SYNFLOOD |\
		RTL8651_DOS_SRC_STEALTHFIN |\
		RTL8651_DOS_SRC_ICMPFLOOD |\
		RTL8651_DOS_SRC_UDPFLOOD |\
		RTL8651_DOS_SRC_TCPCONNOVERFLOW |\
		RTL8651_DOS_SRC_UDPCONNOVERFLOW |\
		RTL8651_DOS_SRC_TCPUDPCONNOVERFLOW |\
		RTL8651_DOS_PORTSCAN)

#define CHECK_SIPBLOCK_OFF_CONTINUE(dsid) \
			if (_rtl8651_dosProc_sipBlock_sw[dsid] == FALSE) \
			{ \
				continue; \
			}
	
#define CHECK_SIPBLOCK_OFF_RETURN_VOID(dsid) \
		do { \
			if (_rtl8651_dosProc_sipBlock_sw[dsid] == FALSE) \
				return; \
		} while (0);

#define CHECK_SIPBLOCK_DISABLE_RETURN_VOID() \
		do { \
			if (_rtl8651_dosProc_sipBlock_enable == FALSE) \
				return;\
		} while (0);

#define CHECK_SIPBLOCK_DISABLE(ret) \
		do { \
			if (_rtl8651_dosProc_sipBlock_enable == FALSE) \
				return ret ;\
		} while (0);

int32 _rtl8651_procDoSPkt_init(rtl8651_fwdEngineInitPara_t *para)
{
	int32 retval = SUCCESS;

	retval |= _rtl8651_dosProc_blockSip_init(para);

	return retval;
}
void _rtl8651_procDosPkt_reinit()
{
	_rtl8651_dosProc_blockSip_reinit();
}
void _rtl8651_procDoSPkt(uint32 item, uint32 dsid, uint16 vid, struct ip*ip, uint8 direction, struct rtl_pktHdr* pktHdr)
{
	_rtl8651_dosProc_blockSip(item, dsid, vid, ip);
}
void _rtl8651_procDoSIP(uint32 item, uint32 dsid, uint16 vid, ipaddr_t ip)
{
	_rtl8651_dosProc_blockIP(item, dsid, vid, ip);
}

int32 _rtl8651_dosProc_blockSip_init(rtl8651_fwdEngineInitPara_t *para)
{
	int32 idx;
	uint32 JailCnt;
	_rtl8651_dosProc_blockIp_t *ptr;

	
	for (idx = 0 ; idx < RTL8651_MAX_DIALSESSION ; idx ++)
		_rtl8651_dosProc_sipBlock_sw[idx] = FALSE;

	if (!para)
		goto errout;

	JailCnt = para->dosProc_sipblock_count;

	if (JailCnt == 0)
		goto disableOut;

	
	emptyJail = (_rtl8651_dosProc_blockIp_t *)rtlglue_malloc(JailCnt*sizeof(_rtl8651_dosProc_blockIp_t));
	if (!emptyJail)
		goto errout;

	inuseJail = NULL;

	
	memset(emptyJail, 0, (JailCnt*sizeof(_rtl8651_dosProc_blockIp_t)));

	
	ptr = emptyJail;
	for (idx = 0 ; idx < JailCnt ; idx ++, ptr ++)
	{
		if (idx != (JailCnt - 1))
		{
			ptr->jail_next = ptr + 1;
		}
	}

	
	for (idx = 0 ; idx < RTL8651_DOS_PROC_BLOCKIP_HASHTBL_SIZE ; idx ++)
		sipBlockHash[idx] = NULL;

	
	for (idx = 0 ; idx < RTL8651_VLAN_NUMBER ; idx ++)
		blockVlanIpCnt[idx] = 0;

	
	for (idx = 0 ; idx < RTL8651_MAX_DIALSESSION ; idx ++)
		_rtl8651_dos_sipBlock_sw[idx] = AttackBlockSip;

	_rtl8651_dosProc_sipBlock_enable = TRUE;
	return SUCCESS;
errout:
	_rtl8651_dosProc_sipBlock_enable = FALSE;
	return FAILED;
disableOut:
	_rtl8651_dosProc_sipBlock_enable = FALSE;
	return SUCCESS;
}
void _rtl8651_dosProc_blockSip_reinit()
{
	CHECK_SIPBLOCK_DISABLE_RETURN_VOID();

	
	rtl8651_dosProc_blockSip_freeAll();
}

static __inline__ uint32 _rtl8651_dosProc_blockSip_hash(ipaddr_t ip)
{
	return (ip & (RTL8651_DOS_PROC_BLOCKIP_HASHTBL_SIZE-1));
}

int32 _rtl8651_dosProc_blockSip_check(uint16 vid, ipaddr_t sip)
{
	_rtl8651_dosProc_blockIp_t *ptr;
	uint32 hash = _rtl8651_dosProc_blockSip_hash(sip);

	CHECK_SIPBLOCK_DISABLE(FAILED);

	ptr = sipBlockHash[hash];

	while (ptr)
	{
		if ((ptr->ip == sip) && (ptr->vid == vid))
		{	
			if (ptr->dischargeTime < _sysUpSeconds)
			{	
				_rtl8651_dosProc_blockSip_discharge(_sysUpSeconds);
				goto not_found;
			}
			return SUCCESS;
		}
		ptr = ptr->next;
	}
not_found:
	return FAILED;
}

static __inline__ void _rtl8651_dosProc_blockSip_imprison(uint16 vid, ipaddr_t sip, uint32 prisonTerm)
{
	_rtl8651_dosProc_blockIp_t *prisoner;
	_rtl8651_dosProc_blockIp_t *ptr, *prev;
	uint32 hash = _rtl8651_dosProc_blockSip_hash(sip);

	
	if (!emptyJail)
		return;
	prisoner = emptyJail;
	emptyJail = emptyJail->jail_next;

	
	
	_rtl8651_removeNaptConnectionByIP(sip);
	
	assert(_rtl8651_connSrcTrackingRemoveByIp(sip) == 0);

	
	prisoner->ip				= sip;
	prisoner->vid				= vid;
	prisoner->dischargeTime	= _sysUpSeconds + prisonTerm;

	
	
	
	ptr = inuseJail;
	prev = NULL;
	while (ptr && ((ptr->dischargeTime) < (prisoner->dischargeTime)))
	{
		prev = ptr;
		ptr = ptr->jail_next;
	}
	
	prisoner->jail_next = ptr;
	if (prev)
	{
		prev->jail_next = prisoner;
	}else
	{
		inuseJail = prisoner;
	}
	
	prisoner->next = sipBlockHash[hash];
	prisoner->hdr = &sipBlockHash[hash];
	sipBlockHash[hash] = prisoner;

	
	blockVlanIpCnt[vid&(RTL8651_VLAN_NUMBER-1)] ++;

	
	_rtl8651_arrangeAllACL();
}

void _rtl8651_dosProc_blockSip_discharge(uint32 currentTime)
{
	_rtl8651_dosProc_blockIp_t *ptr, *prev, *hashptr;
	int32 AclArrange = FALSE;

	
	while (inuseJail && (inuseJail->dischargeTime < currentTime))
	{
		AclArrange = TRUE;
	
		
		blockVlanIpCnt[(inuseJail->vid)&(RTL8651_VLAN_NUMBER-1)] --;

		
		ptr = inuseJail;
		inuseJail = inuseJail->jail_next;

		
		if (*(ptr->hdr) == ptr)
		{	
			*(ptr->hdr) = ptr->next;
		}else
		{
			hashptr = *(ptr->hdr);
			prev = NULL;
			while (hashptr)
			{
				if (hashptr == ptr)
				{
					assert(prev);		
					if (prev)
						prev->next = hashptr->next;
					goto jailGiveBack;
				}
				prev = hashptr;
				hashptr = hashptr->next;
			}
			assert(hashptr);	
		}

jailGiveBack:
		
		ptr->jail_next = emptyJail;
		emptyJail = ptr;
	}

	if (AclArrange == TRUE)
		_rtl8651_arrangeAllACL();
}

static __inline__ void _rtl8651_dosProc_blockSip(uint32 item, uint32 dsid, uint16 vid, struct ip*ip)
{

	CHECK_SIPBLOCK_DISABLE_RETURN_VOID();
	CHECK_SIPBLOCK_OFF_RETURN_VOID(dsid);

	if (item & _rtl8651_dos_sipBlock_sw[dsid])
	{
		ipaddr_t sip = PKTGET_UINT32_UNALIGNED(&ip->ip_src);
	
		_rtl8651_dosProc_blockSip_imprison(vid, sip, _rtl8651_dos_sipBlock_prisonTerm);
	}
}

static __inline__ void _rtl8651_dosProc_blockIP(uint32 item, uint32 dsid, uint16 vid, ipaddr_t ip)
{
	CHECK_SIPBLOCK_DISABLE_RETURN_VOID();

	
	if ( dsid == RTL8651_DSID_ALL )
	{
		int idx;
		for ( idx = 0 ; idx < RTL8651_MAX_DIALSESSION ; idx ++ )
		{
			CHECK_SIPBLOCK_OFF_CONTINUE(idx);

			if (item & _rtl8651_dos_sipBlock_sw[idx])
			{
				goto imprison;
			}
		}
	} else
	{
		CHECK_SIPBLOCK_OFF_RETURN_VOID(dsid);

		if (item & _rtl8651_dos_sipBlock_sw[dsid])
		{
			goto imprison;
		}
	}

	return;

imprison:
	_rtl8651_dosProc_blockSip_imprison(vid, ip, _rtl8651_dos_sipBlock_prisonTerm);
}


#define RTL8651_IP_ACL_RULE(ipRule, action, sip, sipm, dip, dipm, iptos, iptosm, iproto, iprotom, ipflag, ipflagm)\
{\
	(ipRule)->srcIpAddr_ = sip;\
	(ipRule)->srcIpAddrMask_ = sipm;\
	(ipRule)->dstIpAddr_ = dip;\
	(ipRule)->dstIpAddrMask_ = dipm;\
	(ipRule)->tos_ = iptos;\
	(ipRule)->tosMask_ = iptosm;\
	(ipRule)->ipProto_ = iproto;\
	(ipRule)->ipProtoMask_ = iprotom;\
	(ipRule)->ipFlag_ = ipflag;\
	(ipRule)->ipFlagMask_ = ipflagm;\
	(ipRule)->ruleType_ = RTL8651_ACL_IP;\
	(ipRule)->actionType_ = action;\
}

uint32 _rtl8651_addAclRuleForSipBlock(uint16 vid, uint32 aclStart, uint32 aclEnd, int8 *enoughFlag)
{
	_rtl8651_dosProc_blockIp_t *ptr;
	_rtl8651_tblDrvAclRule_t rule;
	uint32 aclIdx;
	uint32 addCnt;

	CHECK_SIPBLOCK_DISABLE(0);

	if (blockVlanIpCnt[vid&(RTL8651_VLAN_NUMBER-1)] == 0)
		return 0;	

	
	aclIdx = aclStart;
	addCnt = 0;
	ptr = inuseJail;
	while (ptr)
	{
		if (ptr->vid == vid)
		{
			if (aclIdx > aclEnd) {
				enoughFlag = 0;	
				goto out;
			}
			memset(&rule, 0, sizeof(_rtl8651_tblDrvAclRule_t));
			RTL8651_IP_ACL_RULE(&rule, RTL8651_ACL_DROP, ptr->ip, 0xffffffff, 0, 0, 0, 0, 0, 0, 0, 0);
			rtl8651_setAsicAclRule(aclIdx, &rule);
			addCnt ++;
			aclIdx ++;
		}
		ptr = ptr->jail_next;
	}
out:
	return addCnt;
}



uint32 rtl8651_dosSIPBlockSw(uint32 item, uint32 dsid, uint32 *flag)
{
	assert(dsid < rtl8651_tblDrvPara.dialSessionNumber);

	rtlglue_drvMutexLock();

	if (!flag)		
	{
		rtlglue_drvMutexUnlock();
		return (item & _rtl8651_dos_sipBlock_sw[dsid]);
	}else
	{	
		if (*flag)
		{	
			_rtl8651_dos_sipBlock_sw[dsid] = (_rtl8651_dos_sipBlock_sw[dsid] | item);
		}else
		{
			_rtl8651_dos_sipBlock_sw[dsid] = (_rtl8651_dos_sipBlock_sw[dsid] & ~item);
		}
		rtlglue_drvMutexUnlock();
		return SUCCESS;
	}
}

void rtl8651_dosProc_blockSip_freeAll(void)
{
	int32 idx;

	CHECK_SIPBLOCK_DISABLE_RETURN_VOID();

	rtlglue_drvMutexLock();

	
	_rtl8651_dosProc_blockSip_discharge(0xffffffff);
	assert(inuseJail == NULL);
	for (idx = 0 ; idx < RTL8651_VLAN_NUMBER ; idx ++)
	{
		assert(blockVlanIpCnt[idx] == 0);
	}

	rtlglue_drvMutexUnlock();
}

int32 rtl8651_dosProc_blockSip_setPrisonTime(uint32 prisonTime)
{
	CHECK_SIPBLOCK_DISABLE(FAILED);
	rtlglue_drvMutexLock();
	_rtl8651_dos_sipBlock_prisonTerm = prisonTime;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}

void rtl8651_dosProc_blockSip_enable(uint32 dsid, uint32 enable)
{
#if 0
	uint32 true = TRUE;
	uint32 false = FALSE;
#endif
	CHECK_SIPBLOCK_DISABLE_RETURN_VOID();

	rtlglue_drvMutexLock();

	if (enable)
	{
		_rtl8651_dosProc_sipBlock_sw[dsid] = TRUE;
#if 0
		rtl8651_dosSIPBlockSw(0xffffffff, dsid, &false);
		rtl8651_dosSIPBlockSw(_rtl8651_dos_sipBlock_sw[dsid], dsid, &true);
#endif
	}else
	{
		if (_rtl8651_dos_sipBlock_sw[dsid])
		{
			rtl8651_dosProc_blockSip_freeAll();
		}
		_rtl8651_dosProc_sipBlock_sw[dsid] = FALSE;
#if 0
		rtl8651_dosSIPBlockSw(0xffffffff, dsid, &false);
#endif
	}

	rtlglue_drvMutexUnlock();
	
}


void rtl8651_dosProc_blockSip_dumpJail(void)
{
	_rtl8651_dosProc_blockIp_t *ptr;
	int32 inuseJailCnt, emptyJailCnt, idx;

	CHECK_SIPBLOCK_DISABLE_RETURN_VOID();

	rtlglue_drvMutexLock();

	rtlglue_printf("\nSource IP Blocking System:\n");

	ptr = inuseJail;
	inuseJailCnt = 0;
	while (ptr)
	{
		inuseJailCnt ++;
		ptr = ptr->jail_next;
	}
	ptr = emptyJail;
	emptyJailCnt = 0;
	while (ptr)
	{
		emptyJailCnt ++;
		ptr = ptr->jail_next;
	}
	rtlglue_printf("\n");

	rtlglue_printf("\tPrison Time:\t\t%d sec.\n\n", _rtl8651_dos_sipBlock_prisonTerm);
	rtlglue_printf("DOS attack Mask:\n");
	rtlglue_printf("\tSession\t\tMask\n");
	rtlglue_printf("\t========================\n");
	for (idx = 0 ; idx < RTL8651_MAX_DIALSESSION ; idx ++)
		rtlglue_printf("\t%d\t\t0x%X\n", idx, _rtl8651_dos_sipBlock_sw[idx]);

	rtlglue_printf("\n");

	rtlglue_printf("System Usage:\n");
	
	rtlglue_printf("\t\tTotal\tUsed\tFree\n");
	rtlglue_printf("\tJail\t%d\t%d\t%d\n", (inuseJailCnt + emptyJailCnt), inuseJailCnt, emptyJailCnt);

	rtlglue_printf("\n");
	
	rtlglue_printf("Blocked IP List:\n\n");
	rtlglue_printf("\t\tIP\t\tVID\tDischarge Time(sec.)\n");
	rtlglue_printf("\t==========================================================\n");

	ptr = inuseJail;
	if (!ptr)
	{
		rtlglue_printf("\t\t\t(Empty Entry)\n");
	}
	else while (ptr)
	{
		rtlglue_printf("\t%03u.%03u.%03u.%03u\t\t0x%X\t\t%d\n", NIPQUAD(ptr->ip), ptr->vid, ptr->dischargeTime);
		ptr = ptr->jail_next;
	}
	rtlglue_printf("\t==========================================================\n");
	rtlglue_printf("\t\t\t\t\tCurrent Time: %d sec.\n", _sysUpSeconds);
	rtlglue_drvMutexUnlock();
}



