#include "types.h"
#include "rtl_glue.h"
#include "mbuf.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_timer.h"
#include "rtl8651_dos.h"
#include "rtl8651_dos_local.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_ipQueue.h"
#include "rtl8651_ipQueue_local.h"


#include "rtl8651_jhash.h"


#ifdef RTL865X_DEBUG
static int IpReasmReqds;
static int IpReasmFails;
static int IpReasmTimeout;
static int IpReasmOKs;
#define IP_INC_STATS_BH(x) x++
#define DLOG(fmt,args...) rtlglue_printf(fmt,#args)
#else
#define DLOG(fmt,args...) 
#define IP_INC_STATS_BH(x)
#endif


#define IPQ_HASHSZ	64

static struct ipq *ipq_hash[IPQ_HASHSZ];
static uint32 ipfrag_hash_rnd=31;


static struct ipq *negative_ipq[IPQ_HASHSZ];


static struct ipq *positive_ipq[IPQ_HASHSZ];


static struct ipq *lru_list = NULL;			
static struct ipq *lru_negative_list = NULL;	
static struct ipq *lru_positive_list = NULL;	





static uint32 sysctl_ipfrag_time = RTL8651_DefaultDrvMaxFragTimeOut;				
static uint32 negative_list_time = RTL8651_DefaultDrvMaxNegativeListTimeOut;			
static uint32 rtl8651_drvMaxNegativeListCnt = RTL8651_DefaultDrvMaxNegativeListCnt;	
static uint32 rtl8651_drvMaxPositiveListCnt = RTL8651_DefaultDrvMaxPositiveListCnt;	
static uint32 rtl8651_drvMaxFragSubPktCnt = RTL8651_DefaultDrvMaxFragSubPktCnt;	
static uint32 rtl8651_drvMaxFragPoolCnt = RTL8651_DefaultDrvMaxFragPoolCnt;		
static uint32 rtl8651_drvMaxFragPktCnt = RTL8651_DefaultDrvMaxFragPktCnt;			
static uint32 negative_list_cnt = 0;												
static uint32 positive_list_cnt = 0;												
static uint32 frag_flow_cnt = 0;													
static uint32 frag_pkt_cnt = 0;													
static uint32 ip_frag_mem = 0;													

static uint8 system_enable = TRUE;

struct ipq *ipq_pool;
static struct ipq* freelist;


extern uint16				*_rtl8651_udpSizeThreshValue;
extern uint8				_rtl8651_fragQueueProcess;
extern uint8				_rtl8651_fragQueueDropChain;
extern int8				rtl8651_drvIcmpRoutingMsgEnable;
#ifdef _RTL_NAPT_REDIRECT
extern ipaddr_t			_rtl8651_fragQueueNaptRedirect_MatchedIP;
#endif 

#define FRAG_CB(pktHdrPtr)	((uint16)(pktHdrPtr->ph_fragOffset))



struct ipq {
	struct ipq		*next;				
	struct ipq		**pprev;			
	uint32			saddr;
	uint32			daddr;
	uint16			id;
	uint8			protocol;
	uint8			last_in;
	#define LAST_IN			0x01		
	#define FIRST_IN			0x02		
	#define NEGATIVE_LIST	0x04		
	#define POSITIVE_LIST	0x08		
	struct rtl_pktHdr 	*fragments;			
	int				subPktCnt;			
	int				len;					
	int				meat;
#if 0
	int				refcnt;
#endif
	struct rtl8651_timer_list	timer;				

	unsigned int		hash_idx;			

	
	struct ipq *prev_lru;					
	struct ipq *next_lru;

};




#if 0
static void dump_ipqhash(void);
#endif
static __inline__ unsigned int ipqhashfn(uint16, uint32, uint32, uint8);
static __inline__ struct ipq *ipq_alloc(void);
static __inline__ void ipq_link(struct ipq *, unsigned int);
static __inline__ void ipq_unlink(struct ipq *);
static __inline__ void ipq_free(struct ipq *);
static __inline__ void frag_freepktMbuf(struct rtl_pktHdr *);
static __inline__  void frag_freeMbufChain(struct ipq *);
static __inline__ void negative_unlink(struct ipq *);
static __inline__ void negative_free(struct ipq *);
static void frag_delNegative(struct ipq *);
static void frag_ipq2Negative(struct ipq *);
static __inline__ void positive_unlink(struct ipq *);
static __inline__ void positive_free(struct ipq *);
static void frag_delPositive(struct ipq *);
static void frag_ipq2Positive(struct ipq *);
static void frag_destroy(struct ipq *);
static void frag_tryDelNegative(void);
static int32 frag_tryExpire(void);
static void frag_expire(unsigned long);
static struct ipq *ip_frag_create(unsigned int, struct ip *);
static struct ipq *ip_frag_find(struct ip *);
int32 ip_frag_queue(struct ipq *, struct rtl_pktHdr *,struct ip *);
static int32 _rtl8651_flushFragQueue(struct rtl_pktHdr *, struct rtl_pktHdr *);
static __inline__ void _rtl8651_IpFragQueueSetInitPara(rtl8651_fwdEngineInitPara_t *para);
static __inline__ void ipqPool_init(void);
static __inline__ void ipqPool_reinit(void);
static void * ipqPool_malloc(uint32);
static void ipqPool_free(void*);


void rtl8651_IpFragQueueSetTimeout(_rtl8651_IpFragQueueTimeout_t *para)
{
	if (!para)
		return;

	rtlglue_drvMutexLock();
	if (para->MaxFragTimeout)
	{
		sysctl_ipfrag_time = para->MaxFragTimeout;
	}
	if (para->MaxNegativeListTimeOut)
	{
		negative_list_time = para->MaxNegativeListTimeOut;
	}
	rtlglue_drvMutexUnlock();
}
void _rtl8651_IpFragQueueSetInitPara(rtl8651_fwdEngineInitPara_t *para)
{
	if (!para)
		return;

	rtl8651_drvMaxFragPktCnt = para->MaxFragPktCnt;
	rtl8651_drvMaxFragSubPktCnt = para->MaxFragSubPktCnt;
	rtl8651_drvMaxNegativeListCnt = para->MaxNegativeListEntryCnt;
	rtl8651_drvMaxPositiveListCnt = para->MaxPositiveListEntryCnt;
	rtl8651_drvMaxFragPoolCnt = para->MaxFragPoolCnt;

	if (para->MaxFragTimeOut)
	{
		sysctl_ipfrag_time = para->MaxFragTimeOut;
	}
	if (para->MaxNegativeListTimeOut)
	{
		negative_list_time = para->MaxNegativeListTimeOut;
	}

	if ((!rtl8651_drvMaxFragPktCnt) ||
		((!rtl8651_drvMaxFragSubPktCnt) && (!rtl8651_drvMaxFragPoolCnt)))
	{
		system_enable = FALSE;
	}
}

int32 rtl8651_IpFragQueueReinit(void)
{
	int32 index;

	if (system_enable == FALSE)
	{
		return SUCCESS;
	}

	rtlglue_drvMutexLock();

	for (index = 0 ; index < IPQ_HASHSZ ; index ++)
	{
		
		while (ipq_hash[index])
		{
			frag_destroy(ipq_hash[index]);	
		}
		
		while(negative_ipq[index])
		{
			frag_delNegative(negative_ipq[index]);
		}
		
		while(positive_ipq[index])
		{
			frag_delPositive(positive_ipq[index]);
		}
	}

	ipqPool_reinit();

	assert(lru_list == NULL);
	assert(lru_negative_list == NULL);
	assert(lru_positive_list == NULL);
	assert(negative_list_cnt == 0);
	assert(positive_list_cnt == 0);
	assert(frag_flow_cnt == 0);
	assert(frag_pkt_cnt == 0);
	assert(ip_frag_mem == 0);

	rtlglue_drvMutexUnlock();

	return SUCCESS;
}

int32 _rtl8651_IpFragQueueInit(rtl8651_fwdEngineInitPara_t *para)
{
	if (para)
	{
		_rtl8651_IpFragQueueSetInitPara(para);
	}
	if (system_enable == FALSE)
	{
		return SUCCESS;
	}
	ipqPool_init();
	rtl8651_IpFragQueueReinit();
	return SUCCESS;
}

#if 0
void dump_ipqhash(void)
{
	struct rtl_pktHdr *pktHdrPtr;
	struct ipq *next;
	int pktnum,i;
	DLOG(" ipqhash:\n");
	for (i=0;i<IPQ_HASHSZ;i++)
	{
		
		 next = ipq_hash[i];
		 while (next)
		{
			
			pktHdrPtr= next->fragments ;
			pktnum = 0;
			while (pktHdrPtr)
			{
				pktnum++;
				pktHdrPtr=(struct rtl_pktHdr *)pktHdrPtr->ph_nextHdr;
			}
			DLOG(" %d: %x queue pkts %d \n",i,(uint32)next,pktnum);
			next=next->next;
		}
	}
}
#endif

static __inline__ unsigned int ipqhashfn(uint16 id, uint32 saddr, uint32 daddr, uint8 prot)
{
	return jhash_3words((uint32)id << 16 | prot, saddr, daddr,
			    ipfrag_hash_rnd) & (IPQ_HASHSZ - 1);
}

void ipqPool_init(void)
{
	rtlglue_drvMutexLock();

	ipq_pool = (struct ipq *)rtlglue_malloc(sizeof(struct ipq) * (rtl8651_drvMaxFragPktCnt + rtl8651_drvMaxNegativeListCnt + rtl8651_drvMaxPositiveListCnt));
	freelist = ipq_pool;

	rtlglue_drvMutexUnlock();
}
void ipqPool_reinit(void)
{
	uint32 idx;
	struct ipq *ptr;

	if ((rtl8651_drvMaxFragPktCnt + rtl8651_drvMaxNegativeListCnt + rtl8651_drvMaxPositiveListCnt) == 0)
	{
		freelist = NULL;
		return;
	}

	
	memset(ipq_pool, 0, (sizeof(struct ipq) * (rtl8651_drvMaxFragPktCnt + rtl8651_drvMaxNegativeListCnt + rtl8651_drvMaxPositiveListCnt)));

	
	freelist = ipq_pool;

	ptr = ipq_pool;

	for (idx = 0 ; idx < (rtl8651_drvMaxFragPktCnt + rtl8651_drvMaxNegativeListCnt + rtl8651_drvMaxPositiveListCnt) ; idx ++, ptr ++)
	{
		
		if (idx == ((rtl8651_drvMaxFragPktCnt + rtl8651_drvMaxNegativeListCnt + rtl8651_drvMaxPositiveListCnt) - 1))	
		{
			ptr->next = NULL;
		}else
		{
			ptr->next = (ptr+1);
		}
	}
	return;
}
void * ipqPool_malloc(uint32 ipq_sz)
{
	struct ipq *ret;

	assert(ipq_sz == sizeof(struct ipq));

	if (!freelist)
	{
		return NULL;
	}

	ret = freelist;

	freelist = ret->next;

	return (void *)ret;
}
void ipqPool_free(void* ipq)
{
	memset(ipq, 0, sizeof(struct ipq));
	((struct ipq *)ipq)->next = freelist;
	freelist = ipq;
}


static __inline__ struct ipq *ipq_alloc(void)
{
	struct ipq *qp;

	qp = (struct ipq*)ipqPool_malloc(sizeof(struct ipq));

	if(!qp)
	{
		goto err;;
	}

	
	qp->next = NULL;
	qp->pprev = NULL;

	
	ip_frag_mem += sizeof(struct ipq);
	frag_flow_cnt ++;

	return qp;
err:
	return NULL;
}

static __inline__ void ipq_link(struct ipq *qp, unsigned int hash)
{
	struct ipq *orig;

	
	_rtl8651_init_timer(&qp->timer);
	qp->timer.data = (uint32) qp;				
	qp->timer.function =(void*)frag_expire;		
	_rtl8651_mod_timer(&qp->timer, _sysUpSeconds + sysctl_ipfrag_time);

	
	orig = ipq_hash[hash];				
	qp->next = orig;
	qp->pprev = &ipq_hash[hash];			
	if (orig)
	{
		orig->pprev = &(qp->next);		
	}
	ipq_hash[hash] = qp;

}

static __inline__ void ipq_unlink(struct ipq *qp)
{
	
	_rtl8651_del_timer(&qp->timer);

	
	if(qp->next)
	{
		qp->next->pprev = qp->pprev;
	}
	if (qp->pprev)
	{
		*qp->pprev = qp->next;
	}

	qp->next = NULL;
	qp->pprev = NULL;

	
	if (qp->next_lru)
	{
		qp->next_lru->prev_lru = qp->prev_lru;
	}
	if (qp->prev_lru)
	{
		qp->prev_lru->next_lru = qp->next_lru;
	}
	if (lru_list == qp)
	{
		lru_list = qp->next_lru;
	}
	qp->next_lru = NULL;
	qp->prev_lru = NULL;

}

static __inline__ void ipq_free(struct ipq *qp)
{
	
	ipqPool_free(qp);

	
	ip_frag_mem -= sizeof(struct ipq);
	frag_flow_cnt --;
}

static __inline__ void frag_freepktMbuf(struct rtl_pktHdr *pkt)
{
	frag_pkt_cnt --;
	ip_frag_mem -= pkt->ph_mbuf->m_extsize;
	pkt->ph_nextHdr = NULL;
	mBuf_freeMbufChain(pkt->ph_mbuf);
}

static __inline__  void frag_freeMbufChain(struct ipq *qp)
{
	struct rtl_pktHdr *head, *fp;

	if ((head = qp->fragments)!=NULL)
	{
		for ( ; head ; )
		{
			fp = head->ph_nextHdr;
			frag_freepktMbuf(head);
			head = fp;
		}
	}
	return;
}

void frag_destroy(struct ipq *qp)
{
	
	ipq_unlink(qp);

	
	frag_freeMbufChain(qp);

	
	frag_ipq2Negative(qp);

	
	frag_flow_cnt --;
}


int32 frag_tryExpire(void)
{
	uint8 expireOK = FALSE;

	while (lru_list)
	{
		if (_sysUpSeconds >= lru_list->timer.expires)
		{	
			frag_expire((unsigned long)lru_list);
			expireOK = TRUE;
		}else
		{
			break;
		}
	}

	if (expireOK)
	{
		return SUCCESS;
	}


	return FAILED;
}

void frag_expire(unsigned long arg)
{
	struct ipq *qp = (struct ipq *) arg;

 	
	IP_INC_STATS_BH(IpReasmTimeout);
	IP_INC_STATS_BH(IpReasmFails);

	
	if (rtl8651_drvIcmpRoutingMsgEnable)
	{
		
		_rtl8651_IcmpPktField_t icmp_info;
		struct rtl_pktHdr *pkthdr = qp->fragments;
		struct ip *pip = (struct ip *)((uint8 *)(pkthdr->ph_mbuf->m_data) + (pkthdr->ph_iphdrOffset));
		rtl8651_tblDrv_vlanTable_t *vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
		uint16 vid = vlanp->vid;

		icmp_info.ic_type = ICMP_TIMXCEED;	
		icmp_info.ic_code = 1;	
		icmp_info.ic_mtu = 0;

		
		qp->fragments = pkthdr->ph_nextHdr;
		pkthdr->ph_nextHdr = NULL;
		ip_frag_mem -= pkthdr->ph_mbuf->m_extsize;
		frag_pkt_cnt --;

		if (_rtl8651_drvIcmpErrorGeneration(pkthdr, pip, &icmp_info, vid) != SUCCESS)
		{
			mBuf_freeMbufChain(pkthdr->ph_mbuf);
		}
	}

	
	frag_destroy(qp);
}


struct ipq *ip_frag_create(unsigned int hash, struct ip *pip)
{
	struct ipq *qp;
	struct ipq *lru;
	struct ipq *ipq_recycle;


	
	if ((frag_flow_cnt >= rtl8651_drvMaxFragPktCnt) &&
		(frag_tryExpire() != SUCCESS))
	{
		goto out_err;
	}

	
	if ((qp = ipq_alloc()) == NULL)
	{
		goto out_nomem;
	}

	
	qp->protocol		= pip->ip_p;
	qp->id			= ntohs(pip->ip_id);
	qp->saddr		= PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	qp->daddr		= PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
	qp->last_in		= 0;
	qp->len			= 0;
	qp->meat		= 0;
	qp->fragments	= NULL;
	qp->subPktCnt	= 0;
	qp->hash_idx		= hash;
	qp->prev_lru		= NULL;
	qp->next_lru		= NULL;

	
	ipq_link(qp, hash);

	
	if (!lru_list)
	{
		lru_list = qp;
	}else
	{
		lru = lru_list;
		while(lru->next_lru)
		{
			lru = lru->next_lru;
		}
		lru->next_lru = qp;
		qp->prev_lru = lru;
	}
	return qp;

out_err:
	
	if ((ipq_recycle = ipqPool_malloc(sizeof(struct ipq))) != NULL)
	{
		ip_frag_mem += sizeof(struct ipq);
		goto add_negative;
	}
	
out_nomem:
	
	if (lru_negative_list)
	{
		ipq_recycle = lru_negative_list;
		
		negative_unlink(ipq_recycle);
		negative_list_cnt --;
	}else
	{	
		goto out;
	}

add_negative:
		
		ipq_recycle->protocol		= pip->ip_p;
		ipq_recycle->id			= ntohs(pip->ip_id);
		ipq_recycle->saddr		= PKTGET_UINT32_UNALIGNED(&pip->ip_src);
		ipq_recycle->daddr		= PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
		ipq_recycle->last_in		= 0;
		ipq_recycle->len			= 0;
		ipq_recycle->meat			= 0;
		ipq_recycle->fragments		= NULL;
		ipq_recycle->subPktCnt		= 0;
		ipq_recycle->hash_idx		= hash;
		ipq_recycle->prev_lru		= NULL;
		ipq_recycle->next_lru		= NULL;

		
		frag_ipq2Negative(ipq_recycle);

out:
	return NULL;
}



static __inline__ void positive_unlink(struct ipq *qp)
{
	
	if(qp->next)
	{
		qp->next->pprev = qp->pprev;
	}
	if (qp->pprev)
	{
		*qp->pprev = qp->next;
	}

	qp->next = NULL;
	qp->pprev = NULL;

	
	if (qp->next_lru)
	{
		qp->next_lru->prev_lru = qp->prev_lru;
	}
	if (qp->prev_lru)
	{
		qp->prev_lru->next_lru = qp->next_lru;
	}
	if (lru_positive_list == qp)
	{
		lru_positive_list = qp->next_lru;
	}
	qp->next_lru = NULL;
	qp->prev_lru = NULL;
}

static __inline__ void positive_free(struct ipq *qp)
{
	
	ipqPool_free(qp);

	
	ip_frag_mem -= sizeof(struct ipq);
}

void frag_delPositive(struct ipq *qp)
{
	
	positive_unlink(qp);
	
	positive_free(qp);
	
	positive_list_cnt --;
}

void frag_ipq2Positive(struct ipq *qp)
{
	unsigned int hash;
	struct ipq *orig;
	struct ipq *lru_positive;

	
	if (positive_list_cnt >= rtl8651_drvMaxPositiveListCnt)
	{
		if (lru_positive_list)
		{	
			frag_delPositive(lru_positive_list);
		} else
		{
			
			frag_ipq2Negative(qp);
			goto out;
		}
	}

	qp->last_in = POSITIVE_LIST;	

	
	hash = qp->hash_idx;
	orig = positive_ipq[hash];					
	qp->next = orig;
	qp->pprev = &positive_ipq[hash];			
	if (orig)
	{
		orig->pprev = &(qp->next);		
	}
	positive_ipq[hash] = qp;

	
	qp->next_lru = NULL;
	qp->prev_lru = NULL;
	
	if (!lru_positive_list)
	{
		lru_positive_list = qp;
	}else
	{
		lru_positive = lru_positive_list;
		while(lru_positive->next_lru)
		{
			lru_positive = lru_positive->next_lru;
		}
		lru_positive->next_lru = qp;
		qp->prev_lru = lru_positive;
	}
	positive_list_cnt ++;

out:
	return;
}



static __inline__ void negative_unlink(struct ipq *qp)
{
	_rtl8651_del_timer(&qp->timer);

	
	if(qp->next)
	{
		qp->next->pprev = qp->pprev;
	}
	if (qp->pprev)
	{
		*qp->pprev = qp->next;
	}

	qp->next = NULL;
	qp->pprev = NULL;

	
	if (qp->next_lru)
	{
		qp->next_lru->prev_lru = qp->prev_lru;
	}
	if (qp->prev_lru)
	{
		qp->prev_lru->next_lru = qp->next_lru;
	}
	if (lru_negative_list == qp)
	{
		lru_negative_list = qp->next_lru;
	}
	qp->next_lru = NULL;
	qp->prev_lru = NULL;
}

static __inline__ void negative_free(struct ipq *qp)
{
	
	ipqPool_free(qp);

	
	ip_frag_mem -= sizeof(struct ipq);
}

void frag_delNegative(struct ipq *qp)
{
	
	negative_unlink(qp);
	
	negative_free(qp);
	
	negative_list_cnt --;
}

void frag_ipq2Negative(struct ipq *qp)
{
	unsigned int hash;
	struct ipq *orig;
	struct ipq *lru_negative;

	
	if (negative_list_cnt >= rtl8651_drvMaxNegativeListCnt)
	{
		if (lru_negative_list)
		{	
			frag_delNegative(lru_negative_list);
		} else
		{
			
			negative_free(qp);
			goto out;
		}
	}

	qp->last_in = NEGATIVE_LIST;	

	
	_rtl8651_init_timer(&qp->timer);
	qp->timer.data = (uint32) qp;						
	qp->timer.function =(void*)frag_delNegative;		
	_rtl8651_mod_timer(&qp->timer, _sysUpSeconds + negative_list_time);

	
	hash = qp->hash_idx;
	orig = negative_ipq[hash];					
	qp->next = orig;
	qp->pprev = &negative_ipq[hash];			
	if (orig)
	{
		orig->pprev = &(qp->next);		
	}
	negative_ipq[hash] = qp;

	
	qp->next_lru = NULL;
	qp->prev_lru = NULL;
	
	if (!lru_negative_list)
	{
		lru_negative_list = qp;
	}else
	{
		lru_negative = lru_negative_list;
		while(lru_negative->next_lru)
		{
			lru_negative = lru_negative->next_lru;
		}
		lru_negative->next_lru = qp;
		qp->prev_lru = lru_negative;
	}

	negative_list_cnt ++;

out:
	return;
}


void frag_tryDelNegative(void)
{
	while (lru_negative_list)
	{
		if (_sysUpSeconds >= lru_negative_list->timer.expires)
		{	
			frag_delNegative(lru_negative_list);
		}else
		{
			goto out;
		}
	}
out:
	return;
}


struct ipq *ip_frag_find(struct ip *pip)
{

	uint16 id = ntohs(pip->ip_id);
	uint32 saddr = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	uint32 daddr = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
	uint8 protocol = pip->ip_p;
	unsigned int hash = ipqhashfn(id, saddr, daddr, protocol);
	struct ipq *qp;
	
	frag_tryDelNegative();

	
	if (positive_ipq[hash] == NULL)
	{
		goto check_negative;
	}

	for (qp = positive_ipq[hash]; qp; qp = qp->next)
	{
		if(qp->id == id		&&
		   qp->saddr == saddr	&&
		   qp->daddr == daddr	&&
		   qp->protocol == protocol)
		{
			return qp;	
		}
	}

check_negative:
	
	if (negative_ipq[hash] == NULL)
	{
		goto check_incomplete;
	}

	for (qp = negative_ipq[hash]; qp; qp = qp->next)
	{
		if(qp->id == id		&&
		   qp->saddr == saddr	&&
		   qp->daddr == daddr	&&
		   qp->protocol == protocol)
		{
 			return NULL;	
		}
	}

check_incomplete:

	
	if (ipq_hash[hash] == NULL)
	{
		goto new_one;
	}
	
	for (qp = ipq_hash[hash]; qp; qp = qp->next)
	{
		if(qp->id == id		&&
		   qp->saddr == saddr	&&
		   qp->daddr == daddr	&&
		   qp->protocol == protocol)
		{
 			return qp;
		}
	}


new_one:
	return ip_frag_create(hash, pip);
}


int32 ip_frag_queue(struct ipq *qp, struct rtl_pktHdr *pktHdrPtr,struct ip * ip)
{
	struct rtl_pktHdr  *prev, *next;
	int flags, offset;
	int ihl, end;
	int32 retval;
	rtl8651_tblDrv_vlanTable_t *vlanp = NULL;
	uint8 direction = 0;
	uint32 dosIgnore = FALSE;
	uint32 ipQueuePassThrough = TRUE;

	
	if ((rtl8651_drvMaxFragSubPktCnt > 0) &&
		(qp->subPktCnt >= rtl8651_drvMaxFragSubPktCnt))
	{
		goto err_destroy_pkt_qp;
	}
	if ((rtl8651_drvMaxFragPoolCnt > 0) &&
		(frag_pkt_cnt >= rtl8651_drvMaxFragPoolCnt))
	{
		goto err_destroy_pkt_qp;
	}

	
	if (rtl8651a_dosFragmentCheck(pktHdrPtr->ph_dsid, ip, pktHdrPtr) == SUCCESS)
	{	
		goto err_dos_detected;
	}
	
 	offset = ntohs(ip->ip_off);
	flags = offset & ~IP_OFFMASK;							
	offset &= IP_OFFMASK;									
	offset <<= 3;											
 	ihl = ((ip->ip_vhl & 0x0f)<< 2);							
	pktHdrPtr->ph_dataLen = ntohs(ip->ip_len) - ihl;			

 	end = offset + pktHdrPtr->ph_dataLen;				

	
	if ((flags & IP_MF) == 0) {
		if ((end < qp->len) ||
			((qp->last_in & LAST_IN) && (end != qp->len)))
		{	
			goto err_TearDrop;
		}
		
		qp->last_in |= LAST_IN;
		qp->len = end;
	} else {
		if (end & 0x07) {
			end &= ~0x07;
		}
		if (end > qp->len) {
			if (qp->last_in & LAST_IN)
			{	
				goto err_TearDrop;
			}
			qp->len = end;
		}
	}

	
	prev = NULL;
	if (qp->fragments == NULL)
	{
		next = NULL;
	}else
	{
		for(next = qp->fragments ; next ; next = next->ph_nextHdr)
		{
			if (FRAG_CB(next) >= offset)
			{
				break;	
			}
			prev = next;
		}
	}

	
	if (prev) {
		int i = (FRAG_CB(prev) + prev->ph_dataLen) - offset;

		if (i > 0) {
 			goto err_TearDrop;
 		}
	}

	if (next && (FRAG_CB(next) < end))
	{
		int i = end - FRAG_CB(next);

		if (i < next->ph_dataLen) {
			 goto err_TearDrop;
 		}
 	}

	pktHdrPtr->ph_fragOffset=offset;

	
	pktHdrPtr->ph_nextHdr = next;
	if (prev)
	{
		prev->ph_nextHdr = pktHdrPtr;
	}else
	{
		qp->fragments = pktHdrPtr;
	}

	qp->meat += pktHdrPtr->ph_dataLen;
	qp->subPktCnt ++;
	ip_frag_mem += pktHdrPtr->ph_mbuf->m_extsize;
	frag_pkt_cnt ++;

	if (offset == 0)
	{
		qp->last_in |= FIRST_IN;
	}

	
	
	vlanp = &DrvTbl.vlan[pktHdrPtr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	direction = (vlanp->internal?0:1);

	dosIgnore= (_rtl8651_dos_enable & _rtl8651_dosIgnoreTypeCheck(direction));

	
	ipQueuePassThrough = TRUE;
	if (dosIgnore == TRUE)
	{
		
		if ((qp->protocol == IPPROTO_UDP) && (_rtl8651_udpSizeThreshValue[qp->fragments->ph_dsid] != 0xffff))
			ipQueuePassThrough = FALSE;
	}else
	{
		
	
		if (_rtl8651_dos_sw[qp->fragments->ph_dsid] & RTL8651_DOS_TEARDROP)
			ipQueuePassThrough = FALSE;
		else if (qp->protocol == IPPROTO_UDP)
		{
			if ((_rtl8651_udpSizeThreshValue[qp->fragments->ph_dsid] != 0xffff) ||
					(_rtl8651_dos_sw[qp->fragments->ph_dsid] & RTL8651_DOS_UDPBOMB))
				ipQueuePassThrough = FALSE;
		}
	}

	
	if (ipQueuePassThrough == TRUE)
	{
		if (qp->last_in & FIRST_IN)
		{
			uint32 pktComplete;

			IP_INC_STATS_BH(IpReasmOKs);
			ipq_unlink(qp);	

			if ((qp->last_in == (FIRST_IN|LAST_IN)) &&
				(qp->meat == qp->len))
			{
				pktComplete = TRUE;
			}else
			{
				pktComplete = FALSE;
			}

			frag_pkt_cnt -= qp->subPktCnt;
			ip_frag_mem -= ((qp->fragments->ph_mbuf->m_extsize) * (qp->subPktCnt));

			retval = _rtl8651_flushFragQueue(qp->fragments, pktHdrPtr);

			if (retval == FAILED)
			{
				retval = SUCCESS;
				if (pktComplete == TRUE)
				{
					ipq_free(qp);
				}else
				{
					frag_ipq2Negative(qp);	
					frag_flow_cnt --;
				}
			}else
			{
				if (pktComplete == TRUE)
				{
					ipq_free(qp);
				}else
				{
					frag_ipq2Positive(qp);		
					frag_flow_cnt --;
				}
			}
			goto out;
		}
	} else
	{	
		if ((qp->last_in == (FIRST_IN|LAST_IN)) &&
			(qp->meat == qp->len))
		{
			
			IP_INC_STATS_BH(IpReasmOKs);
			
			ipq_unlink(qp);	

			
			if (qp->protocol == IPPROTO_UDP)
			{
				
				if ((_rtl8651_udpSizeThreshValue[qp->fragments->ph_dsid] != 0xffff) &&	
					(qp->len > _rtl8651_udpSizeThreshValue[qp->fragments->ph_dsid]))	
				{	
					frag_freeMbufChain(qp);
					retval = SUCCESS;
					goto flow_done;
				}

				
				{
					struct rtl_pktHdr *first = qp->fragments;
					struct ip *p;
					struct udphdr *u;

					p = (struct ip *)((uint32)(first->ph_mbuf->m_data) + (uint32)(first->ph_iphdrOffset));
					u = (struct udphdr *)((uint32)p + (uint32)((p->ip_vhl & 0x0f) << 2));

					if (rtl8651a_dosFragUdpBombCheck(first->ph_dsid, vlanp->vid, p, u, direction, qp->len, first) == SUCCESS)
					{	
						frag_freeMbufChain(qp);
						retval = SUCCESS;
						goto flow_done;
					}
				}
			}
			{	
				frag_pkt_cnt -= qp->subPktCnt;
				ip_frag_mem -= ((qp->fragments->ph_mbuf->m_extsize) * (qp->subPktCnt));
				retval = _rtl8651_flushFragQueue(qp->fragments, pktHdrPtr);
				if (retval == FAILED)
					retval = SUCCESS;
			}
flow_done:
			ipq_free(qp);	
			goto out;
		}
	}

	{
		
		if ((rtl8651_drvMaxFragSubPktCnt > 0) && 
			(qp->subPktCnt >= rtl8651_drvMaxFragSubPktCnt))
		{	
			goto err_destroy_qp;
		}else
		{
			
			
			struct rtl_pktHdr *newHdr;
			struct rtl_mBuf  *newmb=NULL, *oldMbuf;
			uint8 *newCluster;
			uint32 newClusterId;
			int16 rxdesc;

			if (pktHdrPtr->ph_rxdesc >= 0)	
			{
				oldMbuf = pktHdrPtr->ph_mbuf;

				
				if((newmb = mBuf_get(MBUF_DONTWAIT,MBUFTYPE_DATA, 1)) == NULL)
				{
					goto err_destroy_qp;
				}
				if((mBuf_getPkthdr(newmb, MBUF_DONTWAIT) == (struct rtl_mBuf *) NULL)||
				(mBuf_reserve(newmb,rtl8651_fwdEnginePara.mbufHeadroom)))
				{
					
					mBuf_freeMbufChain(newmb);
					goto err_destroy_qp;
				}
	
				assert(newmb);
				
				newHdr = newmb->m_pkthdr;
				newCluster = newmb->m_extbuf;
				newClusterId = newmb->m_extClusterId;
				assert(newHdr);
				*newmb = *oldMbuf;
				newmb->m_next = NULL;
				newmb->m_pkthdr = newHdr;
		
				rxdesc = newHdr->ph_rxdesc;
				*newHdr = *pktHdrPtr;
				newHdr->ph_mbuf = newmb;
				newHdr->ph_rxdesc = rxdesc;
				oldMbuf->m_extClusterId = newClusterId;
				oldMbuf->m_data = newCluster;
				oldMbuf->m_extbuf = newCluster;
				oldMbuf->m_len = 0;
				pktHdrPtr->ph_nextHdr = 0;
				mBuf_freeMbufChain(pktHdrPtr->ph_mbuf);		
		
				
				newHdr->ph_nextHdr = next;
				if (prev)
					{
					prev->ph_nextHdr = newHdr;
				}else
				{
					qp->fragments = newHdr;
				}
			}
			retval = FWDENG_QUEUED;
		}
	}

out:
 	return retval;

err_TearDrop:
	
	{
		struct rtl_pktHdr *pkthdr;
		struct ip *pip;
		ipaddr_t dip;

		if (offset)
		{	
			pkthdr = qp->fragments;	
			pip = (struct ip *)((uint8 *)(pkthdr->ph_mbuf->m_data) + (pkthdr->ph_iphdrOffset));
		}else
		{	
			pkthdr = pktHdrPtr;
			pip = ip;
		}

		dip = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);

		if (vlanp == NULL)
		{	
			vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
			direction = (vlanp->internal?0:1);	
		}
		rtl8651a_dosTearDropLogging(pkthdr->ph_dsid, vlanp->vid, pip, direction, pkthdr);
	}
err_dos_detected:
err_destroy_pkt_qp:
	mBuf_freeMbufChain(pktHdrPtr->ph_mbuf);		
err_destroy_qp:
	IP_INC_STATS_BH(IpReasmFails);
	
	frag_destroy(qp);
	return SUCCESS;
}


inline int32 _rtl8651_frag_NeedFragmentProcess(uint16 id, ipaddr_t sip, ipaddr_t dip, uint8 l4proto)
{
	unsigned int hash;
	struct ipq *qp;

	
	if (_rtl8651_fragQueueProcess)
		return FALSE;

	hash = ipqhashfn(id, sip, dip, l4proto);

	
	if (negative_ipq[hash] == NULL)
	{
		goto check_incomplete;
	}

	for (qp = negative_ipq[hash]; qp; qp = qp->next)
	{
		if(qp->id == id		&&
		   qp->saddr == sip	&&
		   qp->daddr == dip	&&
		   qp->protocol == l4proto)
		{
 			return TRUE;	
		}
	}

check_incomplete:

	
	if (ipq_hash[hash] == NULL)
		return FALSE;

	
	for (qp = ipq_hash[hash]; qp; qp = qp->next)
	{
		if(qp->id == id		&&
		   qp->saddr == sip	&&
		   qp->daddr == dip	&&
		   qp->protocol == l4proto)
		{
 			return TRUE;
		}
	}
	return FALSE;
}



int32 _rtl8651_ip_defrag(struct rtl_pktHdr * pkthdrPtr, struct ip *pip)
{
	struct ipq *qp;

	if (system_enable == FALSE)
	{
		return FWDENG_IP_CANTFRAG;
	}

	
	IP_INC_STATS_BH(IpReasmReqds);

 	
	if ((qp = ip_frag_find(pip)) != NULL) {

		if (qp->last_in & POSITIVE_LIST)
		{	
			return FAILED;
		}

		
		return ip_frag_queue(qp, pkthdrPtr, pip);
	}else
	{
		
		mBuf_freeMbufChain(pkthdrPtr->ph_mbuf);
	}

	IP_INC_STATS_BH(IpReasmFails);
	return SUCCESS;
}

static int32 _rtl8651_flushFragQueue(struct rtl_pktHdr *head, struct rtl_pktHdr *currentPkt)
{
	struct rtl_pktHdr  *hp, *fp;
	int val;

	if (!head)
	{
		return FAILED;
	}

	_rtl8651_fragQueueDropChain = FALSE;
	_rtl8651_fragQueueProcess = TRUE;
#ifdef _RTL_NAPT_REDIRECT
	_rtl8651_fragQueueNaptRedirect_MatchedIP = 0;
#endif 

	hp = head;
	while(hp)
	{
		fp = hp->ph_nextHdr;
		hp->ph_nextHdr = NULL;
		val = rtl8651_fwdEngineInput(hp);

		if (_rtl8651_fragQueueDropChain == TRUE)
		{	
			hp = fp;
			goto errout;
		}

		if (val != SUCCESS)
		{
			if (hp == head)
			{	
				hp->ph_nextHdr = fp;
				goto concatenate;
			}else
			{	
				hp->ph_nextHdr = fp;
				goto errout;
			}
		}
		hp = fp;
	}

	_rtl8651_fragQueueProcess = FALSE;
	return SUCCESS;

errout:
	
	while(hp)
	{
		fp = hp->ph_nextHdr;
		hp->ph_nextHdr = NULL;
		mBuf_freeMbufChain(hp->ph_mbuf);
		hp = fp;
	}
	_rtl8651_fragQueueProcess = FALSE;
	return FAILED;

concatenate:
	
	{

		if (head == currentPkt)
		{
			goto concatenateOut;
		}

		
		
		hp = head;
		while (hp)
		{
			fp = hp->ph_nextHdr;
	
			if (fp == currentPkt)
			{
				if (hp == head)
				{	
					head->ph_nextHdr = currentPkt->ph_nextHdr;
					currentPkt->ph_nextHdr = head;
				}else
				{
					struct rtl_pktHdr *tmpHdr;

					hp->ph_nextHdr = head;
					
					tmpHdr = head->ph_nextHdr;
					head->ph_nextHdr = currentPkt->ph_nextHdr;
					currentPkt->ph_nextHdr = tmpHdr;
				}
				goto exchange_pkt;
			}
			hp = fp;
		}

		
		hp = head;
		goto errout;

exchange_pkt:
		
		{
			struct rtl_mBuf *mbuf_h;
			struct rtl_mBuf *mbuf_c;
			void *tmp_s;
			uint32 tmp_u32;
			uint16 tmp_u16;
			int8 buff[sizeof(struct rtl_pktHdr)];
			int16 rxdesc_h, rxdesc_c;
			struct rtl_pktHdr *nextHdr_h, *nextHdr_c;

			
			mbuf_h = head->ph_mbuf;
			mbuf_c = currentPkt->ph_mbuf;

			
			rxdesc_h = head->ph_rxdesc;
			rxdesc_c = currentPkt->ph_rxdesc;
			nextHdr_h = head->ph_nextHdr;
			nextHdr_c = currentPkt->ph_nextHdr;

			bzero((void *)buff,sizeof(buff));
			rtl8651_memcpy((void *)buff, (void *)head, sizeof(struct rtl_pktHdr));
			rtl8651_memcpy((void *)head, (void *)currentPkt, sizeof(struct rtl_pktHdr));
			rtl8651_memcpy((void *)currentPkt, buff, sizeof(struct rtl_pktHdr));

			head->ph_mbuf = mbuf_h;
			currentPkt->ph_mbuf = mbuf_c;
			head->ph_rxdesc = rxdesc_h;
			currentPkt->ph_rxdesc = rxdesc_c;
			head->ph_nextHdr = nextHdr_h;
			currentPkt->ph_nextHdr = nextHdr_c;

			
			
			tmp_s = mbuf_h->m_data;
			mbuf_h->m_data = mbuf_c->m_data;
			mbuf_c->m_data = tmp_s;
			
			tmp_s = mbuf_h->m_extbuf;
			mbuf_h->m_extbuf = mbuf_c->m_extbuf;
			mbuf_c->m_extbuf = tmp_s;
			
			tmp_u32 = mbuf_h->m_extClusterId;
			mbuf_h->m_extClusterId = mbuf_c->m_extClusterId;
			mbuf_c->m_extClusterId = tmp_u32;
			
			tmp_u16 = mbuf_h->m_extsize;
			mbuf_h->m_extsize = mbuf_c->m_extsize;
			mbuf_c->m_extsize = tmp_u16;
			
			tmp_u16 = mbuf_h->m_len;
			mbuf_h->m_len = mbuf_c->m_len;
			mbuf_c->m_len = tmp_u16;
		}
	}

concatenateOut:	
	
#ifdef _RTL_NAPT_REDIRECT
	
	if ( _rtl8651_fragQueueNaptRedirect_MatchedIP )
	{	
		struct rtl_pktHdr *m_pkt;
		struct ip *m_ip;
		ipaddr_t dip;

		m_pkt = currentPkt;
		dip = 0;
		while (m_pkt)
		{
			m_ip = (struct ip *)(((uint32)(m_pkt->ph_mbuf->m_data) + (m_pkt->ph_iphdrOffset)));

			
			if (IS_FRAGMENT_REMAIN(m_ip->ip_off))
			{	
				if (!dip)
					dip = PKTGET_UINT32_UNALIGNED(&m_ip->ip_dst);

				PKTSET_UINT32_UNALIGNED(_rtl8651_fragQueueNaptRedirect_MatchedIP, &m_ip->ip_dst);
				_RTL8651_L3_ADJUST_CHECKSUM(_rtl8651_fragQueueNaptRedirect_MatchedIP, dip, m_ip->ip_sum);
			}
			
			m_pkt = m_pkt->ph_nextHdr;
		}
	}
#endif 

	_rtl8651_fragQueueProcess = FALSE;
	return FWDENG_QUEUED_TO_GW;
}




