

#include "rtl_types.h"
#include "rtl_queue.h"
#include "rtl8651_tblAsicDrv.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvFwd.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "voip_support.h"
#include "asicRegs.h"
#ifdef CONFIG_RTL865X_PCM
#include "../voip/pcm2.h"
#endif


#undef _VOIP_DEBUG_


typedef SLIST_HEAD( , voip_callback_s) ListCallback;
ListCallback inuseCallback;
ListCallback freeCallback;
typedef SLIST_HEAD( , voip_rtp_port_s) ListRtp;
ListRtp inuseRtp;
ListRtp freeRtp;



int32 voip_init( int32 numChannel, int32 numRtp )
{
	voip_callback_t *spoolCallback;
	voip_rtp_port_t *spoolRtp;
	int i;
	
	SLIST_INIT( &inuseCallback );
	SLIST_INIT( &freeCallback );
	SLIST_INIT( &inuseRtp );
	SLIST_INIT( &freeRtp );
	
	TBL_MEM_ALLOC( spoolCallback, voip_callback_t, numChannel );
	for( i = 0; i<numChannel; i++ )
		SLIST_INSERT_HEAD( &freeCallback, &spoolCallback[i], nextCallback );

	TBL_MEM_ALLOC( spoolRtp, voip_rtp_port_t, numRtp );
	for( i = 0; i<numRtp; i++ )
		SLIST_INSERT_HEAD( &freeRtp, &spoolRtp[i], nextRtp );
	
	return SUCCESS;
}



int32 voip_collect( void )
{
	voip_callback_t *one;
	voip_rtp_port_t *two;
	
	while ( ( one = SLIST_FIRST( &inuseCallback ) ) )
 	{
		SLIST_REMOVE( &inuseCallback, one, voip_callback_s, nextCallback );
		SLIST_INSERT_HEAD( &freeCallback, one, nextCallback );
	}

	while ( ( two = SLIST_FIRST( &inuseRtp ) ) )
 	{
		SLIST_REMOVE( &inuseRtp, two, voip_rtp_port_s, nextRtp );
		SLIST_INSERT_HEAD( &freeRtp, two, nextRtp );
	}

	SLIST_INIT( &inuseCallback );
	SLIST_INIT( &inuseRtp );
	
	return SUCCESS;
}



int32 voip_register_callback( uint32 chid, voip_callback_t *pReg )
{
	voip_callback_t *one;
	
	for( one = SLIST_FIRST( &inuseCallback );
	     one;
 	     one = SLIST_NEXT( one, nextCallback ) )
 	{
 		if ( one->chid == chid )
 		goto found;
 	}

 	
 	one = SLIST_FIRST( &freeCallback );
 	if ( one == NULL ) return FAILED;
	SLIST_REMOVE( &freeCallback, one, voip_callback_s, nextCallback );
	SLIST_INSERT_HEAD( &inuseCallback, one, nextCallback );

found:

	if ( pReg ) 
	{
		one->chid = chid;
		one->fpPcmInt = pReg->fpPcmInt;
		one->fpPktRx = pReg->fpPktRx;
		one->fpTick = pReg->fpTick;
		one->msTick = pReg->msTick;
		one->msRemain = pReg->msTick;
	}
	else        
	{
		SLIST_REMOVE( &inuseCallback, one, voip_callback_s, nextCallback );
		SLIST_INSERT_HEAD( &freeCallback, one, nextCallback );
	}
	
	return SUCCESS;
}



void voip_dump_callback( void )
{
	voip_callback_t *one;
	
	rtlglue_printf( "chid  tick(ms)   pcmInt   pktRx\n" );
	rtlglue_printf( "--------------------------------------------------\n" );
	for( one = SLIST_FIRST( &inuseCallback );
	     one;
 	     one = SLIST_NEXT( one, nextCallback ) )
 	{
		rtlglue_printf( "  %1d  %p(%d) %p %p\n", 
		                one->chid, one->fpTick, one->msTick,
		                one->fpPcmInt, one->fpPktRx );
 	}
}



int32 voip_ms_routine( int32 msPassed )
{
	voip_callback_t *one;
	
	for( one = SLIST_FIRST( &inuseCallback );
	     one;
 	     one = SLIST_NEXT( one, nextCallback ) )
 	{
 		one->msRemain -= msPassed;
 		if ( one->msRemain <= 0 )
 		{
 			if ( one->fpTick )
	 			one->fpTick( one->chid, msPassed );
 			one->msRemain += one->msTick;
 		}
 	}
 	
	return SUCCESS;
}



int32 voip_pollPcmInt( void )
{
	voip_callback_t *one;
	uint32 pisr;

	pisr = READ_MEM32( PCM_ISR );
	WRITE_MEM32( PCM_ISR, pisr ); 
	for( one = SLIST_FIRST( &inuseCallback );
	     one;
 	     one = SLIST_NEXT( one, nextCallback ) )
 	{
 		
 		if ( pisr & (0xff000000>>(one->chid*8)) )
 		{
#ifdef CONFIG_RTL865X_PCM

#endif
 			if ( one->fpPcmInt )
 			{
	 			one->fpPcmInt( pisr );
 			}
 		}
 	}
 	
	return SUCCESS;
}



int32 voip_query_RTPport( uint32 *chid, uint32 *mid, voip_rtp_port_t *pRtp )
{
	voip_rtp_port_t *one;
	
#ifdef _VOIP_DEBUG_
	rtlglue_printf( "%s( isTcp=%d, remIp=0x%08x, rempPort=%d, extIp=0x%08x, extPort=%d ).\n\n", 
	                __FUNCTION__, pRtp->isTcp, pRtp->remIp, pRtp->remPort, pRtp->extIp, pRtp->extPort );
#endif
	
	for( one = SLIST_FIRST( &inuseRtp );
	     one;
 	     one = SLIST_NEXT( one, nextRtp ) )
 	{
 		if ( one->chid == pRtp->chid &&
 		     one->mid == pRtp->mid )
 		goto found;
 	}

 	
	return FAILED;

found:
	if ( chid ) *chid = one->chid;
	if ( mid ) *mid = one->mid;
	
	return SUCCESS;
}



int32 voip_register_RTPport( uint32 chid, uint32 mid, voip_rtp_port_t* pRtp )
{
	voip_rtp_port_t *one;
	int32 ret;

	if ( pRtp && pRtp->isTcp )
	{
		
		ret = FAILED;
		goto out;
	}
	
#ifdef _VOIP_DEBUG_
	rtlglue_printf( "%s( isTcp=%d, remIp=0x%08x, rempPort=%d, extIp=0x%08x, extPort=%d ).\n\n", 
	                __FUNCTION__, pRtp->isTcp, pRtp->remIp, pRtp->remPort, pRtp->extIp, pRtp->extPort );
#endif
	
	for( one = SLIST_FIRST( &inuseRtp );
	     one;
 	     one = SLIST_NEXT( one, nextRtp ) )
 	{
 		if ( one->chid == chid &&
 		     one->mid == mid )
 		goto found;
 	}

 	
 	one = SLIST_FIRST( &freeRtp );
 	if ( one == NULL )
 	{
		printk("SLIST_FIRST failed\n");
 		ret = FAILED;
 		goto out;
 	}
	SLIST_REMOVE( &freeRtp, one, voip_rtp_port_s, nextRtp );
	SLIST_INSERT_HEAD( &inuseRtp, one, nextRtp );

found:
	if ( pRtp )
	{
		rtl8651_tblDrvAclRule_t acl;
		rtl8651_tblDrv_networkIntfTable_t *intf;
		
		one->chid = chid;
		one->mid = mid;
		one->isTcp = pRtp->isTcp;
		one->remIp = pRtp->remIp;
		one->remPort = pRtp->remPort;
		one->extIp = pRtp->extIp;
		one->extPort = pRtp->extPort;

		
		ret = rtl8651_registerTrappingDispatchEntry( one->isTcp?IPPROTO_TCP:IPPROTO_UDP,
		                                             one->remIp, one->extIp,
		                                             one->remPort, one->extPort,
		                                             VOIP_CATEGORY( one->chid, one->mid ) );
		if ( ret != SUCCESS )
		{
			printk("rtl8651_registerTrappingDispatchEntry failed\n");
			goto out;
		}

		
		intf = _rtl8651_getIpInfo( one->extIp, NULL, NULL, NULL );
		if ( intf==NULL )
		{
			printk("_rtl8651_getIpInfo failed\n");
			ret = FAILED;
			goto out;
		}
		
		
		memset((void*)&acl,0,sizeof(acl));
		acl.ruleType_=RTL8651_ACL_UDP;
		acl.srcIpAddr_		= one->remIp;
		acl.srcIpAddrMask_	= 0xffffffff;
		acl.dstIpAddr_		= one->extIp;
		acl.dstIpAddrMask_	= 0xffffffff;
		acl.udpSrcPortLB_=one->remPort;
		acl.udpSrcPortUB_=one->remPort;
		acl.udpDstPortLB_=one->extPort;
		acl.udpDstPortUB_=one->extPort;
		acl.actionType_ = RTL8651_ACL_CPU;
		ret = rtl8651a_addAclRule( 0, intf->ifName, TRUE,  &acl );
		if ( ret != SUCCESS )
		{
			printk("Add acl rule fail\n");
			goto out;
		}
	}
	else
	{
		
		ret = rtl8651_unregisterTrappingDispatchEntry( one->isTcp?IPPROTO_TCP:IPPROTO_UDP,
		                                               one->remIp, one->extIp,
		                                               one->remPort, one->extPort );

		SLIST_REMOVE( &inuseRtp, one, voip_rtp_port_s, nextRtp );
		SLIST_INSERT_HEAD( &freeRtp, one, nextRtp );
	}

	ret = SUCCESS;
out:
	return ret;
}



int32 voip_pktRx( struct rtl_pktHdr *pPkt )
{
	voip_callback_t *one;
	uint32 chid, mid;

	VOIP_CATEGORY_MAP( pPkt->ph_category, &chid, &mid );
	
#ifdef _VOIP_DEBUG_
	rtlglue_printf( "%s( pPkt->ph_category=%08x ), chid=%d, mid=%d.\n", 
	                __FUNCTION__, pPkt->ph_category, chid, mid );
#endif

	for( one = SLIST_FIRST( &inuseCallback );
	     one;
 	     one = SLIST_NEXT( one, nextCallback ) )
 	{
#ifdef _VOIP_DEBUG_
		rtlglue_printf( "%s():%d one->chid=%d.\n", 
		                __FUNCTION__, __LINE__, one->chid );
#endif
 		if ( one->chid == chid )
 		{
 			
 			struct ip *ipHdr;
			struct udphdr *udp;

 			ipHdr = (struct ip *)(pPkt->ph_mbuf->m_data + pPkt->ph_iphdrOffset);
			udp = (struct udphdr *)L4HDR_ADDR(ipHdr);
 			
			if ( one->fpPktRx )
			{
#ifdef _VOIP_DEBUG_
				rtlglue_printf( "%s():%d udp payload length=%d.\n", 
				                __FUNCTION__, __LINE__, ntohs(udp->uh_ulen)-sizeof(*udp) );				
#endif
				one->fpPktRx( chid, mid, ((uint8*)udp)+sizeof(*udp), ntohs(udp->uh_ulen)-sizeof(*udp), TYPE_NONE );
			}
 			
 			break;
 		}
 	}

	mBuf_freeMbufChain( pPkt->ph_mbuf );
	return SUCCESS;
}



void voip_dump_RTPport( void )
{
	voip_rtp_port_t *one;
	
	rtlglue_printf( "chid mid Pro ExternalIp:port  RemoteIp:port\n" );
	rtlglue_printf( "--------------------------------------------------\n" );
	for( one = SLIST_FIRST( &inuseRtp );
	     one;
 	     one = SLIST_NEXT( one, nextRtp ) )
 	{
		rtlglue_printf( "  %1d   %1d  %3s "NIPQUAD2STR2":%5d "NIPQUAD2STR2":%5d\n", 
		                one->chid, one->mid, one->isTcp?"TCP":"UDP",
		                NIPQUAD2(one->extIp), one->extPort, NIPQUAD2(one->remIp), one->remPort );
 	}
}



int32 voip_TxRtpPkt( uint32 chid, uint32 mid, void* packet, uint32 pktLen, enum VOIP_PKT_TYPE flags )
{
	voip_rtp_port_t *one;
	int32 ret;
	static uint16 ipIdCount = 0x5566; 
	
#ifdef _VOIP_DEBUG_
	rtlglue_printf( "%s(chid=%d, mid=%d, packet=%p, pktLen=%d, flags=0x%08x).\n\n", 
	                __FUNCTION__, chid, mid, packet, pktLen, flags );
#endif

	for( one = SLIST_FIRST( &inuseRtp );
	     one;
 	     one = SLIST_NEXT( one, nextRtp ) )
 	{
 		if ( one->chid == chid &&
 		     one->mid == mid )
 		break;
 	}

	if ( one )
	{
		struct ip *pip;
		struct udphdr *udp;
		uint8 *rtp;
		struct rtl_pktHdr *pkt;
		struct rtl_mBuf  *mbuf;

		
		if( (mbuf = mBuf_get(MBUF_DONTWAIT,MBUFTYPE_DATA, 1)) == NULL )
		{
			ret = FAILED;
#ifdef _VOIP_DEBUG_
			rtlglue_printf( "%s():%d return %d.\n\n", 
			                __FUNCTION__, __LINE__, ret );
#endif
			goto out;
		}
		if( (mBuf_getPkthdr(mbuf, MBUF_DONTWAIT) == (struct rtl_mBuf *) NULL)||
			(mBuf_reserve(mbuf,128+sizeof(struct ether_header)+2)) )
		{
			mBuf_freeMbufChain(mbuf);
			ret = FAILED;
#ifdef _VOIP_DEBUG_
			rtlglue_printf( "%s():%d return %d.\n\n", 
			                __FUNCTION__, __LINE__, ret );
#endif
			goto out;
		}

		pkt = mbuf->m_pkthdr;
		
		pip = (struct ip *)((uint8*)mbuf->m_data);
		udp = (struct udphdr *)(((uint8*)pip)+sizeof(struct ip));
		rtp = (uint8*)((uint8*)udp+sizeof(struct udphdr));

		
		pip->ip_vhl = 0x45;
		pip->ip_tos = 0x00;
		pip->ip_len = ntohs( sizeof(struct ip) + sizeof(struct udphdr) + pktLen );
#ifdef RTL865X_TEST
		pip->ip_id = ntohs(0x7788); 
#else
		pip->ip_id = ntohs(((uint8*)packet)[0]<<8|((uint8*)packet)[1])^((uint32)packet)^((uint32)packet>>16)^(ipIdCount++); 
#endif
		pip->ip_off = 0;
		pip->ip_ttl = 147; 
		pip->ip_p = IPPROTO_UDP;
		
		pip->ip_src.s_addr = ntohl( one->extIp );
		pip->ip_dst.s_addr = ntohl( one->remIp );

		
		udp->uh_ulen = ntohs( pktLen + sizeof(struct udphdr) );
		udp->uh_sport = ntohs( one->extPort );
		udp->uh_dport = ntohs( one->remPort );
		udp->uh_sum = 0x5566; 

		
		memcpy( rtp, packet, pktLen );

		ret = rtl8651_offloadEngineL3Send( pkt, pktLen+sizeof(struct ip)+sizeof(struct udphdr), 
		                                   OFFLOAD_IP_CKSUM | OFFLOAD_L4_CKSUM, 0 );
		if ( ret != SUCCESS )
		{
#ifdef _VOIP_DEBUG_
			rtlglue_printf( "%s():%d return %d.\n\n", 
			                __FUNCTION__, __LINE__, ret );
#endif
			goto out;
		}                                     
	}
	else
	{
		
		ret = FAILED;
		goto out;
	}

	ret = SUCCESS;
out:
	return ret;
}


