#ifndef RTONCE_HAL_CACHE_H
#define RTONCE_HAL_CACHE_H

//=============================================================================
//
//      hal_cache.h
//
//      HAL cache control API
//
//=============================================================================
//
// Purpose:     Cache control API
// Description: The macros defined here provide the HAL APIs for handling
//              cache control operations.
// Usage:
//              #include <rt/hal/hal_cache.h>
//
//=============================================================================

/*
#include <include/pkgconf/hal.h>
#include <include/infra/rt_type.h>

#include <include/hal/var_cache.h>
*/

#include "../pkgconf/hal.h"
#include "../infra/rt_type.h"
#include "var_cache.h"

// Use this macro to allow the assembler to accept "cache" instructions,
// which are MIPS ISA 3. This is useful if someone is compiling
// with -mips2, but the architecture is really MIPS ISA 3.

/*
// Since LX4180 support only mips1 architecure, we don't need to switch
// ISA level anywhere. removed by xavier 11/13/2000
//
#define _hal_asm_mips_cpp_stringize( _x_ ) #_x_
#define _HAL_ASM_SET_MIPS_ISA( _isal_ ) asm volatile ( \
      ".set mips" _hal_asm_mips_cpp_stringize(_isal_) )
*/

//=============================================================================
// Default Implementation. This uses the standard MIPS CP0 registers and
// cache instructions. Note that not all variants will have all of the
// functionality defined here. 

//-----------------------------------------------------------------------------
// Cache dimensions.
// These really should be defined in var_cache.h. If they are not, then provide
// a set of numbers that are typical of many variants.

#ifndef HAL_DCACHE_SIZE

// Data cache
#define HAL_DCACHE_SIZE                 4096    // Size of data cache in bytes
#define HAL_DCACHE_LINE_SIZE            16      // Size of a data cache line
#define HAL_DCACHE_WAYS                 1       // Associativity of the cache

// Instruction cache
#define HAL_ICACHE_SIZE                 4096    // Size of cache in bytes
#define HAL_ICACHE_LINE_SIZE            16      // Size of a cache line
#define HAL_ICACHE_WAYS                 1       // Associativity of the cache

#define HAL_DCACHE_SETS (HAL_DCACHE_SIZE/(HAL_DCACHE_LINE_SIZE*HAL_DCACHE_WAYS))
#define HAL_ICACHE_SETS (HAL_ICACHE_SIZE/(HAL_ICACHE_LINE_SIZE*HAL_ICACHE_WAYS))

#endif

//-----------------------------------------------------------------------------
// Cache instruction uses LSBs or MSBs (depending on the
// implementation) of the virtual address to specify which WAY to
// affect. The _ALL_WAYS macro defines the necessary cache instructions
// to affect all ways.

#ifdef HAL_MIPS_CACHE_INSN_USES_LSB
# define _IWAY(_n_) ((_n_)*HAL_ICACHE_SIZE/HAL_ICACHE_WAYS+(_n_))
# define _DWAY(_n_) ((_n_)*HAL_DCACHE_SIZE/HAL_DCACHE_WAYS+(_n_))
#else
# define _IWAY(_n_) ((_n_)*HAL_ICACHE_SIZE/HAL_ICACHE_WAYS)
# define _DWAY(_n_) ((_n_)*HAL_DCACHE_SIZE/HAL_DCACHE_WAYS)
#endif

#if (HAL_DCACHE_WAYS == 1)
#define _HAL_ASM_DCACHE_ALL_WAYS( _cmd_ , _addr_ )       \
    asm volatile ("cache %0,0(%1);"                      \
                    : : "I" ((_cmd_) | 1), "r"(_addr_) )
#elif (HAL_DCACHE_WAYS == 2)
#define _HAL_ASM_DCACHE_ALL_WAYS( _cmd_ , _addr_ )      \
    asm volatile ("cache %0,0(%1);"                     \
                  "cache %0,%2(%1);"                    \
                    : : "I" ((_cmd_) | 1), "r"(_addr_), \
                        "I" (_DWAY(1)))
#elif (HAL_DCACHE_WAYS == 4)
#define _HAL_ASM_DCACHE_ALL_WAYS( _cmd_ , _addr_ )      \
    asm volatile ("cache %0,0(%1);"                     \
                  "cache %0,%2(%1);"                    \
                  "cache %0,%3(%1);"                    \
                  "cache %0,%4(%1);"                    \
                    : : "I" ((_cmd_) | 1), "r"(_addr_), \
                        "I" (_DWAY(1)),                  \
                        "I" (_DWAY(2)),                  \
                        "I" (_DWAY(3)))
#else
# error "Unsupported number of ways"
#endif

#if (HAL_ICACHE_WAYS == 1)
#define _HAL_ASM_ICACHE_ALL_WAYS( _cmd_ , _addr_ )       \
    asm volatile ("cache %0,0(%1);"                      \
                    : : "I" (_cmd_), "r"(_addr_) )
#elif (HAL_ICACHE_WAYS == 2)
#define _HAL_ASM_ICACHE_ALL_WAYS( _cmd_ , _addr_ )      \
    asm volatile ("cache %0,0(%1);"                     \
                  "cache %0,%2(%1);"                    \
                    : : "I" (_cmd_), "r"(_addr_),       \
                        "I" (_IWAY(1)))
#elif (HAL_ICACHE_WAYS == 4)
#define _HAL_ASM_ICACHE_ALL_WAYS( _cmd_ , _addr_ )      \
    asm volatile ("cache %0,0(%1);"                     \
                  "cache %0,%2(%1);"                    \
                  "cache %0,%3(%1);"                    \
                  "cache %0,%4(%1);"                    \
                    : : "I" (_cmd_), "r"(_addr_),       \
                        "I" (_IWAY(1)),                  \
                        "I" (_IWAY(2)),                  \
                        "I" (_IWAY(3)))
#else
# error "Unsupported number of ways"
#endif

//-----------------------------------------------------------------------------
// Global control of data cache

// Enable the data cache
// There is no default mechanism for enabling or disabling the caches.
#ifndef HAL_DCACHE_ENABLE_DEFINED
#define HAL_DCACHE_ENABLE()
#endif

// Disable the data cache
#ifndef HAL_DCACHE_DISABLE_DEFINED
#define HAL_DCACHE_DISABLE()
#endif

#ifndef HAL_DCACHE_IS_ENABLED_DEFINED
#define HAL_DCACHE_IS_ENABLED(_state_) (_state_) = 1;
#endif

// Invalidate the entire cache
// We simply use HAL_DCACHE_SYNC() to do this. For writeback caches this
// is not quite what we want, but there is no index-invalidate operation
// available.
#ifndef HAL_DCACHE_INVALIDATE_ALL_DEFINED
#define HAL_DCACHE_INVALIDATE_ALL() HAL_DCACHE_SYNC()
#endif

// Synchronize the contents of the cache with memory.
// This uses the index-writeback-invalidate operation.
#ifndef HAL_DCACHE_SYNC_DEFINED
/*
#define HAL_DCACHE_SYNC()                                               \
{									\
	asm volatile					\
	(								\
		"ori	$8, $0, 1;"			\
		".set	noreorder;"			\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$8, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		".set	reorder;"			\
		:							\
		:							\
		: "$8"						\
	);								\
}
*/

#ifdef	__COMPILER_MIPS16

#define HAL_DCACHE_SYNC()                                               \
{									\
	asm volatile					\
	(								\
		"li		$4, 1;"				\
		"addiu	$29,-24;"\
		"sw		$31,20($29);" \
		"sw		$4,16($29);" \
		"jalx	____hal_cache_invalidate_all;"	\
		"lw		$4,16($29);" \
		"addiu	$29,24;"\
		:							\
		:							\
		: "$4","$31"						\
	);								\
}
#else
#define HAL_DCACHE_SYNC()                                               \
{									\
	asm volatile					\
	(								\
		"ori	$8, $0, 1;"			\
		".set	noreorder;"			\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$8, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		".set	reorder;"			\
		:							\
		:							\
		: "$8"						\
	);								\
}
#endif

/*
{	    RT_MACRO_START                                                     \
    register RT_ADDRESS _baddr_ = 0x80000000;                          \
    register RT_ADDRESS _addr_ = 0x80000000;                           \
    register RT_WORD _size_ = HAL_DCACHE_SIZE;                         \
    _HAL_ASM_SET_MIPS_ISA(3);                                           \
    for( ; _addr_ <= _baddr_+_size_; _addr_ += HAL_DCACHE_LINE_SIZE )   \
    { _HAL_ASM_DCACHE_ALL_WAYS(0x01, _addr_); }                         \
    _HAL_ASM_SET_MIPS_ISA(0);                                           \
    RT_MACRO_END
*/
#endif

// Set the data cache refill burst size
//#define HAL_DCACHE_BURST_SIZE(_size_)

// Set the data cache write mode
//#define HAL_DCACHE_WRITE_MODE( _mode_ )

//#define HAL_DCACHE_WRITETHRU_MODE       0
//#define HAL_DCACHE_WRITEBACK_MODE       1

// Load the contents of the given address range into the data cache
// and then lock the cache so that it stays there.
// This uses the fetch-and-lock cache operation.
/*
#ifndef HAL_DCACHE_LOCK_DEFINED
#define HAL_DCACHE_LOCK(_base_, _asize_)                                  \
    RT_MACRO_START                                                       \
    register RT_ADDRESS _baddr_ = (RT_ADDRESS)(_base_);                 \
    register RT_ADDRESS _addr_ = (RT_ADDRESS)(_base_);                  \
    register RT_WORD _size_ = (_asize_);                                 \
    register RT_WORD _state_;                                            \
    HAL_DCACHE_IS_ENABLED( _state_ );                                     \
    if( _state_ ) {                                                       \
        _HAL_ASM_SET_MIPS_ISA(3);                                         \
        for( ; _addr_ <= _baddr_+_size_; _addr_ += HAL_DCACHE_LINE_SIZE ) \
        { _HAL_ASM_DCACHE_ALL_WAYS(0x1d, _addr_); }                       \
        _HAL_ASM_SET_MIPS_ISA(0);                                         \
    }                                                                     \
    RT_MACRO_END
#endif
*/
// Undo a previous lock operation.
// Do this by flushing the cache, which is defined to clear the lock bit.
/*
#ifndef HAL_DCACHE_UNLOCK_DEFINED
#define HAL_DCACHE_UNLOCK(_base_, _size_) \
        HAL_DCACHE_FLUSH( _base_, _size_ )
#endif
*/

// Unlock entire cache
/*
#ifndef HAL_DCACHE_UNLOCK_ALL_DEFINED
#define HAL_DCACHE_UNLOCK_ALL() \
        HAL_DCACHE_INVALIDATE_ALL()
#endif
*/

//-----------------------------------------------------------------------------
// Data cache line control

// Allocate cache lines for the given address range without reading its
// contents from memory.
//#define HAL_DCACHE_ALLOCATE( _base_ , _size_ )

// Write dirty cache lines to memory and invalidate the cache entries
// for the given address range.
// This uses the hit-writeback-invalidate cache operation.
#ifndef HAL_DCACHE_FLUSH_DEFINED
#define HAL_DCACHE_FLUSH( _base_ , _asize_ )                              \
    RT_MACRO_START                                                       \
    register RT_ADDRESS _baddr_ = (RT_ADDRESS)(_base_);                 \
    register RT_ADDRESS _addr_ = (RT_ADDRESS)(_base_);                  \
    register RT_WORD _size_ = (_asize_);                                 \
    register RT_WORD _state_;                                            \
    register RT_WORD _ori_c0_sr_;												\
    HAL_DCACHE_IS_ENABLED( _state_ );                                     \
    if( _state_ )																\
    {																			\
		asm volatile															\
		(																		\
			"lui	$13, 0x1;"													\
			".set	noreorder;"													\
			"mfc0	%0, $12;"													\
			"nop;"																\
			"or		$9, %0, $13;"												\
			"mtc0	$9, $12;"													\
			".set	reorder;"													\
			: "=r"(_ori_c0_sr_)													\
			:																	\
			: "$9", "$13"														\
		);																		\
        for( ; _addr_ <= _baddr_+_size_; _addr_ += HAL_DCACHE_LINE_SIZE ) \
        {																		\
        	asm volatile ("sb $0, 0(%0)" : : "r"(_addr_) );						\
        }																		\
		asm volatile ("mtc0 %0, $12" : : "r"(_ori_c0_sr_) );					\
    }                                                                     \
    RT_MACRO_END
#endif

// Invalidate cache lines in the given range without writing to memory.
// This uses the hit-invalidate cache operation.
//
// There is no support for this kind of operation in LX4180,
// so we just flush it.
// - xavier
//
#ifndef HAL_DCACHE_INVALIDATE_DEFINED
#define HAL_DCACHE_INVALIDATE( _base_ , _asize_ )                       \
RT_MACRO_START																	\
    register RT_ADDRESS _baddr_ = (RT_ADDRESS)(_base_);						\
    register RT_ADDRESS _addr_ = (RT_ADDRESS)(_base_);                        \
    register RT_WORD _size_ = (_asize_);                                       \
    register RT_WORD _state_;                                                  \
    register RT_WORD _ori_c0_sr_;												\
    HAL_DCACHE_IS_ENABLED( _state_ );                                           \
    if( _state_ )																\
    {																			\
		asm volatile															\
		(																		\
			"lui	$13, 0x1;"													\
			".set	noreorder;"													\
			"mfc0	%0, $12;"													\
			"nop;"																\
			"or		$9, %0, $13;"												\
			"mtc0	$9, $12;"													\
			".set	reorder;"													\
			: "=r"(_ori_c0_sr_)													\
			:																	\
			: "$9", "$13"														\
		);																		\
        for( ; _addr_ <= _baddr_+_size_; _addr_ += HAL_DCACHE_LINE_SIZE )       \
        {																		\
        	asm volatile ("sb $0, 0(%0)" : : "r"(_addr_) );						\
        }																		\
		asm volatile ("mtc0 %0, $12" : : "r"(_ori_c0_sr_) );					\
    }                                                                           \
RT_MACRO_END
/*
    RT_MACRO_START                                                     \
    register RT_ADDRESS _baddr_ = (RT_ADDRESS)(_base_);               \
    register RT_ADDRESS _addr_ = (RT_ADDRESS)(_base_);                \
    register RT_WORD _size_ = (_asize_);                               \
    _HAL_ASM_SET_MIPS_ISA(3);                                           \
    for( ; _addr_ <= _baddr_+_size_; _addr_ += HAL_DCACHE_LINE_SIZE )   \
    { _HAL_ASM_DCACHE_ALL_WAYS(0x11, _addr_); }                         \
    _HAL_ASM_SET_MIPS_ISA(0);                                           \
    RT_MACRO_END
*/
#endif

// Write dirty cache lines to memory for the given address range.
// This uses the hit-writeback cache operation.
//
// There is no support for this kind of operation in LX4180,
// so we just flush it.
// - xavier
//
#ifndef HAL_DCACHE_STORE_DEFINED
#define HAL_DCACHE_STORE( _base_ , _asize_ )                              \
		HAL_DCACHE_FLUSH( _base_, _asize_ )
/*
    RT_MACRO_START                                                       \
    register RT_ADDRESS _baddr_ = (RT_ADDRESS)(_base_);                 \
    register RT_ADDRESS _addr_ = (RT_ADDRESS)(_base_);                  \
    register RT_WORD _size_ = (_asize_);                                 \
    register RT_WORD _state_;                                            \
    HAL_DCACHE_IS_ENABLED( _state_ );                                     \
    if( _state_ ) {                                                       \
        _HAL_ASM_SET_MIPS_ISA(3);                                         \
        for( ; _addr_ <= _baddr_+_size_; _addr_ += HAL_DCACHE_LINE_SIZE ) \
        { _HAL_ASM_DCACHE_ALL_WAYS(0x19, _addr_); }                       \
        _HAL_ASM_SET_MIPS_ISA(0);                                         \
    }                                                                     \
    RT_MACRO_END
*/
#endif

// Preread the given range into the cache with the intention of reading
// from it later.
//#define HAL_DCACHE_READ_HINT( _base_ , _size_ )

// Preread the given range into the cache with the intention of writing
// to it later.
//#define HAL_DCACHE_WRITE_HINT( _base_ , _size_ )

// Allocate and zero the cache lines associated with the given range.
//#define HAL_DCACHE_ZERO( _base_ , _size_ )

//-----------------------------------------------------------------------------
// Global control of Instruction cache

// Enable the instruction cache
// There is no default mechanism for enabling or disabling the caches.
#ifndef HAL_ICACHE_ENABLE_DEFINED
#define HAL_ICACHE_ENABLE()
#endif

// Disable the instruction cache
#ifndef HAL_ICACHE_DISABLE_DEFINED
#define HAL_ICACHE_DISABLE()
#endif

#ifndef HAL_ICACHE_IS_ENABLED_DEFINED
#define HAL_ICACHE_IS_ENABLED(_state_) (_state_) = 1;
#endif

// Invalidate the entire cache
// This uses the index-invalidate cache operation.
#ifndef HAL_ICACHE_INVALIDATE_ALL_DEFINED

#if (CROSS_COMPILER == COMPILER_GNU_GCC)
/*
#define HAL_ICACHE_INVALIDATE_ALL() \
    RT_MACRO_START                 \
	asm volatile					\
	(								\
		"ori	$8, $0, 2;"			\
		".set	noreorder;"			\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$8, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		".set	reorder;"			\
		:							\
		:							\
		: "$8"						\
	);								\
    RT_MACRO_END
*/

#ifdef	__COMPILER_MIPS16
#define HAL_ICACHE_INVALIDATE_ALL() \
    RT_MACRO_START              \
	asm volatile				\
	(							\
		"addiu	$29,-24;"		\
		"sw		$31,20($29);"	\
		"sw		$4,16($29);"	\
		"li		$4,2;"			\
		"jalx ____hal_cache_invalidate_all;nop;" \
		"lw		$4,16($29);"	\
		"addiu	$29,24;"		\
		".set	reorder;"		\
		:						\
		:						\
		: "$4","$31"			\
	);							\
    RT_MACRO_END
#else
#define HAL_ICACHE_INVALIDATE_ALL() \
    RT_MACRO_START                 \
	asm volatile					\
	(								\
		"ori	$8, $0, 2;"			\
		".set	noreorder;"			\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$8, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		".set	reorder;"			\
		:							\
		:							\
		: "$8"						\
	);								\
    RT_MACRO_END
#endif

#endif

#if (CROSS_COMPILER == COMPILER_GREEN_HILLS)
asm void HAL_ICACHE_INVALIDATE_ALL(void)
{
	ori		$8,$0,2
	.set	noreorder
	mtc0	$0,$20
	nop
	nop
	mtc0	$8,$20
	nop
	nop
	mtc0	$0,$20
	nop
	nop
	.set	reorder
}
#endif

#endif

// Synchronize the contents of the cache with memory.
// Simply force the cache to reload.
#ifndef HAL_ICACHE_SYNC_DEFINED
#define HAL_ICACHE_SYNC() HAL_ICACHE_INVALIDATE_ALL()
#endif

// Set the instruction cache refill burst size
//#define HAL_ICACHE_BURST_SIZE(_size_)

// Load the contents of the given address range into the instruction cache
// and then lock the cache so that it stays there.
// This uses the fetch-and-lock cache operation.

// There is no support for this kind of operation in LX4180,
// - xavier
//
/*
#ifndef HAL_ICACHE_LOCK_DEFINED
#define HAL_ICACHE_LOCK(_base_, _asize_)                                  \
    RT_MACRO_START                                                       \
    register RT_ADDRESS _baddr_ = (RT_ADDRESS)(_base_);                 \
    register RT_ADDRESS _addr_ = (RT_ADDRESS)(_base_);                  \
    register RT_WORD _size_ = (_asize_);                                 \
    register RT_WORD _state_;                                            \
    HAL_ICACHE_IS_ENABLED( _state_ );                                     \
    if( _state_ ) {                                                       \
        _HAL_ASM_SET_MIPS_ISA(3);                                         \
        for( ; _addr_ <= _baddr_+_size_; _addr_ += HAL_DCACHE_LINE_SIZE ) \
        { _HAL_ASM_ICACHE_ALL_WAYS(0x1c, _addr_); }                       \
        _HAL_ASM_SET_MIPS_ISA(0);                                         \
    }                                                                     \
    RT_MACRO_END
#endif
*/

// Undo a previous lock operation.
// Do this by invalidating the cache, which is defined to clear the lock bit.
/*
#ifndef HAL_ICACHE_UNLOCK_DEFINED
#define HAL_ICACHE_UNLOCK(_base_, _size_) \
        HAL_ICACHE_INVALIDATE( _base_, _size_ )
#endif
*/

// Unlock entire cache
//#define HAL_ICACHE_UNLOCK_ALL()

//-----------------------------------------------------------------------------
// Instruction cache line control

// Invalidate cache lines in the given range without writing to memory.
// This uses the hit-invalidate cache operation.
#ifndef HAL_ICACHE_INVALIDATE_DEFINED
#define HAL_ICACHE_INVALIDATE( _base_ , _asize_ )                       \
    RT_MACRO_START                                                     \
    register RT_ADDRESS _baddr_ = (RT_ADDRESS)(_base_);               \
    register RT_ADDRESS _addr_ = (RT_ADDRESS)(_base_);                \
    register RT_WORD _size_ = (_asize_);                               \
    register RT_WORD _ori_c0_sr_;										\
	asm volatile														\
	(																	\
		"lui	$13, 0x3;"												\
		".set	noreorder;"												\
		"mfc0	%0, $12;"												\
		"nop;"															\
		"or		$9, %0, $13;"											\
		"mtc0	$9, $12;"												\
		".set	reorder;"												\
		: "=r"(_ori_c0_sr_)												\
		:																\
		: "$9", "$13"													\
	);																	\
    for( ; _addr_ <= _baddr_+_size_; _addr_ += HAL_ICACHE_LINE_SIZE )   \
    {																	\
    	asm volatile ("sb $0, 0(%0)" : : "r"(_addr_) );					\
    }																	\
	asm volatile ("mtc0 %0, $12" : : "r"(_ori_c0_sr_) );				\
    RT_MACRO_END
#endif

//-----------------------------------------------------------------------------
// Check that a supported configuration has actually defined some macros.

#ifndef HAL_DCACHE_ENABLE

#error Unsupported MIPS configuration

#endif

//-----------------------------------------------------------------------------
#endif // ifndef RTONCE_HAL_CACHE_H
// End of hal_cache.h
