#ifndef RTONCE_HAL_HAL_INTR_H
#define RTONCE_HAL_HAL_INTR_H

//==========================================================================
//
//      hal_intr.h
//
//      HAL Interrupt and clock support
//
//==========================================================================

/*
#include <include\pkgconf\hal.h>

#include <include\infra\rt_type.h>
#include <include\hal\hal_io.h>

#include <include\hal\var_intr.h>
*/

#include "../pkgconf/hal.h"
#include "../infra/rt_type.h"
#include "hal_io.h"
#include "var_intr.h"


//--------------------------------------------------------------------------
// MIPS vectors.

// These are the exception codes presented in the Cause register and
// correspond to VSRs. These values are the ones to use for HAL_VSR_GET/SET

// External interrupt
#define RTNUM_HAL_VECTOR_INTERRUPT            0
// TLB modification exception
#define RTNUM_HAL_VECTOR_TLB_MOD              1
// TLB miss (Load or IFetch)
#define RTNUM_HAL_VECTOR_TLB_LOAD_REFILL      2
// TLB miss (Store)
#define RTNUM_HAL_VECTOR_TLB_STORE_REFILL     3
// Address error (Load or Ifetch)
#define RTNUM_HAL_VECTOR_LOAD_ADDRESS         4
// Address error (store)
#define RTNUM_HAL_VECTOR_STORE_ADDRESS        5
// Bus error (Ifetch)
#define RTNUM_HAL_VECTOR_IBE                  6
// Bus error (data load or store)
#define RTNUM_HAL_VECTOR_DBE                  7
// System call
#define RTNUM_HAL_VECTOR_SYSTEM_CALL          8
// Break point
#define RTNUM_HAL_VECTOR_BREAKPOINT           9
// Reserved instruction
#define RTNUM_HAL_VECTOR_RESERVED_INSTRUCTION 10
// Coprocessor unusable
#define RTNUM_HAL_VECTOR_COPROCESSOR          11
// Arithmetic overflow
#define RTNUM_HAL_VECTOR_OVERFLOW             12
// Reserved
#define RTNUM_HAL_VECTOR_RESERVED_13          13
// Division-by-zero [reserved vector]
// This is caused by 'trap 0x7' which GCC puts in the code to check
// for division by zero. The break_vsr_springboard in vectors.S is the
// only caller of this vector.
#define RTNUM_HAL_VECTOR_DIV_BY_ZERO          14
// Floating point exception
#ifdef  RTHWR_HAL_MIPS_FPU
#define RTNUM_HAL_VECTOR_FPE                  15
#endif

#define RTNUM_HAL_VSR_MIN                     0
#define RTNUM_HAL_VSR_MAX                     15
#define RTNUM_HAL_VSR_COUNT                   16

// Exception vectors. These are the values used when passed out to an
// external exception handler using rt_hal_deliver_exception()

#define RTNUM_HAL_EXCEPTION_DATA_TLBERROR_ACCESS RTNUM_HAL_VECTOR_TLB_MOD
#define RTNUM_HAL_EXCEPTION_DATA_TLBMISS_ACCESS \
          RTNUM_HAL_VECTOR_TLB_LOAD_REFILL
#define RTNUM_HAL_EXCEPTION_DATA_TLBMISS_WRITE \
          RTNUM_HAL_VECTOR_TLB_STORE_REFILL
#define RTNUM_HAL_EXCEPTION_DATA_UNALIGNED_ACCESS \
          RTNUM_HAL_VECTOR_LOAD_ADDRESS
#define RTNUM_HAL_EXCEPTION_DATA_UNALIGNED_WRITE \
          RTNUM_HAL_VECTOR_STORE_ADDRESS
#define RTNUM_HAL_EXCEPTION_CODE_ACCESS    RTNUM_HAL_VECTOR_IBE
#define RTNUM_HAL_EXCEPTION_DATA_ACCESS    RTNUM_HAL_VECTOR_DBE
#define RTNUM_HAL_EXCEPTION_SYSTEM_CALL    RTNUM_HAL_VECTOR_SYSTEM_CALL
#define RTNUM_HAL_EXCEPTION_INSTRUCTION_BP RTNUM_HAL_VECTOR_BREAKPOINT
#define RTNUM_HAL_EXCEPTION_ILLEGAL_INSTRUCTION \
          RTNUM_HAL_VECTOR_RESERVED_INSTRUCTION
#define RTNUM_HAL_EXCEPTION_COPROCESSOR    RTNUM_HAL_VECTOR_COPROCESSOR
#define RTNUM_HAL_EXCEPTION_OVERFLOW       RTNUM_HAL_VECTOR_OVERFLOW
#define RTNUM_HAL_EXCEPTION_DIV_BY_ZERO    RTNUM_HAL_VECTOR_DIV_BY_ZERO
#ifdef  RTHWR_HAL_MIPS_FPU
#define RTNUM_HAL_EXCEPTION_FPU            RTNUM_HAL_VECTOR_FPE
#endif

#define RTNUM_HAL_EXCEPTION_INTERRUPT      RTNUM_HAL_VECTOR_BREAKPOINT

// decoded exception vectors
#define RTNUM_HAL_EXCEPTION_FPU_INEXACT     (-1)
#define RTNUM_HAL_EXCEPTION_FPU_DIV_BY_ZERO (-2)
#define RTNUM_HAL_EXCEPTION_FPU_OVERFLOW    (-3)
#define RTNUM_HAL_EXCEPTION_FPU_UNDERFLOW   (-4)
#define RTNUM_HAL_EXCEPTION_FPU_INVALID     (-5)


// Min/Max exception numbers and how many there are
#define RTNUM_HAL_EXCEPTION_MIN                RTNUM_HAL_EXCEPTION_FPU_INVALID
#define RTNUM_HAL_EXCEPTION_MAX                RTNUM_HAL_VSR_MAX

#define RTNUM_HAL_EXCEPTION_COUNT           \
                 ( RTNUM_HAL_EXCEPTION_MAX - RTNUM_HAL_EXCEPTION_MIN + 1 )


#ifndef RTHWR_HAL_INTERRUPT_VECTORS_DEFINED

// the default for all MIPS variants is to use the 6 bits
// in the cause register.

#define RTNUM_HAL_INTERRUPT_0                0
#define RTNUM_HAL_INTERRUPT_1                1
#define RTNUM_HAL_INTERRUPT_2                2
#define RTNUM_HAL_INTERRUPT_3                3
#define RTNUM_HAL_INTERRUPT_4                4
#define RTNUM_HAL_INTERRUPT_5                5

// Min/Max ISR numbers and how many there are
#define RTNUM_HAL_ISR_MIN                     0
#define RTNUM_HAL_ISR_MAX                     5
#define RTNUM_HAL_ISR_COUNT                   6

// The vector used by the Real time clock. The default here is to use
// interrupt 5, which is connected to the counter/comparator registers
// in many MIPS variants.

#ifndef RTNUM_HAL_INTERRUPT_RTC
#define RTNUM_HAL_INTERRUPT_RTC            RTNUM_HAL_INTERRUPT_1
#endif

#define RTHWR_HAL_INTERRUPT_VECTORS_DEFINED

#endif

//--------------------------------------------------------------------------
// Static data used by HAL

// ISR tables
externC volatile RT_ADDRESS    hal_interrupt_handlers[RTNUM_HAL_ISR_COUNT];
externC volatile RT_ADDRWORD   hal_interrupt_data[RTNUM_HAL_ISR_COUNT];
externC volatile RT_ADDRESS    hal_interrupt_objects[RTNUM_HAL_ISR_COUNT];

// VSR table
externC volatile RT_ADDRESS    hal_vsr_table[RTNUM_HAL_VSR_MAX+1];

//--------------------------------------------------------------------------
// Default ISR
// The #define is used to test whether this routine exists, and to allow
// us to call it.

externC rt_uint32 hal_default_isr(RT_ADDRWORD vector, RT_ADDRWORD data);

#define HAL_DEFAULT_ISR hal_default_isr

//--------------------------------------------------------------------------
// Interrupt state storage

typedef rt_uint32 RT_INTERRUPT_STATE;



#if (CROSS_COMPILER == COMPILER_GNU_GCC)
//--------------------------------------------------------------------------
// Interrupt control macros
// Beware of nops in this code. They fill delay slots and avoid CP0 hazards
// that might otherwise cause following code to run in the wrong state or
// cause a resource conflict.
// Note that _old_ need to be declared as a "register" variable before using
// any of the following macros
/*
#define HAL_DISABLE_INTERRUPTS(_old_)           \
{                                               \
    asm volatile (                              \
        "mfc0   $8,$12; nop;"                   \
        "move   %0,$8;"                         \
        "and    $8,$8,0XFFFFFFFE;"              \
        "mtc0   $8,$12;"                        \
        "nop; nop; nop;"                        \
        "and    %0,%0,0X1;"                     \
        : "=r"(_old_)                           \
        :                                       \
        : "$8"                                  \
        );                                      \
}
*/

#ifdef	__COMPILER_MIPS16
#define HAL_DISABLE_INTERRUPTS(_old_)	\
{	\
	asm volatile ( \
		"addiu	$29,-24;"\
		"sw		$31,20($29);" \
		"sw		$4,16($29);" \
		"sw		$5,12($29);" \
		"jalx ____hal_disable_interrupts;nop;"	\
		"move %0,$4;" \
		"lw		$4,16($29);" \
		"lw		$5,12($29);" \
		"addiu	$29,24;"\
		: "=r"(_old_) \
		:	\
		: "$4","$5","$31"	\
		);\
}
#else
#define HAL_DISABLE_INTERRUPTS(_old_)           \
{                                               \
    asm volatile (                              \
        "mfc0   $8,$12; nop;"                   \
        "move   %0,$8;"                         \
        "and    $8,$8,0XFFFFFFFE;"              \
        "mtc0   $8,$12;"                        \
        "nop; nop; nop;"                        \
        "and    %0,%0,0X1;"                     \
        : "=r"(_old_)                           \
        :                                       \
        : "$8"                                  \
        );                                      \
}
#endif


/*

#define HAL_DISABLE_INTERRUPT()			        \
{                                               \
    asm volatile (                              \
        "mfc0   $8,$12; nop;"                   \
        "and    $8,$8,0XFFFFFFFE;"              \
        "mtc0   $8,$12;"                        \
        "nop; nop; nop;"                        \
        :			                            \
        :                                       \
        : "$8"                                  \
        );                                      \
}

*/


#ifdef	__COMPILER_MIPS16
#define HAL_DISABLE_INTERRUPT()			        \
{                                               \
    asm volatile (                              \
		"addiu	$29,-24;"\
		"sw		$31,20($29);" \
		"sw		$4,16($29);" \
        "jalx ____hal_disable_interrupt;nop;"	\
		"lw		$31,20($29);" \
		"lw		$4,16($29);" \
		"addiu	$29,24;"\
        :			                            \
        :                                       \
        : "$4","$31"                                  \
        );                                      \
}
#else
#define HAL_DISABLE_INTERRUPT()			        \
{                                               \
    asm volatile (                              \
        "mfc0   $8,$12; nop;"                   \
        "and    $8,$8,0XFFFFFFFE;"              \
        "mtc0   $8,$12;"                        \
        "nop; nop; nop;"                        \
        :			                            \
        :                                       \
        : "$8"                                  \
        );                                      \
}
#endif

/*
#define HAL_ENABLE_INTERRUPTS()                 \
{                                               \
    asm volatile (                              \
        "mfc0   $8,$12; nop;"                   \
        "or     $8,$8,1;"                       \
        "mtc0   $8,$12;"                        \
        "nop; nop; nop;"                        \
        :                                       \
        :                                       \
        : "$8"                                  \
        );                                      \
}
*/

#ifdef	__COMPILER_MIPS16
#define HAL_ENABLE_INTERRUPT()			        \
{                                               \
    asm volatile (                              \
		"addiu	$29,-24;"\
		"sw		$31,20($29);" \
		"sw		$4,16($29);" \
        "jalx ____hal_enable_interrupts;nop;"	\
		"lw		$31,20($29);" \
		"lw		$4,16($29);" \
		"addiu	$29,24;"\
        :			                            \
        :                                       \
        : "$4","$31"                                  \
        );                                      \
}
#else
#define HAL_ENABLE_INTERRUPTS()                 \
{                                               \
    asm volatile (                              \
        "mfc0   $8,$12; nop;"                   \
        "or     $8,$8,1;"                       \
        "mtc0   $8,$12;"                        \
        "nop; nop; nop;"                        \
        :                                       \
        :                                       \
        : "$8"                                  \
        );                                      \
}
#endif

/*
#define HAL_RESTORE_INTERRUPTS(_old_)           \
{                                               \
    asm volatile (                              \
        "mfc0   $8,$12;"		                \
        "and    %0,%0,0x1;"                     \
        "and    $8,$8,0XFFFFFFFE;"              \
        "or     $8,$8,%0;"                      \
        "mtc0   $8,$12;"                        \
        "nop; nop; nop;"                        \
        :                                       \
        : "r"(_old_)                            \
        : "$8"                                  \
        );                                      \
}
*/

#ifdef	__COMPILER_MIPS16
#define HAL_RESTORE_INTERRUPTS(_old_)			        \
{                   \
    asm volatile ( \
		"addiu	$29,-24;"\
		"sw		$31,20($29);" \
		"sw		$4,16($29);" \
		"sw		$5,12($29);" \
		"move $4,%0;"	\
		"jalx ____hal_restore_interrupts;nop;"	\
		"lw		$4,16($29);" \
		"lw		$5,12($29);" \
		"addiu	$29,24;"\
		:	\
		: "r"(_old_) \
		: "$31","$4","$5"	\
		);\
}
#else
#define HAL_RESTORE_INTERRUPTS(_old_)           \
{                                               \
    asm volatile (                              \
        "mfc0   $8,$12;"		                \
        "and    %0,%0,0x1;"                     \
        "and    $8,$8,0XFFFFFFFE;"              \
        "or     $8,$8,%0;"                      \
        "mtc0   $8,$12;"                        \
        "nop; nop; nop;"                        \
        :                                       \
        : "r"(_old_)                            \
        : "$8"                                  \
        );                                      \
}
#endif


/*
#define HAL_QUERY_INTERRUPTS( _state_ )         \
{                                               \
    asm volatile (                              \
        "mfc0   %0,$12; nop;"                   \
        "and    %0,%0,0x1;"                     \
        : "=r"(_state_)                         \
        :                                       \
        : "$8"                                  \
        );                                      \
}
*/
#ifdef	__COMPILER_MIPS16
#define HAL_QUERY_INTERRUPTS( _state_ )			\
{                                               \
    asm volatile (                              \
		"addiu	$29,-24;"\
		"sw		$31,20($29);" \
		"sw		$4,16($29);" \
		"jalx ____hal_query_interrupts;nop;"	\
		"move %0,$4;"	\
		"lw		$31,20($29);" \
		"lw		$4,16($29);" \
		"addiu	$29,24;"\
        : "=r"(_state_)                         \
        :                                       \
        : "$4","$31"                                  \
        );                                      \
}
#else
#define HAL_QUERY_INTERRUPTS( _state_ )         \
{                                               \
    asm volatile (                              \
        "mfc0   %0,$12; nop;"                   \
        "and    %0,%0,0x1;"                     \
        : "=r"(_state_)                         \
        :                                       \
        : "$8"                                  \
        );                                      \
}
#endif

#endif


#if (CROSS_COMPILER == COMPILER_GREEN_HILLS)
/*
 *	Green Hills compiler can't recognize all the inline assembly syntax of GNU.
 *	In fact, it also defines its own inline assembly macro facility.
 */

asm void HAL_DISABLE_INTERRUPTS(_old_)
{
%reg _old_
; # _old_ is declared as a "register" variable
	mfc0	$8,$12
	nop
	move	_old_,$8
	and		$8,$8,0XFFFFFFFE
	mtc0	$8,$12
	nop
	nop
	nop
	and		_old_,_old_,0X1
%mem _old_
; # _old_ is declared as a "memory" variable
	mfc0	$8,$12	
	nop
	move	$9,$8
	and		$8,$8,0XFFFFFFFE
	mtc0	$8,$12
	nop
	nop
	nop
	and		$9,$9,0X1
	sw		$9,_old_
	nop
%error
}

asm void HAL_DISABLE_INTERRUPT(void)
{
	mfc0	$8,$12
	nop
	and		$8,$8,0XFFFFFFFE
	mtc0	$8,$12
	nop
	nop
	nop
}

asm void HAL_ENABLE_INTERRUPTS(void)
{
	mfc0	$8,$12
	nop
	or		$8,$8,1
	mtc0	$8,$12
	nop
	nop
	nop
}

asm void HAL_RESTORE_INTERRUPTS(_old_)
{
%reg _old_
	mfc0	$8,$12
	and		_old_,_old_,0x1
	and		$8,$8,0XFFFFFFFE
	or		$8,$8,_old_
	mtc0	$8,$12
	nop
	nop
	nop
%mem _old_
	mfc0	$8,$12
	lw		$9,_old_
	and		$9,$9,0x1
	and		$8,$8,0XFFFFFFFE
	or		$8,$8,$9
	mtc0	$8,$12
	nop
	nop
	nop
%error
}

asm void HAL_QUERY_INTERRUPTS(_state_)
{
%reg _state_
	mfc0	_state_,$12
	nop
	and		_state_,_state_,0x1
%mem _state_
	mfc0	$8,$12
	nop
	and		$8,$8,0x1
	sw		S8,_state_
	nop
%error
}

#endif	// END of (CROSS_COMPILER == COMPILER_GREEN_HILLS)


//--------------------------------------------------------------------------
// Routine to execute DSRs using separate interrupt stack

#ifdef  RTIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK
extern void hal_interrupt_stack_call_pending_DSRs(void);
#define HAL_INTERRUPT_STACK_CALL_PENDING_DSRS() \
    hal_interrupt_stack_call_pending_DSRs()

// these are offered solely for stack usage testing
// if they are not defined, then there is no interrupt stack.
#define HAL_INTERRUPT_STACK_BASE rt_interrupt_stack_base
#define HAL_INTERRUPT_STACK_TOP  rt_interrupt_stack
// use them to declare these extern however you want:
//       extern char HAL_INTERRUPT_STACK_BASE[];
//       extern char HAL_INTERRUPT_STACK_TOP[];
// is recommended
#endif

//--------------------------------------------------------------------------
// Vector translation.
// For chained interrupts we only have a single vector though which all
// are passed. For unchained interrupts we have a vector per interrupt.

#ifndef HAL_TRANSLATE_VECTOR

#if defined(RTIMP_HAL_COMMON_INTERRUPTS_CHAIN)

#define HAL_TRANSLATE_VECTOR(_vector_,_index_) (_index_) = 0

#else

#define HAL_TRANSLATE_VECTOR(_vector_,_index_) (_index_) = (_vector_)

#endif

#endif

//--------------------------------------------------------------------------
// Interrupt and VSR attachment macros

#define HAL_INTERRUPT_IN_USE( _vector_, _state_)                          \
    RT_MACRO_START                                                       \
    rt_uint32 _index_;                                                   \
    HAL_TRANSLATE_VECTOR ((_vector_), _index_);                           \
                                                                          \
    if( hal_interrupt_handlers[_index_] == (RT_ADDRESS)HAL_DEFAULT_ISR ) \
        (_state_) = 0;                                                    \
    else                                                                  \
        (_state_) = 1;                                                    \
    RT_MACRO_END

#define HAL_INTERRUPT_ATTACH( _vector_, _isr_, _data_, _object_ )           \
{                                                                           \
    rt_uint32 _index_;                                                     \
    HAL_TRANSLATE_VECTOR( _vector_, _index_ );                              \
                                                                            \
    if( hal_interrupt_handlers[_index_] == (RT_ADDRESS)HAL_DEFAULT_ISR )   \
    {                                                                       \
        hal_interrupt_handlers[_index_] = (RT_ADDRESS)_isr_;               \
        hal_interrupt_data[_index_] = (RT_ADDRWORD)_data_;                 \
        hal_interrupt_objects[_index_] = (RT_ADDRESS)_object_;             \
    }                                                                       \
}

#define HAL_INTERRUPT_DETACH( _vector_, _isr_ )                         \
{                                                                       \
    rt_uint32 _index_;                                                 \
    HAL_TRANSLATE_VECTOR( _vector_, _index_ );                          \
                                                                        \
    if( hal_interrupt_handlers[_index_] == (RT_ADDRESS)_isr_ )         \
    {                                                                   \
        hal_interrupt_handlers[_index_] = (RT_ADDRESS)HAL_DEFAULT_ISR; \
        hal_interrupt_data[_index_] = 0;                                \
        hal_interrupt_objects[_index_] = 0;                             \
    }                                                                   \
}

#define HAL_VSR_GET( _vector_, _pvsr_ )                 \
    *(_pvsr_) = (void (*)())hal_vsr_table[_vector_];


#define HAL_VSR_SET( _vector_, _vsr_, _poldvsr_ ) RT_MACRO_START         \
    if( (void*)_poldvsr_ != NULL)                                         \
        *(RT_ADDRESS *)_poldvsr_ = (RT_ADDRESS)hal_vsr_table[_vector_]; \
    hal_vsr_table[_vector_] = (RT_ADDRESS)_vsr_;                         \
RT_MACRO_END

// This is an ugly name, but what it means is: grab the VSR back to ROS
// internal handling, or if you like, the default handler.  But if
// cooperating with GDB and RtMon, the default behaviour is to pass most
// exceptions to RtMon.  This macro undoes that so that ROS handles the
// exception.  So use it with care.

externC void __default_exception_vsr(void);
externC void __default_interrupt_vsr(void);
externC void __break_vsr_springboard(void);

#define HAL_VSR_SET_TO_ECOS_HANDLER( _vector_, _poldvsr_ ) RT_MACRO_START  \
    HAL_VSR_SET( _vector_, _vector_ == RTNUM_HAL_VECTOR_INTERRUPT          \
                              ? (RT_ADDRESS)__default_interrupt_vsr        \
                              : _vector_ == RTNUM_HAL_VECTOR_BREAKPOINT    \
                                ? (RT_ADDRESS)__break_vsr_springboard      \
                                : (RT_ADDRESS)__default_exception_vsr,     \
                 _poldvsr_ );                                               \
RT_MACRO_END

//--------------------------------------------------------------------------
// Interrupt controller access
// The default code here simply uses the fields present in the CP0 status
// and cause registers to implement this functionality.
// Beware of nops in this code. They fill delay slots and avoid CP0 hazards
// that might otherwise cause following code to run in the wrong state or
// cause a resource conflict.

#ifndef RTHWR_HAL_INTERRUPT_CONTROLLER_ACCESS_DEFINED

#if (CROSS_COMPILER == COMPILER_GNU_GCC)
/*
#define HAL_INTERRUPT_MASK( _vector_ )          \
RT_MACRO_START                                 \
    asm volatile (                              \
        "mfc0   $3,$12\n"                       \
        "la     $2,0x00000400\n"                \
        "sllv   $2,$2,%0\n"                     \
        "nor    $2,$2,$0\n"                     \
        "and    $3,$3,$2\n"                     \
        "mtc0   $3,$12\n"                       \
        "nop; nop; nop\n"                       \
        :                                       \
        : "r"(_vector_)                         \
        : "$2", "$3"                            \
        );                                      \
RT_MACRO_END
*/
#ifdef	__COMPILER_MIPS16
#define HAL_INTERRUPT_MASK( _vector_ )          \
RT_MACRO_START                                 \
    asm volatile (                              \
		"addiu	$29,-24;"\
		"sw		$31,20($29);" \
		"sw		$4,16($29);" \
		"sw		$5,12($29);" \
		"sw		$6,8($29);" \
		"move	$6,%0;"		\
		"jalx ____hal_interrup_mask;nop;"	\
		"lw		$31,20($29);" \
		"lw		$4,16($29);" \
		"lw		$5,12($29);" \
		"lw		$6,8($29);" \
		"addiu	$29,24;"\
        :							         \
        : "r"(_vector_)                        \
        : "$4","5","$31"                        \
        );                                      \
RT_MACRO_END

#else
#define HAL_INTERRUPT_MASK( _vector_ )          \
RT_MACRO_START                                 \
    asm volatile (                              \
        "mfc0   $3,$12\n"                       \
        "la     $2,0x00000400\n"                \
        "sllv   $2,$2,%0\n"                     \
        "nor    $2,$2,$0\n"                     \
        "and    $3,$3,$2\n"                     \
        "mtc0   $3,$12\n"                       \
        "nop; nop; nop\n"                       \
        :                                       \
        : "r"(_vector_)                         \
        : "$2", "$3"                            \
        );                                      \
RT_MACRO_END
#endif

/*
#define HAL_INTERRUPT_UNMASK( _vector_ )        \
RT_MACRO_START                                 \
    asm volatile (                              \
        "mfc0   $3,$12\n"                       \
        "la     $2,0x00000400\n"                \
        "sllv   $2,$2,%0\n"                     \
        "or     $3,$3,$2\n"                     \
        "mtc0   $3,$12\n"                       \
        "nop; nop; nop\n"                       \
        :                                       \
        : "r"(_vector_)                         \
        : "$2", "$3"                            \
        );                                      \
RT_MACRO_END
*/

#ifdef	__COMPILER_MIPS16
#define HAL_INTERRUPT_UNMASK( _vector_ )          \
RT_MACRO_START                                 \
    asm volatile (                              \
		"addiu	$29,-24;"\
		"sw		$31,20($29);" \
		"sw		$4,16($29);" \
		"sw		$5,12($29);" \
		"sw		$6,8($29);" \
		"move	$6,%0;"		\
		"jalx ____hal_interrup_unmask;nop;"	\
		"lw		$31,20($29);" \
		"lw		$4,16($29);" \
		"lw		$5,12($29);" \
		"lw		$6,8($29);" \
		"addiu	$29,24;"\
        :							         \
        : "r"(_vector_)                        \
        : "$4","5","$31"                        \
        );                                      \
RT_MACRO_END
#else
#define HAL_INTERRUPT_UNMASK( _vector_ )        \
RT_MACRO_START                                 \
    asm volatile (                              \
        "mfc0   $3,$12\n"                       \
        "la     $2,0x00000400\n"                \
        "sllv   $2,$2,%0\n"                     \
        "or     $3,$3,$2\n"                     \
        "mtc0   $3,$12\n"                       \
        "nop; nop; nop\n"                       \
        :                                       \
        : "r"(_vector_)                         \
        : "$2", "$3"                            \
        );                                      \
RT_MACRO_END
#endif

/*
#define HAL_INTERRUPT_ACKNOWLEDGE( _vector_ )   \
RT_MACRO_START                                 \
    asm volatile (                              \
        "mfc0   $3,$13\n"                       \
        "la     $2,0x00000400\n"                \
        "sllv   $2,$2,%0\n"                     \
        "nor    $2,$2,$0\n"                     \
        "and    $3,$3,$2\n"                     \
        "mtc0   $3,$13\n"                       \
        "nop; nop; nop\n"                       \
        :                                       \
        : "r"(_vector_)                         \
        : "$2", "$3"                            \
        );                                      \
RT_MACRO_END
*/

#ifdef	__COMPILER_MIPS16
#define HAL_INTERRUPT_ACKNOWLEDGE( _vector_ )          \
RT_MACRO_START                                 \
    asm volatile (                              \
		"addiu	$29,-24;"\
		"sw		$31,20($29);" \
		"sw		$4,16($29);" \
		"sw		$5,12($29);" \
		"sw		$6,8($29);" \
		"move	$6,%0;"		\
		"jalx ____hal_interrup_acknowledge;nop;"	\
		"lw		$31,20($29);" \
		"lw		$4,16($29);" \
		"lw		$5,12($29);" \
		"lw		$6,8($29);" \
		"addiu	$29,24;"\
        :							         \
        : "r"(_vector_)                        \
        : "$4","5","$31"                        \
        );                                      \
RT_MACRO_END
#else
#define HAL_INTERRUPT_ACKNOWLEDGE( _vector_ )   \
RT_MACRO_START                                 \
    asm volatile (                              \
        "mfc0   $3,$13\n"                       \
        "la     $2,0x00000400\n"                \
        "sllv   $2,$2,%0\n"                     \
        "nor    $2,$2,$0\n"                     \
        "and    $3,$3,$2\n"                     \
        "mtc0   $3,$13\n"                       \
        "nop; nop; nop\n"                       \
        :                                       \
        : "r"(_vector_)                         \
        : "$2", "$3"                            \
        );                                      \
RT_MACRO_END
#endif

#endif	/* END of (CROSS_COMPILER == COMPILER_GNU_GCC)	*/

/*
 *	Green Hills compiler can't recognize all the inline assembly syntax of GNU.
 *	In fact, it also defines its own inline assembly macro facility.
 */
#if (CROSS_COMPILER == COMPILER_GREEN_HILLS)

asm void HAL_INTERRUPT_MASK( _vector_ )
{       
%reg _vector_
; # _vector_ is declared as a "register" variable
	mfc0   $3,$12
    la     $2,0x00000400
    sllv   $2,$2,_vector_
    nor    $2,$2,$0
    and    $3,$3,$2
    mtc0   $3,$12
    nop
	nop
	nop        
%mem _vector_

%error
}                             

asm void HAL_INTERRUPT_UNMASK( _vector_ )
{
%con _vector_
; # _vector_ is a constant value
    mfc0	$3,$12
    la		$2,0x00000400
    sll		$2,$2,_vector_
    or		$3,$3,$2
    mtc0	$3,$12
    nop
	nop
	nop
%reg _vector_
; # _vector_ is declared as a "register" variable
    mfc0	$3,$12
    la		$2,0x00000400
    sllv	$2,$2,_vector_
    or		$3,$3,$2
    mtc0	$3,$12
    nop
	nop
	nop
%mem _vector_
	lw		$3,_vector_
    la		$2,0x00000400
    sllv	$2,$2,$3
    mfc0	$3,$12
	nop
    or		$3,$3,$2
    mtc0	$3,$12
    nop
	nop
	nop
%error

}

asm void HAL_INTERRUPT_ACKNOWLEDGE( _vector_ )
{
%reg _vector_
; # _vector_ is declared as a "register" variable
    mfc0   $3,$13
    la     $2,0x00000400
    sllv   $2,$2,_vector_  
    nor    $2,$2,$0
    and    $3,$3,$2
    mtc0   $3,$13
    nop
	nop
	nop
%mem _vector_

%error
}

#endif	/* END of (CROSS_COMPILER == COMPILER_GREEN_HILLS) */


#define HAL_INTERRUPT_CONFIGURE( _vector_, _level_, _up_ )

#define HAL_INTERRUPT_SET_LEVEL( _vector_, _level_ )

#define RTHWR_HAL_INTERRUPT_CONTROLLER_ACCESS_DEFINED

#endif	// END of #ifndef RTHWR_HAL_INTERRUPT_CONTROLLER_ACCESS_DEFINED

//--------------------------------------------------------------------------
// Clock control.
// This code uses the count and compare registers that are present in many
// MIPS variants.
// Beware of nops in this code. They fill delay slots and avoid CP0 hazards
// that might otherwise cause following code to run in the wrong state or
// cause a resource conflict.

#ifndef RTHWR_HAL_CLOCK_CONTROL_DEFINED

externC RT_WORD32 rt_hal_clock_period;
#define RTHWR_HAL_CLOCK_PERIOD_DEFINED

/*
 #-------------------------------------------
 # definitions for timer control
 #

 # parameters
 #
#define TIMER_ID		0
#define TIMER_PERIOD	1000000
#define TIMER_INT		IRQ1

#define CLOCK_INTERRUPT_MASK	(0x0800 << TIMER_INT)

 # interrupt controller
 #
#define ADDR_BASE		0xbd010000
#define ADDR_GIMR		(ADDR_BASE + 0x0)
#define ADDR_GISR		(ADDR_BASE + 0x4)
#define ADDR_IRR		(ADDR_BASE + 0x8)
#define GIMR_TCIE		(1 << 0)
#define GISR_TCIP		(1 << 0)
#define IRR_TCIRS		(1 << 0)
#define IRQ1			0
#define IRQ2			1
#define IRQ3			2
#define IRQ4			3

 # timer/counter registers
 #
#define ADDR_TCDATA		(ADDR_BASE + 0x60 + TIMER_ID * 4)
#define ADDR_TCCNT		(ADDR_BASE + 0x70 + TIMER_ID * 4)
#define ADDR_TCCNR		(ADDR_BASE + 0x50)
#define ADDR_TCIR		(ADDR_BASE + 0x54)

 # timer/counter control register
 #
#define TCCNR_TCEN		(1 << (TIMER_ID * 2))
#define TCCNR_TCMODE	(2 << (TIMER_ID * 2))
#define TCCNR_TCSRC		(1 << (TIMER_ID + 8))

 # timer/counter interrupt register
 #
#define TCIR_TCIE		(1 << TIMER_ID)
#define TCIR_TCIP		(1 << (TIMER_ID + 4))
*/


// clock initialization for RTL8601
// this routine does the following actions:
//	- set timer interrupt routine in IRR
//	- set timer period to the timer data register TCDATA
//	- set timer mode/enable/source in the timer control register TCCNR
//	- enable timer interrupt in the timer interrupt register TCIR
//	- enable timer interrupt in GIMR
//
#if (CROSS_COMPILER == COMPILER_GNU_GCC)
/*
#define HAL_CLOCK_INITIALIZE( _period_ )	\
RT_MACRO_START								\
	asm volatile							\
	(										\
		"li		$10, 0xbd010000;"			\
		"lw		$8, 8($10);"				\
		"li		$9, 0xfffc;"				\
		"and	$8, $9;"					\
		"ori	$8, 0x0;"					\
		"sw		$8, 8($10);"				\
		"sw		%0, 0x60($10);"				\
		"lw		$8, 0x50($10);"				\
		"nop;"								\
		"ori	$8, 0x3;"					\
		"sw		$8, 0x50($10);"				\
		"lw		$8, 0x54($10);"				\
		"nop;"								\
		"ori	$8, 0x1;"					\
		"sw		$8, 0x54($10);"				\
		"lw		$8, 0($10);"				\
		"nop;"								\
		"ori	$8, 0x1;"					\
		"sw		$8, 0($10);"				\
		:									\
		: "r"(_period_)						\
		: "$8", "$9", "$10"					\
	);										\
	rt_hal_clock_period = _period_;		\
RT_MACRO_END


#define HAL_CLOCK_RESET( _vector_, _period_ )	HAL_CLOCK_INITIALIZE( _period_)


// HAL_CLOCK_READ() is used to read the number of hardware clock ticks
// (not the number of timer interrupt) since last timer interrupt
// - xavier
//
#define HAL_CLOCK_READ(_pvalue_)				\
RT_MACRO_START									\
	register RT_WORD32 result;					\
	asm volatile								\
	(											\
		"li		$8, 0xbd010000;"				\
		"lw		$9, 0x70($8);"					\
		"li		$8, 0x00ffffff;"				\
		"and	%0, $8, $9;"					\
		: "=r"(result)							\
		:										\
		: "$8", "$9"							\
	);											\
	*(_pvalue_) = RTNUM_HAL_RTC_PERIOD - result;\
RT_MACRO_END


// #--- acknowledge the timer interrupt
// #--- CLEAR_TIMER0_INT
// - xavier
//
#define HAL_CLOCK_ACKNOWLEDGE()				\
RT_MACRO_START								\
	asm volatile							\
	(										\
		"li		$8, 0xbd010000;"			\
		"lw		$9, 0x54($8);"				\
		"nop;"								\
		"ori	$9, 0x10;"					\
		"sw		$9, 0x54($8);"				\
		:									\
		:									\
		: "$8", "$9"						\
	);										\
RT_MACRO_END
*/
#endif

/*
 *	Green Hills compiler can't recognize all the inline assembly syntax of GNU.
 *	In fact, it also defines its own inline assembly macro facility.
 */
#if (CROSS_COMPILER == COMPILER_GREEN_HILLS)
#define HAL_CLOCK_INITIALIZE( _period_ )	\
RT_MACRO_START								\
	asm volatile							\
	(										\
		"li		$10, 0xbd010000;"			\
		"lw		$8, 8($10);"				\
		"li		$9, 0xfffc;"				\
		"and	$8, $9;"					\
		"ori	$8, 0x0;"					\
		"sw		$8, 8($10);"				\
		"sw		%0, 0x60($10);"				\
		"lw		$8, 0x50($10);"				\
		"nop;"								\
		"ori	$8, 0x3;"					\
		"sw		$8, 0x50($10);"				\
		"lw		$8, 0x54($10);"				\
		"nop;"								\
		"ori	$8, 0x1;"					\
		"sw		$8, 0x54($10);"				\
		"lw		$8, 0($10);"				\
		"nop;"								\
		"ori	$8, 0x1;"					\
		"sw		$8, 0($10);"				\
		:									\
		: "r"(_period_)						\
		: "$8", "$9", "$10"					\
	);										\
	rt_hal_clock_period = _period_;		\
RT_MACRO_END


#define HAL_CLOCK_RESET( _vector_, _period_ )	HAL_CLOCK_INITIALIZE( _period_)


// HAL_CLOCK_READ() is used to read the number of hardware clock ticks
// (not the number of timer interrupt) since last timer interrupt

asm void HAL_CLOCK_READ(_pvalue_)
{
%reg _pvalue_
; # _pvalue_ is declared as a "register" variable
	li		$8, 0xbd010000
	lw		$9, 0x70($8)
	li		$8, 0x00ffffff
	and		_pvalue_, $8, $9
%mem _pvalue_
; # _pvalue_ is declared as a memory variable
	li		$8, 0xbd010000
	lw		$9, 0x70($8)
	li		$8, 0x00ffffff
	and		$9, $9, $8
	sw		$9, _pvalue_
%error
}
/*
#define HAL_CLOCK_READ(_pvalue_)				\
RT_MACRO_START									\
	register RT_WORD32 result;					\
	asm volatile								\
	(											\
		"li		$8, 0xbd010000;"				\
		"lw		$9, 0x70($8);"					\
		"li		$8, 0x00ffffff;"				\
		"and	%0, $8, $9;"					\
		: "=r"(result)							\
		:										\
		: "$8", "$9"							\
	);											\
	*(_pvalue_) = RTNUM_HAL_RTC_PERIOD - result;\
RT_MACRO_END
*/

// #--- acknowledge the timer interrupt
// #--- CLEAR_TIMER0_INT
//
#define HAL_CLOCK_ACKNOWLEDGE()				\
RT_MACRO_START								\
	asm volatile							\
	(										\
		"li		$8, 0xbd010000;"			\
		"lw		$9, 0x54($8);"				\
		"nop;"								\
		"ori	$9, 0x10;"					\
		"sw		$9, 0x54($8);"				\
		:									\
		:									\
		: "$8", "$9"						\
	);										\
RT_MACRO_END

#endif	// END of (CROSS_COMPILER == COMPILER_GREEN_HILLS)


#define RTHWR_HAL_CLOCK_CONTROL_DEFINED

#endif

#if defined(RTVAR_KERNEL_COUNTERS_CLOCK_LATENCY) && \
    !defined(HAL_CLOCK_LATENCY)
#define HAL_CLOCK_LATENCY( _pvalue_ )			HAL_CLOCK_READ(_pvalue_)
#endif


//--------------------------------------------------------------------------
// Microsecond delay function provided in hal_misc.c
externC void hal_delay_us(int us);

#define HAL_DELAY_US(n)          hal_delay_us(n)

//--------------------------------------------------------------------------
#endif // ifndef RTONCE_HAL_HAL_INTR_H
// End of hal_intr.h
