#ifndef RTONCE_HAL_PLF_STUB_H
#define RTONCE_HAL_PLF_STUB_H

//=============================================================================
//
//      plf_stub.h
//
//      Platform header for GDB stub support.
//
//=============================================================================

//=============================================================================
//
// Purpose:     Platform HAL stub support for MIPS/JMR3904 boards.
// Usage:       #include <rt/hal/plf_stub.h>
//
//=============================================================================

/*
#include <include\pkgconf\hal.h>
#include <include\pkgconf\hal_mips_4180_rt8601.h>
*/

#include "../pkgconf/hal.h"

#if defined (RTDBG_HAL_DEBUG_GDB_INCLUDE_STUBS) || defined (RTDBG_HAL_MIPS_DEBUG_GDB_CTRLC_SUPPORT)
// #if defined (RTDBG_HAL_DEBUG_GDB_INCLUDE_STUBS)

/*
#include <include\infra\rt_type.h>         // RT_UNUSED_PARAM

#include <include\hal\mips-stub.h>          // architecture stub support
*/

#include "../infra/rt_type.h"         // RT_UNUSED_PARAM

#include "mips-stub.h"          // architecture stub support
// #include <rt/io/serialio.h>

///////////////////////////////////////////////////////////////////////
// data structures for serial driver
// copied from <rt/io/serialio.h>
// - xavier

// Supported baud rates
typedef enum {
    RTNUM_SERIAL_BAUD_50 = 1,
    RTNUM_SERIAL_BAUD_75,
    RTNUM_SERIAL_BAUD_110,
    RTNUM_SERIAL_BAUD_134_5,
    RTNUM_SERIAL_BAUD_150,
    RTNUM_SERIAL_BAUD_200,
    RTNUM_SERIAL_BAUD_300,
    RTNUM_SERIAL_BAUD_600,
    RTNUM_SERIAL_BAUD_1200,
    RTNUM_SERIAL_BAUD_1800,
    RTNUM_SERIAL_BAUD_2400,
    RTNUM_SERIAL_BAUD_3600,
    RTNUM_SERIAL_BAUD_4800,
    RTNUM_SERIAL_BAUD_7200,
    RTNUM_SERIAL_BAUD_9600,
    RTNUM_SERIAL_BAUD_14400,
    RTNUM_SERIAL_BAUD_19200,
    RTNUM_SERIAL_BAUD_38400,
    RTNUM_SERIAL_BAUD_57600,
    RTNUM_SERIAL_BAUD_115200,
    RTNUM_SERIAL_BAUD_230400
} rt_serial_baud_rate_t;
#define RTNUM_SERIAL_BAUD_MIN RTNUM_SERIAL_BAUD_50 
#define RTNUM_SERIAL_BAUD_MAX RTNUM_SERIAL_BAUD_230400

// Note: two levels of macro are required to get proper expansion.
#define _RT_SERIAL_BAUD_RATE(n) RTNUM_SERIAL_BAUD_##n
#define RT_SERIAL_BAUD_RATE(n) _RT_SERIAL_BAUD_RATE(n)

// Stop bit selections
typedef enum {
    RTNUM_SERIAL_STOP_1 = 1,
    RTNUM_SERIAL_STOP_1_5,
    RTNUM_SERIAL_STOP_2
} rt_serial_stop_bits_t;

// Parity modes
typedef enum {
    RTNUM_SERIAL_PARITY_NONE = 0,
    RTNUM_SERIAL_PARITY_EVEN,
    RTNUM_SERIAL_PARITY_ODD,
    RTNUM_SERIAL_PARITY_MARK,
    RTNUM_SERIAL_PARITY_SPACE
} rt_serial_parity_t;

// Word length
typedef enum {
    RTNUM_SERIAL_WORD_LENGTH_5 = 5,
    RTNUM_SERIAL_WORD_LENGTH_6,
    RTNUM_SERIAL_WORD_LENGTH_7,
    RTNUM_SERIAL_WORD_LENGTH_8
} rt_serial_word_length_t;

typedef struct {
    rt_serial_baud_rate_t   baud;
    rt_serial_stop_bits_t   stop;
    rt_serial_parity_t      parity;
    rt_serial_word_length_t word_length;
    rt_uint32               flags;
} rt_serial_info_t;


//----------------------------------------------------------------------------
// Define serial stuff.
extern void hal_rt8601_init_serial( void );
extern int  hal_rt8601_get_char( void );
extern void hal_rt8601_put_char( int c );
extern void	hal_rt8601_set_baud( int baud );

#define HAL_STUB_PLATFORM_INIT_SERIAL()       hal_rt8601_init_serial()
#define HAL_STUB_PLATFORM_GET_CHAR()          hal_rt8601_get_char()
#define HAL_STUB_PLATFORM_PUT_CHAR(c)         hal_rt8601_put_char((c))
#define HAL_STUB_PLATFORM_SET_BAUD_RATE(baud) hal_rt8601_set_baud(baud)
//#define HAL_STUB_PLATFORM_SET_BAUD_RATE(baud) RT_UNUSED_PARAM(int, (baud))
#define HAL_STUB_PLATFORM_INTERRUPTIBLE       0
#define HAL_STUB_PLATFORM_INIT_BREAK_IRQ()    RT_EMPTY_STATEMENT

//----------------------------------------------------------------------------
// Stub initializer.
#define HAL_STUB_PLATFORM_STUBS_INIT()        RT_EMPTY_STATEMENT


///////////////////////////////////////////////////////////////////////
// definitions for 16x5x serial driver
//

#define RTPRI_IO_SERIAL_GENERIC_16X5X_STEP 4

#define SER_REG(_x_) ((_x_)*RTPRI_IO_SERIAL_GENERIC_16X5X_STEP)

// Receive control Registers
#define REG_rhr SER_REG(0)    // Receive holding register
#define REG_isr SER_REG(2)    // Interrupt status register
#define REG_lsr SER_REG(5)    // Line status register
#define REG_msr SER_REG(6)    // Modem status register
#define REG_scr SER_REG(7)    // Scratch register

// Transmit control Registers
#define REG_thr SER_REG(0)    // Transmit holding register
#define REG_ier SER_REG(1)    // Interrupt enable register
#define REG_fcr SER_REG(2)    // FIFO control register
#define REG_lcr SER_REG(3)    // Line control register
#define REG_mcr SER_REG(4)    // Modem control register
#define REG_ldl SER_REG(0)    // LSB of baud rate
#define REG_mdl SER_REG(1)    // MSB of baud rate

// Interrupt Enable Register
#define IER_RCV 0x01
#define IER_XMT 0x02
#define IER_LS  0x04
#define IER_MS  0x08

// Line Control Register
#define LCR_WL5 0x00    // Word length
#define LCR_WL6 0x01
#define LCR_WL7 0x02
#define LCR_WL8 0x03
#define LCR_SB1 0x00    // Number of stop bits
#define LCR_SB1_5 0x04  // 1.5 -> only valid with 5 bit words
#define LCR_SB2 0x04
#define LCR_PN  0x00    // Parity mode - none
#define LCR_PE  0x0C    // Parity mode - even
#define LCR_PO  0x08    // Parity mode - odd
#define LCR_PM  0x28    // Forced "mark" parity
#define LCR_PS  0x38    // Forced "space" parity
#define LCR_DL  0x80    // Enable baud rate latch

// Line Status Register
#define LSR_RSR 0x01
#define LSR_OE  0x02
#define LSR_PE  0x04
#define LSR_FE  0x08
#define LSR_BI  0x10
#define LSR_THE 0x20
#define LSR_TEMT 0x40
#define LSR_FIE 0x80

// Modem Control Register
#define MCR_DTR 0x01
#define MCR_RTS 0x02
#define MCR_INT 0x08   // Enable interrupts
#define MCR_LOOP 0x10
#define MCR_AFE 0x20	// Auto flow control

// Interrupt status Register
#define ISR_MS  0x00
#define ISR_nIP 0x01
#define ISR_Tx  0x02
#define ISR_LS  0x06
#define ISR_Rx  0x04
#define ISR_RxTO  0x0C

// Modem Status Register
#define MSR_DCTS 0x01
#define MSR_DDSR 0x02
#define MSR_TERI 0x04
#define MSR_DDCD 0x08
#define MSR_CTS  0x10
#define MSR_DSR  0x20
#define MSR_RI   0x40
#define MSR_CD   0x80

// FIFO Control Register
#define FCR_FE   0x01    // FIFO enable
#define FCR_CRF  0x02    // Clear receive FIFO
#define FCR_CTF  0x04    // Clear transmit FIFO
#define FCR_DMA  0x08    // DMA mode select
#define FCR_F64  0x20    // Enable 64 byte fifo (16750+)
#define FCR_RT14 0xC0    // Set Rx trigger at 14
#define FCR_RT8  0x80    // Set Rx trigger at 8
#define FCR_RT4  0x40    // Set Rx trigger at 4
#define FCR_RT1  0x00    // Set Rx trigger at 1

///////////////////////////////////////////////////////////////////////
// prototype of public functions
//

// configure serial hardware
// in
//	new_config: channel parameters, NULL to use default settings
//	init: re-init hardware
// out:
//	true if succeed, false otherwise
//
bool serial_config_port(rt_serial_info_t *new_config, bool init);

// init serial device
// in
//	new_config: channel parameters, NULL to use default settings
// out:
//	true if succeed, false otherwise
//
bool serial_init(rt_serial_info_t* config);

// Send a character to the device output buffer.
// Return 'true' if character is sent to device
bool serial_putc(unsigned char c);

// Fetch a character from the device input buffer, waiting if necessary
unsigned char serial_getc(void);

// check if receive buffer is empty
bool serial_rx_empty(void);

#endif // ifdef RTDBG_HAL_DEBUG_GDB_INCLUDE_STUBS

//-----------------------------------------------------------------------------
#endif // RTONCE_HAL_PLF_STUB_H
// End of plf_stub.h
