#ifndef RTONCE_IMP_CACHE_H
#define RTONCE_IMP_CACHE_H

//=============================================================================
//
//      var_cache.h
//
//      HAL cache control API
//
//=============================================================================
//
// Purpose:     Cache control API
// Description: The macros defined here provide the HAL APIs for handling
//              cache control operations.
// Usage:
//              #include <rt/hal/var_cache.h>
//
//=============================================================================

/*
#include <include/pkgconf/hal.h>
#include <include/infra/rt_type.h>

#include <include/hal/plf_cache.h>
*/

#include "../pkgconf/hal.h"
#include "../infra/rt_type.h"
#include "plf_cache.h"

//=============================================================================
// Toshiba TX3904

#ifdef RTPKG_HAL_MIPS_TX3904

//-----------------------------------------------------------------------------
// Cache dimensions

// Data cache
#define HAL_DCACHE_SIZE                 4096	// Size of data cache in bytes
#define HAL_DCACHE_LINE_SIZE            16		// Size of a data cache line
#define HAL_DCACHE_WAYS                 1		// Associativity of the cache

// Instruction cache
#define HAL_ICACHE_SIZE                 4096    // Size of cache in bytes
#define HAL_ICACHE_LINE_SIZE            16      // Size of a cache line
#define HAL_ICACHE_WAYS                 1       // Associativity of the cache

#define HAL_DCACHE_SETS (HAL_DCACHE_SIZE/(HAL_DCACHE_LINE_SIZE*HAL_DCACHE_WAYS))
#define HAL_ICACHE_SETS (HAL_ICACHE_SIZE/(HAL_ICACHE_LINE_SIZE*HAL_ICACHE_WAYS))

//-----------------------------------------------------------------------------
// Global control of data cache

// Enable the data cache
// This uses a bit in the config register, which is TX39 specific.
#define HAL_DCACHE_ENABLE_DEFINED
#define HAL_DCACHE_ENABLE()
// cache is always enabled in generic R3000
/*
#define HAL_DCACHE_ENABLE()                     \
{                                               \
    asm volatile ("mfc0 $2,$3;"                 \
                  "ori  $2,$2,0x0010;"          \
                  "mtc0 $2,$3;"                 \
                  :                             \
                  :                             \
                  : "$2"                        \
                 );                             \
                                                \
}
*/

// Disable the data cache
#define HAL_DCACHE_DISABLE_DEFINED
#define HAL_DCACHE_DISABLE()
// cache is always enabled in generic R3000
/*
#define HAL_DCACHE_DISABLE()                    \
{                                               \
    asm volatile ("mfc0 $2,$3;"                 \
                  "la   $3,0xFFFFFFEF;"         \
                  "and  $2,$2,$3;"              \
                  "mtc0 $2,$3;"                 \
                  :                             \
                  :                             \
                  : "$2", "$3"                  \
                 );                             \
                                                \
}
*/

// Invalidate the entire cache
// The TX39 only has hit-invalidate on the DCACHE, not
// index-invalidate, so we cannot just empty the cache out without
// knowing what is in it. This is annoying. So, the best we can do is
// fill the cache with data that is unlikely to be there
// otherwise. Hence we read bytes from the ROM space since this is
// most likely to be code, and will not get out of sync even if it is not.
#define HAL_DCACHE_INVALIDATE_ALL_DEFINED

#if (CROSS_COMPILER == COMPILER_GNU_GCC)
/*
#define HAL_DCACHE_INVALIDATE_ALL() \
{                                   \
	asm volatile					\
	(								\
		"ori	$8, $0, 1;"			\
		".set	noreorder;"			\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$8, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		".set	reorder;"			\
		:							\
		:							\
		: "$8"						\
	);								\
}
*/

#ifdef	__COMPILER_MIPS16
#define HAL_DCACHE_INVALIDATE_ALL() \
    RT_MACRO_START              \
	asm volatile				\
	(							\
		"addiu	$29,-24;"		\
		"sw		$31,20($29);"	\
		"sw		$4,16($29);"	\
		"li		$4,1;"			\
		"jalx ____hal_cache_invalidate_all;nop;"	\
		"lw		$31,20($29);"	\
		"lw		$4,16($29);"	\
		"addiu	$29,24;"		\
		".set	reorder;"		\
		:						\
		:						\
		: "$4","$31"			\
	);							\
    RT_MACRO_END
#else

#define HAL_DCACHE_INVALIDATE_ALL() \
{                                   \
	asm volatile					\
	(								\
		"ori	$8, $0, 1;"			\
		".set	noreorder;"			\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$8, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		".set	reorder;"			\
		:							\
		:							\
		: "$8"						\
	);								\
}
#endif



#endif

#if (CROSS_COMPILER == COMPILER_GREEN_HILLS)
asm void HAL_DCACHE_INVALIDATE_ALL(void)
{
	ori		$8, $0, 1
	.set	noreorder
	mtc0	$0, $20
	nop
	nop
	mtc0	$8, $20
	nop
	nop	
	mtc0	$0, $20
	nop
	nop	
	.set	reorder
}
#endif


/*
{																				\
    volatile RT_BYTE *addr = (RT_BYTE *)(0x9fc00000);                 \
    volatile RT_BYTE tmp = 0;                                          \
    int i;                                                              \
    for( i = 0; i < (HAL_DCACHE_SIZE*2); i += HAL_DCACHE_LINE_SIZE )    \
    {                                                                   \
        tmp = addr[i];                                                  \
    }                                                                   \
}
*/

// Synchronize the contents of the cache with memory.
#define HAL_DCACHE_SYNC_DEFINED
#define HAL_DCACHE_SYNC() HAL_DCACHE_INVALIDATE_ALL()

// Set the data cache refill burst size
//#define HAL_DCACHE_BURST_SIZE(_size_)

// Set the data cache write mode
//#define HAL_DCACHE_WRITE_MODE( _mode_ )

//#define HAL_DCACHE_WRITETHRU_MODE       0
//#define HAL_DCACHE_WRITEBACK_MODE       1

// Load the contents of the given address range into the data cache
// and then lock the cache so that it stays there.
/*
#define HAL_DCACHE_LOCK_DEFINED
#define HAL_DCACHE_LOCK(_base_, _size_)         \
{                                               \
    asm volatile ("mfc0 $2,$7;"                 \
                  "ori  $2,$2,0x0100;"          \
                  "mtc0 $2,$7;"                 \
                  :                             \
                  :                             \
                  : "$2"                        \
                 );                             \
}
*/

// Undo a previous lock operation
/*
#define HAL_DCACHE_UNLOCK_DEFINED
#define HAL_DCACHE_UNLOCK(_base_, _size_)       \
{                                               \
    asm volatile ("mfc0 $2,$7;"                 \
                  "la   $3,0xFFFFFEFF;"         \
                  "and  $2,$2,$3;"              \
                  "mtc0 $2,$7;"                 \
                  :                             \
                  :                             \
                  : "$2", "$3"                  \
                 );                             \
}
*/

// Unlock entire cache
/*
#define HAL_DCACHE_UNLOCK_ALL_DEFINED
#define HAL_DCACHE_UNLOCK_ALL() HAL_DCACHE_UNLOCK(0,HAL_DCACHE_SIZE)
*/

//-----------------------------------------------------------------------------
// Data cache line control

// Allocate cache lines for the given address range without reading its
// contents from memory.
//#define HAL_DCACHE_ALLOCATE( _base_ , _size_ )

// Write dirty cache lines to memory and invalidate the cache entries
// for the given address range.
//#define HAL_DCACHE_FLUSH( _base_ , _size_ )
#define HAL_DCACHE_FLUSH_DEFINED        // Ensure no default definition
//
// ?? no HAL_DCACHE_FLUSH()
// the following is added by xavier 11/22/2000
//
#define HAL_DCACHE_FLUSH( _base_ , _asize_ )                                    \
RT_MACRO_START                                                                 \
    register RT_ADDRESS _baddr_ = (RT_ADDRESS)(_base_);                       \
    register RT_ADDRESS _addr_ = (RT_ADDRESS)(_base_);                        \
    register RT_WORD _size_ = (_asize_);                                       \
    register RT_WORD _state_;                                                  \
    register RT_WORD _ori_c0_sr_;												\
    HAL_DCACHE_IS_ENABLED( _state_ );                                           \
    if( _state_ ) {                                                             \
		asm volatile															\
		(																		\
			"lui	$13, 0x1;"													\
			".set	noreorder;"													\
			"mfc0	%0, $12;"													\
			"nop;"																\
			"or		$9, %0, $13;"												\
			"mtc0	$9, $12;"													\
			".set	reorder;"													\
			: "=r"(_ori_c0_sr_)													\
			:																	\
			: "$9", "$13"														\
		);																		\
        for( ; _addr_ <= _baddr_+_size_; _addr_ += HAL_DCACHE_LINE_SIZE )       \
        {																		\
        	asm volatile ("sb $0, 0(%0)" : : "r"(_addr_) );						\
        }																		\
		asm volatile ("mtc0 %0, $12" : : "r"(_ori_c0_sr_) );					\
    }                                                                           \
RT_MACRO_END

// the following segment is added by xavier 11/22/2000
//
#define HAL_DCACHE_INVALIDATE_DEFINED
#define HAL_DCACHE_INVALIDATE( _base_ , _asize_ )                               \
RT_MACRO_START                                                                 \
    register RT_ADDRESS _baddr_ = (RT_ADDRESS)(_base_);                       \
    register RT_ADDRESS _addr_ = (RT_ADDRESS)(_base_);                        \
    register RT_WORD _size_ = (_asize_);                                       \
    register RT_WORD _state_;                                                  \
    register RT_WORD _ori_c0_sr_;												\
    HAL_DCACHE_IS_ENABLED( _state_ );                                           \
    if( _state_ ) {                                                             \
		asm volatile															\
		(																		\
			"lui	$13, 0x1;"													\
			".set	noreorder;"													\
			"mfc0	%0, $12;"													\
			"nop;"																\
			"or		$9, %0, $13;"												\
			"mtc0	$9, $12;"													\
			".set	reorder;"													\
			: "=r"(_ori_c0_sr_)													\
			:																	\
			: "$9", "$13"														\
		);																		\
        for( ; _addr_ <= _baddr_+_size_; _addr_ += HAL_DCACHE_LINE_SIZE )       \
        {																		\
        	asm volatile ("sb $0, 0(%0)" : : "r"(_addr_) );						\
        }																		\
		asm volatile ("mtc0 %0, $12" : : "r"(_ori_c0_sr_) );					\
    }                                                                           \
RT_MACRO_END

// Write dirty cache lines to memory for the given address range.
//#define HAL_DCACHE_STORE( _base_ , _size_ )
#define HAL_DCACHE_STORE_DEFINED        // Disable default definition

//-----------------------------------------------------------------------------
// Global control of Instruction cache

// Enable the instruction cache
// This uses a bit in the config register, which is TX39 specific.
#define HAL_ICACHE_ENABLE_DEFINED
#define HAL_ICACHE_ENABLE()
/*
// cache is always enabled in generic R3000
#define HAL_ICACHE_ENABLE()                     \
{                                               \
    asm volatile ("mfc0 $2,$3;"                 \
                  "ori  $2,$2,0x0020;"          \
                  "mtc0 $2,$3;"                 \
                  :                             \
                  :                             \
                  : "$2"                        \
                 );                             \
                                                \
}
*/

// Disable the instruction cache
#define HAL_ICACHE_DISABLE_DEFINED
#define HAL_ICACHE_DISABLE()
/*
// cache is always enabled in generic R3000
#define HAL_ICACHE_DISABLE()                    \
{                                               \
    asm volatile ("mfc0 $2,$3;"                 \
                  "la   $3,0xFFFFFFDF;"         \
                  "and  $2,$2,$3;"              \
                  "mtc0 $2,$3;"                 \
                  "j    1f;"                    \
                  "nop;"                        \
                  ".balign 16,0;"               \
                  "1:;"                         \
                  :                             \
                  :                             \
                  : "$2", "$3"                  \
                 );                             \
                                                \
}
*/

// Load the contents of the given address range into the instruction cache
// and then lock the cache so that it stays there.
/*
#define HAL_ICACHE_LOCK_DEFINED
#define HAL_ICACHE_LOCK(_base_, _size_)         \
{                                               \
    asm volatile ("mfc0 $2,$7;"                 \
                  "ori  $2,$2,0x0200;"          \
                  "mtc0 $2,$7;"                 \
                  :                             \
                  :                             \
                  : "$2"                        \
                 );                             \
}
*/

// Undo a previous lock operation
/*
#define HAL_ICACHE_UNLOCK_DEFINED
#define HAL_ICACHE_UNLOCK(_base_, _size_)       \
{                                               \
    asm volatile ("mfc0 $2,$7;"                 \
                  "la   $3,0xFFFFFDFF;"         \
                  "and  $2,$2,$3;"              \
                  "mtc0 $2,$7;"                 \
                  :                             \
                  :                             \
                  : "$2", "$3"                  \
                 );                             \
}
*/

// Unlock entire cache
/*
#define HAL_ICACHE_UNLOCK_ALL_DEFINED
#define HAL_ICACHE_UNLOCK_ALL() HAL_ICACHE_UNLOCK(0, HAL_ICACHE_SIZE)
*/

//-----------------------------------------------------------------------------
// Instruction cache line control
// On the TX39, the instruction cache must be disabled to use the index-invalidate
// cache operation.

// Invalidate the entire cache
// This uses the index-invalidate cache operation.
#define HAL_ICACHE_INVALIDATE_ALL_DEFINED

#if (CROSS_COMPILER == COMPILER_GNU_GCC)
/*
#define HAL_ICACHE_INVALIDATE_ALL() \
    RT_MACRO_START                 \
	asm volatile					\
	(								\
		"ori	$8, $0, 2;"			\
		".set	noreorder;"			\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$8, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		".set	reorder;"			\
		:							\
		:							\
		: "$8"						\
	);								\
    RT_MACRO_END
*/

#ifdef	__COMPILER_MIPS16
#define HAL_ICACHE_INVALIDATE_ALL() \
    RT_MACRO_START              \
	asm volatile				\
	(							\
		"addiu	$29,-24;"		\
		"sw		$31,20($29);"	\
		"sw		$4,16($29);"	\
		"li		$4,2;"			\
		"jalx ____hal_cache_invalidate_all;nop;"	\
		"lw		$31,20($29);"	\
		"lw		$4,16($29);"	\
		"addiu	$29,24;"		\
		".set	reorder;"		\
		:						\
		:						\
		: "$4","$31"			\
	);							\
    RT_MACRO_END
#else
#define HAL_ICACHE_INVALIDATE_ALL() \
    RT_MACRO_START                 \
	asm volatile					\
	(								\
		"ori	$8, $0, 2;"			\
		".set	noreorder;"			\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$8, $20;"			\
		"nop;"						\
		"nop;"						\
		"mtc0	$0, $20;"			\
		"nop;"						\
		"nop;"						\
		".set	reorder;"			\
		:							\
		:							\
		: "$8"						\
	);								\
    RT_MACRO_END
#endif

#endif

#if (CROSS_COMPILER == COMPILER_GREEN_HILLS)
asm void HAL_ICACHE_INVALIDATE_ALL(void)
{
	ori		$8,$0,2
	.set	noreorder
	mtc0	$0,$20
	nop
	nop
	mtc0	$8,$20
	nop
	nop
	mtc0	$0,$20
	nop
	nop
	.set	reorder
}
#endif

/*
{                                                                               \
    register RT_ADDRESS _baddr_ = 0x80000000;                                  \
    register RT_ADDRESS _addr_ = 0x80000000;                                   \
    register RT_WORD _state_;                                                  \
    HAL_ICACHE_IS_ENABLED(_state_);                                             \
    HAL_ICACHE_DISABLE();                                                       \
    for( ; _addr_ < _baddr_+HAL_ICACHE_SIZE; _addr_ += HAL_ICACHE_LINE_SIZE )   \
    {                                                                           \
        asm volatile ("cache 0x00,0(%0)" : : "r"(_addr_) );                     \
    }                                                                           \
    if( _state_ ) HAL_ICACHE_ENABLE();                                          \
}
*/

// Invalidate cache lines in the given range without writing to memory.
// This uses the index-invalidate cache operation since the TX39 does not
// have hit-invalidate on the instruction cache.
#define HAL_ICACHE_INVALIDATE_DEFINED
#define HAL_ICACHE_INVALIDATE( _base_ , _asize_ )                       \
{                                                                       \
    register RT_ADDRESS _baddr_ = (RT_ADDRESS)(_base_);               \
    register RT_ADDRESS _addr_ = (RT_ADDRESS)(_base_);                \
    register RT_WORD _size_ = (_asize_);                               \
    register RT_WORD _ori_c0_sr_;										\
	asm volatile														\
	(																	\
		"lui	$13, 0x3;"												\
		".set	noreorder;"												\
		"mfc0	%0, $12;"												\
		"nop;"															\
		"or		$9, %0, $13;"											\
		"mtc0	$9, $12;"												\
		".set	reorder;"												\
		: "=r"(_ori_c0_sr_)												\
		:																\
		: "$9", "$13"													\
	);																	\
    for( ; _addr_ <= _baddr_+_size_; _addr_ += HAL_ICACHE_LINE_SIZE )   \
    {																	\
    	asm volatile ("sb $0, 0(%0)" : : "r"(_addr_) );				\
    }																	\
	asm volatile ("mtc0 %0, $12" : : "r"(_ori_c0_sr_) );				\
}
/*																		\
    register RT_ADDRESS _addr_ = (RT_ADDRESS)(_base_);                \
    register RT_WORD _size_ = (_asize_);                               \
    register RT_WORD _state_;                                          \
    HAL_ICACHE_IS_ENABLED(_state_);                                     \
    HAL_ICACHE_DISABLE();                                               \
    for( ; _addr_ <= _addr_+_size_; _addr_ += HAL_ICACHE_LINE_SIZE )    \
    {                                                                   \
        asm volatile ("cache 0,0(%0)" : : "r"(_addr_) );                \
    }                                                                   \
    if( _state_ ) HAL_ICACHE_ENABLE();                                  \
}
*/

#endif // RTPKG_HAL_MIPS_TX3904

//-----------------------------------------------------------------------------
#endif // ifndef RTONCE_IMP_CACHE_H
// End of imp_cache.h
