/*----------------------------------------------------------------------------------------------
*
*  TFTP server for Realtek Router Bootloader
*
*  This is a naive implementation of TFTP Server, It is just support TFTP Client WRQ .
* 
*  Any question,please mail to alva_zhang@realsil.com.cn .             09/19/2005
*
*-----------------------------------------------------------------------------------------------
*/
//#include "stdlib.h"
#include <linux/autoconf.h>
#include "tftpnaive.h"
#include "net.h"

//random generation
// additive congruential random number generators
//        (see R. Sedgewick, Algorithms, Chapter 35)
//
#ifdef CONFIG_RTL865X_LOADER_SUPPORT_TFTP_SERVER
#define TFTPD_RANDOM_PORT 0
#if TFTPD_RANDOM_PORT

#define m1    10000
#define m    100000000

static unsigned long j;
static unsigned long a[55];

static unsigned long mult(unsigned long p, unsigned long q)
{
  unsigned long p1,p0,q1,q0;
  
  p1 = p/m1; p0 = p % m1;
  q1 = q/m1; q0 = q % m1;
  return((((p0*q1 + p1*q0) % m1) * m1 + p0*q0) % m);
}

void srandom(unsigned long s)
{
  a[0] = s;
  j = 0;
  do 
    {
      ++j;
      a[j] = (mult(31,a[j-1]) + 1) % m;
    } while (j<54);
}

/* additive congruential method    
*    random() returns an unsigned long random number in the range 0 to r-1
*    The random number generator is initialized by srandom()
*/
unsigned long random(unsigned long r)
{
  int x,y;
  
  j = (j + 1) % 55;
  x = (j+23)%55;
  y = (j+54)%55;
  a[j] = (a[x] + a[y]) % m;
  return( ((a[j] / m1) * r) / m1 );
}

#endif

//This function process TFTP server command transfer.
//Here we just accept WRQ because there is NONE file manage system
//so can only put the file received into buffer and can't find that file 
//client request. Any REQ will be responsed with an error(no1) message by server.
// note 1: No retranmitting for ACK WRQ here.
// note 2: server port(69) is binded with TFTP Command(REQ/WRQ),any other 
//         transaction will be considered as an error message
void processTFTPC (TFTPDpkt * tp)
{
  TFTPDpkt tftpd_pkt;
  unsigned int pkg_len;
  unsigned int dport;
  register int nbtry;
  
  memset (&tftpd_pkt, 0, sizeof (tftpd_pkt));
  
  if(tftpd_recd->status==IDLE)
    { 
      
#if TFTPD_RANDOM_PORT       
      // generation a random TFTP server port with the seed of client port
      dport= tp->udphdr.src_port;
      for(; ;)
        {         
          srandom(dport);
          dport=random(65536);
          if((dport!= tp->udphdr.src_port)&&(dport!= TFTP_SRV_PORT)&&(dport>69))
            break;
        }
#else

      dport = TFTPD_DEFAULT_PORT ; //defined in file net.h      
#endif      
      
      if(tp->opcode==WRQ)
        {
          tftpd_recd->status=WRP;      
          tftpd_recd->dport = dport; 
          tftpd_recd->client=tp->udphdr.iphdr.src_ip;
          tftpd_recd->cport = tp->udphdr.src_port;        
          
          tftpd_pkt.opcode = ACK;
          tftpd_pkt.num = 0; // for the WRQ PKG,the first ackpkg blkno=0;
          pkg_len=4;
		  printf("Begin to Transfer Data from Client to TFTP Server(Board RAM) ......\n");
        }
      
      else if(tp->opcode==REQ)
        {
          // here should prepare tftp_data_pkg & send it
          // because there is not filesystem in the bootloader
          // we can NOT get the file client want to get, so we send
          // a error packet indicate that file can't find.
          tftpd_pkt.opcode = ERROR;
          tftpd_pkt.num = 1; // error 1 indicate that file can't be found;
          pkg_len=5;
        }
      else // include ACK/ERROR/DATA and other undefine opcode
        { // do some  error handle;
          tftpd_pkt.opcode = ERROR;
          tftpd_pkt.num = 4; // error 4            
          pkg_len = 5;
        }
    } //end of "if(tftpd_recd->status==IDLE)"
  else // during the process of Write/Read transfer between Client and Host,    
    {  // Any other command will be considered as an error message .
      tftpd_pkt.opcode = ERROR;
      tftpd_pkt.num = 2; // error 2 indicate Access violation;          
      pkg_len = 5;

      memset(tftpd_recd,0,sizeof(TFTPD_Record));
      tftpd_recd->terminate=2;
    }

  for (nbtry = 5; nbtry; --nbtry) 
    {
      if (sendUDP((UDPpkt *) &tftpd_pkt, pkg_len, tp->udphdr.iphdr.src_ip,
                  dport,tp->udphdr.src_port) == 0)
        break;
      tick_Delay100ms (1);
    }
  
  if (nbtry == 0) printf ("TFTP: error sending ACK\n");                  
}


// this function process TFTP server data transfer
// here we just process that client write data to server
// and any error will result in transfer termination  
void processTFTPD (TFTPDpkt * tp)
{
  register int nbtry;
  TFTPDpkt tftpd_pkt;
  unsigned short pkg_len;
  memset (&tftpd_pkt, 0, sizeof tftpd_pkt);
  //check the packet format
  /*  
  tftpd_pkt.opcode = ERROR;
  pkg_len = 5;
  
  if(tp->udphdr.iphdr.src_ip!=tftpd_recd->client)
    {
      tftpd_pkt.num = 7; // error 7 ,No such user          
      tftpd_recd->terminate=3;
    }
  if(tp->udphdr.src_port!=tftpd_recd->cport)
    {
      tftpd_pkt.num = 5; // error 5 indicate unkowned TID;          
      tftpd_recd->terminate=4;
    }
  
  tftpd_recd->timeout=0;
  
  if((tp->opcode==REQ)||(tp->opcode==WRQ))
    {
      tftpd_pkt.num = 4; // error 4 
      tftpd_recd->terminate=5;
    }
  if(tp->opcode==ERROR)
    {
      tftpd_recd->terminate=6;
    }
  */      

  tftpd_recd->timeout=0;

  if(tftpd_recd->status ==WRP)
    {
      if(tp->opcode==DATA)
        {
         //if((tp->udphdr.iphdr.src_ip==tftpd_recd->client)&&(tp->udphdr.src_port==tftpd_recd->cport))
         //Since the TFTP Server just service 4 a intrannet client,It need NOT check the client's ip 
          if(tp->udphdr.src_port==tftpd_recd->cport)
            {
              if(tp->num == (tftpd_recd->blkno + 1))
                {
                  register unsigned short len = tp->udphdr.length - 12;
                  if(len)
                    {
                      memcpy ((void *) (tftpd_recd->buff + tftpd_recd->bcnt), tp->data, len);
                      tftpd_recd->bcnt = tftpd_recd->bcnt + len;
                    }
                  tftpd_recd->blkno++ ;           
                  if(len<512) //TFTP transfer terminate
                    {
                      tftpd_recd->terminate= 20; //normal terminate 
                      printf ("OK,Transfer Data Completed,TFTP Client UpLoad %d Bytes\n",tftpd_recd->bcnt);
                    }                                
                  tftpd_pkt.opcode = ACK;
                  tftpd_pkt.num = tftpd_recd->blkno; 
                  pkg_len = 4;      
                  tftpd_recd->retry=0;
                } //end of "if(tp->num == tftpd_recd->blkno+1)"
              //receive a packet just stored, it indicate that the pkg server sended
              //has lost, so we have to retransmit again.
              else if(tp->num == tftpd_recd->blkno)
                {
                  tftpd_pkt.opcode = ACK;
                  tftpd_pkt.num = tftpd_recd->blkno; 
                  pkg_len = 4;      
                  
                  //  pkg retransmit
                  if( tftpd_recd->retry<6){ tftpd_recd->retry++; }
                  else { tftpd_recd->terminate=1; }
                } // end of "else if(tp->num == tftpd_recd->blkno)"
              else 
                {
                  tftpd_pkt.opcode = ERROR;
                  tftpd_pkt.num = 4; // error 4 
                  pkg_len = 5;      
                  tftpd_recd->terminate=5;
                }
            }//end of "if((tp->udphdr.iphdr.src_ip==tftpd_recd->client)||
             //           (tp->udphdr.src_port==tftpd_recd->cport))"
          else
            {
              tftpd_pkt.opcode = ERROR;
              tftpd_pkt.num = 7; // error 7 ,No such user          
              pkg_len = 5;
              tftpd_recd->terminate=6;
            }
        } //end of "if(tp->opcode==DATA)"
      else
        {
          tftpd_pkt.opcode = ERROR;
          tftpd_pkt.num = 4; // error 4 
          pkg_len = 5;      
          tftpd_recd->terminate=7;                  
        }
    } //end of "if(tftpd_recd->status ==WRP)"
  else 
    {
      tftpd_pkt.opcode = ERROR;
      tftpd_pkt.num = 4; // error 4 
      pkg_len = 5;      
      tftpd_recd->terminate=8;                  
    }
  
  // send pkg
  for (nbtry = 5; nbtry; --nbtry) 
    {
      if (sendUDP((UDPpkt *) & tftpd_pkt, pkg_len, tftpd_recd->client,
                  tftpd_recd->dport,tftpd_recd->cport) == 0)
        break;
      tick_Delay100ms (1);
    }
  if (nbtry == 0) printf ("TFTP: error sending ACK\n");                   
}


void reSendAck(void)
{
  register int nbtry;
  TFTPDpkt tftpd_pkt;
  
  tftpd_pkt.opcode = ACK;
  tftpd_pkt.num = tftpd_recd->blkno;  
  
  for (nbtry = 5; nbtry; --nbtry) 
    {
      if (sendUDP((UDPpkt *) & tftpd_pkt, 4, tftpd_recd->client,
                  tftpd_recd->dport,tftpd_recd->cport) == 0)
        break;
      tick_Delay100ms (1);
    }
  if (nbtry == 0) printf ("TFTP: error sending ACK\n");         
}

#endif
